/*
 *  PINT Pint Is Not TWAIN - common scanner driver interface for UN*X
 *  Copyright (C) 1994 Kenneth Stailey ken@spacenet.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * scandd.b.c: bottom half of PINT
 *
 * This file contains the bottom half of the device driver.
 * All functions and variables are pinned into memory when the scanner is 
 * opened.
 * There is no stratagy() since this would have made the pinned code larger
 * than the pageable code.
 *
 * routines in this file:
 *
 *  scan_scsi_cmd
 *  scaniodone
 *  adapter_error
 *  scanwait
 *
 * ken@spacenet.com
 *
 * third release, version 0.3
 *
 */

#define _KERNEL		   /* let includes know that we are in the kernel */

#include <sys/types.h>
#include <sys/device.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/scsi.h>
#include <sys/iostat.h>
#include <sys/watchdog.h>
#include <sys/sleep.h>
#include "scandd.h"

/*
 *  global data
 */
struct scan_hash_table_df scanner_list[SCANNER_HASHSIZE] = { NULL };
int scanner_chain_lock = LOCK_AVAIL; /* for exclusive access to scanner_list */
int scanner_config_count = 0;	     /* # of devices configured		 */

uchar wd[4][256 + 8];		/* space for the SET_WINDOW data	  */
				/* (the FS1 needs the biggest wd so far)  */

/*
 * scan_scsi_cmd:  send a given command to the scanner
 *
 * calling sequence:
 * scan{open|read} -> scan_scsi_cmd -> sc_strategy -> iodone -> scaniodone
 *
 * 1) fill in the sc_buf struct
 *   a) forge a buf struct
 * 2) call devstrat()
 */
void
scan_scsi_cmd(uchar cmd, struct scan_df *scan_ptr)
{
  char   escape_codes[256];	/* string of control codes for jx600, etc */

  /* for SCSI_MODE_SELECT command... */
  struct ricoh_mode_params rmp;

  int cs;			/* Current Scanner (array index into wd) */
  int i;

  /* for the SCSI_TRIGGER_SCAN command */
  static char window_zero[1] = { '\0' };

#ifdef DEBUG_SCSI_CMD
    printf("scan_scsi_cmd: cmd = 0x%x\n", cmd);
#endif

  /* forge a buf struct (inside sc_buf struct) with default values */

  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_READ;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_forw  = NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_back  = NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.av_forw = NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.av_back = NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_iodone = scaniodone;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_vp = NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_dev = scan_ptr->s_dds.parent_devno;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_blkno = (daddr_t)0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_error = (char)0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_resid = (uint)0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_options = 0;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;

  scan_ptr->cmd_buf.scsi_buf.bp = NULL;

  /* fill in the length of the command with default of 6 (override if wrong) */
  scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)6;

  scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_id =
    (uchar)scan_ptr->s_dds.scsi_id;
  scan_ptr->cmd_buf.scsi_buf.scsi_command.resvd = (uchar)0;

  /* fill in the flags field */

  if (scan_ptr->retry) {	/* is this command a retry? */
#ifdef DEBUG_SCSI_CMD
    printf("scan_scsi_cmd: setting SC_RESUME flag\n");
#endif
    scan_ptr->cmd_buf.scsi_buf.flags |= (uchar)SC_RESUME;
  } else {
#ifdef DEBUG_SCSI_CMD
    printf("scan_scsi_cmd: NOT setting SC_RESUME flag\n");
#endif
    scan_ptr->cmd_buf.scsi_buf.flags &= ~(uchar)SC_RESUME;
  }

  /* the Ricoh FS1 & IS50 don't support sync i/o at all */
  if (scan_ptr->s_dds.scanner_type == RICOH_FS1  ||
      scan_ptr->s_dds.scanner_type == RICOH_IS50 ||
      scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
      scan_ptr->s_dds.scanner_type == UMAX_UG630 ||
      scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC) {
#ifdef DEBUG_SCSI_CMD
    printf("scan_scsi_cmd: setting SC_ASYNC flag\n");
#endif
    scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= (uchar)SC_ASYNC;
  }

  scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_op_code = cmd;

  scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.lun =
    scan_ptr->s_dds.lun << 5;

  /* command-specific portion of command block */
  switch (cmd) {
  case SCSI_RESERVE_UNIT:
  case SCSI_RELEASE_UNIT:
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    /* SCB4 is the transfer length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    break;

  case SCSI_REQUEST_SENSE:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)&scan_ptr->rs;

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_FS1:
    case RICOH_IS50:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 12;
      break;
    case RICOH_IS410:
    case IBM_2456:
    case FUJITSU:
    case SHARP_JX600:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount =
	(uint)sizeof(struct scan_req_sense_data);
      break;
    case UMAX_UC630:
    case UMAX_UG630:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 18;
      break;

    case HP_SCANJET_IIC:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 8;
      break;

#ifdef DEBUG_SCSI_CMD
    default:
      printf("scan_scsi_cmd: unsupported scanner type %d\n",
	     scan_ptr->s_dds.scanner_type);
#endif
    }

    /* bytes 2-3 are reserved */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;

    /* byte 4 is the allocation length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] =
      (uchar)sizeof(struct scan_req_sense_data);

    /* byte 5 is the "control byte" whatever that means */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    break;

  case SCSI_SET_WINDOW:

    cs = minor(scan_ptr->devno);

    /****   fill in the window descriptor arrays  ***/

    /* first eight bytes are the header */
    wd[cs][WDH_RESERVED0] = (uchar)0;
    wd[cs][WDH_RESERVED1] = (uchar)0;
    wd[cs][WDH_RESERVED2] = (uchar)0;
    wd[cs][WDH_RESERVED3] = (uchar)0;
    wd[cs][WDH_RESERVED4] = (uchar)0;
    wd[cs][WDH_RESERVED5] = (uchar)0;

    switch  (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_MSB] = (uchar)0;
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_LSB] = (uchar)64;
      break;

    case RICOH_IS410:		/* 320 bytes = 0x0140 */
    case IBM_2456:
    case RICOH_IS50:
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_MSB] = (uchar)0x01;
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_LSB] = (uchar)0x40;
      break;

    case RICOH_FS1:		/* 256 (0x0100) */
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_MSB] = (uchar)0x01;
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_LSB] = (uchar)0x00;
      break;

    case UMAX_UC630:		/* 46 (0x002e) */
    case UMAX_UG630:		/* 46 (0x002e) */
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_MSB] = (uchar)0x00;
      wd[cs][WDH_WINDOW_DESCRIPTOR_BLOCK_LENGTH_LSB] = (uchar)0x2e;
      break;
   }

    wd[cs][WD_WINDOW_ID]	  = (uchar)0;
    wd[cs][WD_RESERVED00]	  = (uchar)0; /* auto bit is unsupported */

    wd[cs][WD_X_RES_MSB] = (uchar)(scan_ptr->s_dds.x_resolution >> 8 & 0xff);
    wd[cs][WD_X_RES_LSB] = (uchar)(scan_ptr->s_dds.x_resolution & 0xff);

    wd[cs][WD_Y_RES_MSB] = (uchar)(scan_ptr->s_dds.y_resolution>>8 & 0xff);
    wd[cs][WD_Y_RES_LSB] = (uchar)(scan_ptr->s_dds.y_resolution & 0xff);

    wd[cs][WD_UPPER_LEFT_X_MSB] =
      (uchar)(scan_ptr->s_dds.x_origin>>24 & 0xff);
    wd[cs][WD_UPPER_LEFT_X_MHB] =
      (uchar)(scan_ptr->s_dds.x_origin>>16 & 0xff);
    wd[cs][WD_UPPER_LEFT_X_MLB] =
      (uchar)(scan_ptr->s_dds.x_origin>> 8 & 0xff);
    wd[cs][WD_UPPER_LEFT_X_LSB] =
      (uchar)(scan_ptr->s_dds.x_origin & 0xff);

    wd[cs][WD_UPPER_LEFT_Y_MSB] =
      (uchar)(scan_ptr->s_dds.y_origin>>24 & 0xff);
    wd[cs][WD_UPPER_LEFT_Y_MHB] =
      (uchar)(scan_ptr->s_dds.y_origin>>16 & 0xff);
    wd[cs][WD_UPPER_LEFT_Y_MLB] =
      (uchar)(scan_ptr->s_dds.y_origin>> 8 & 0xff);
    wd[cs][WD_UPPER_LEFT_Y_LSB] =
      (uchar)(scan_ptr->s_dds.y_origin & 0xff);

    wd[cs][WD_WIDTH_MSB] = (uchar)(scan_ptr->s_dds.window_width>>24 & 0xff);
    wd[cs][WD_WIDTH_MHB] = (uchar)(scan_ptr->s_dds.window_width>>16 & 0xff);
    wd[cs][WD_WIDTH_MLB] = (uchar)(scan_ptr->s_dds.window_width>> 8 & 0xff);
    wd[cs][WD_WIDTH_LSB] = (uchar)(scan_ptr->s_dds.window_width & 0xff);

    wd[cs][WD_LENGTH_MSB] = (uchar)(scan_ptr->s_dds.window_length>>24 & 0xff);
    wd[cs][WD_LENGTH_MHB] = (uchar)(scan_ptr->s_dds.window_length>>16 & 0xff);
    wd[cs][WD_LENGTH_MLB] = (uchar)(scan_ptr->s_dds.window_length>> 8 & 0xff);
    wd[cs][WD_LENGTH_LSB] = (uchar)(scan_ptr->s_dds.window_length & 0xff);

    switch  (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:
    case RICOH_FS1:
      wd[cs][WD_BRIGHTNESS] = (uchar)0x00; /* unsupported */
      break;

    case RICOH_IS410:
    case IBM_2456:
    case UMAX_UC630:
    case UMAX_UG630:
      wd[cs][WD_BRIGHTNESS] = (uchar)(scan_ptr->s_dds.brightness);
      break;

    case RICOH_IS50:

      /* if you throw the MRIF bit the brighness control reverses too */
      /* so I reverse the reversal in software for symmetry's sake */

      switch (scan_ptr->s_dds.image_mode) {
      case GRAYSCALE:
      case DITHERED_MONOCHROME:
	wd[cs][WD_BRIGHTNESS] = 256 - (uchar)(scan_ptr->s_dds.brightness);
	break;

      default:
	wd[cs][WD_BRIGHTNESS] = (uchar)(scan_ptr->s_dds.brightness);
	break;
      }
    }

    switch  (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:
      if (scan_ptr->s_dds.image_mode == BINARY_MONOCHROME)
	wd[cs][WD_THRESHOLD]  = (uchar)(scan_ptr->s_dds.brightness);
      else
	wd[cs][WD_THRESHOLD]  = (uchar)0x00; /* dynamic threshold */
      break;

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
      wd[cs][WD_THRESHOLD]  = (uchar)0x00; /* dynamic threshold */
      break;

    case RICOH_FS1:
      wd[cs][WD_THRESHOLD]  = (uchar)0x01; /* "normal" threshold */

    case UMAX_UC630:		/* follow brightness control */
    case UMAX_UG630:
      wd[cs][WD_THRESHOLD]  =  (uchar)(scan_ptr->s_dds.brightness);
    }

    switch  (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:
      wd[cs][WD_CONTRAST]	= (uchar)0;
      break;

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_FS1:
    case UMAX_UC630:
    case UMAX_UG630:
      wd[cs][WD_CONTRAST]	= (uchar)(scan_ptr->s_dds.contrast);
      break;

    case RICOH_IS50:

      /* if you throw the MRIF bit the contrast control reverses too */
      /* so I reverse the reversal in software for symmetry's sake */

      switch (scan_ptr->s_dds.image_mode) {
      case GRAYSCALE:
      case DITHERED_MONOCHROME:
	wd[cs][WD_CONTRAST] = 256 - (uchar)(scan_ptr->s_dds.contrast);
	break;

      default:
	wd[cs][WD_CONTRAST] = (uchar)(scan_ptr->s_dds.contrast);
	break;
      }
    }

    switch (scan_ptr->s_dds.image_mode) {
    case RED:
    case GREEN:
    case BLUE:
      wd[cs][WD_IMAGE_COMPOSTION] = GRAYSCALE;
      break;

    default:
      wd[cs][WD_IMAGE_COMPOSTION] = scan_ptr->s_dds.image_mode;
    }

    switch (scan_ptr->s_dds.image_mode) {
    case GRAYSCALE:
    case RED:
    case GREEN:
    case BLUE:
      wd[cs][WD_BITS_PER_PIXEL]	  = (uchar)8;
      break;

    case DITHERED_MONOCHROME:
    case BINARY_MONOCHROME:
      wd[cs][WD_BITS_PER_PIXEL]	  = (uchar)1;
      break;

    case COLOR:
      wd[cs][WD_BITS_PER_PIXEL]	  = (uchar)24;
      break;
    }

    switch  (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:
      wd[cs][WDF_HALFTONE_PATTERN_MSB] = (uchar)0;
      wd[cs][WDF_HALFTONE_PATTERN_LSB] = (uchar)BUILTIN_HALFTONE_PATTERN;

      wd[cs][WD_RIF_AND_PADDING]  = (uchar)0;

      wd[cs][WD_BIT_ORDERING_MSB] = (uchar)0;
      wd[cs][WD_BIT_ORDERING_LSB] = (uchar)0;

      break;

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
      wd[cs][WDR_HALFTONE_CODE] = (uchar)2; /* dithering */
      wd[cs][WDR_HALFTONE_ID]   = (uchar)0x0A; /* 8x8 Bayer pattern */

      wd[cs][WD_RIF_AND_PADDING] = (uchar)0x03;

      wd[cs][WD_BIT_ORDERING_MSB] = (uchar)0;
      wd[cs][WD_BIT_ORDERING_LSB] = (uchar)0x07;
      break;

    case RICOH_FS1:
      wd[cs][WDR_HALFTONE_CODE] = (uchar)0;	/* reserved */
      wd[cs][WDR_HALFTONE_ID]   = (uchar)0;	/* reserved */

      wd[cs][WD_RIF_AND_PADDING] = (uchar)0x03;

      wd[cs][WD_BIT_ORDERING_MSB] = (uchar)0;
      wd[cs][WD_BIT_ORDERING_LSB] = (uchar)0;
      break;

    case UMAX_UC630:
    case UMAX_UG630:
      wd[cs][WDR_HALFTONE_CODE] = (uchar)0x00;
      wd[cs][WDR_HALFTONE_ID]   = (uchar)0x0c; /* 4x4 pattern */

      wd[cs][WD_RIF_AND_PADDING] = (uchar)0x03;

      wd[cs][WD_BIT_ORDERING_MSB] = (uchar)0;
      wd[cs][WD_BIT_ORDERING_LSB] = (uchar)0;
      break;
    }

    wd[cs][WD_COMPRESSION_TYPE] = (uchar)0;
    wd[cs][WD_COMPRESSION_ARG]  = (uchar)0;
    wd[cs][WD_RESERVED01]	= (uchar)0;
    wd[cs][WD_RESERVED02]	= (uchar)0;
    wd[cs][WD_RESERVED03]	= (uchar)0;
    wd[cs][WD_RESERVED04]	= (uchar)0;
    wd[cs][WD_RESERVED05]	= (uchar)0;
    wd[cs][WD_RESERVED06]	= (uchar)0;

    switch  (scan_ptr->s_dds.scanner_type) {

    case FUJITSU:
      wd[cs][WDF_IDENTIFICATION_CODE]	= (uchar)0;
      wd[cs][WDF_RESERVED07]		= (uchar)0;
      wd[cs][WDF_OUTLINE]		= (uchar)0;
      wd[cs][WDF_EMPHASIS]		= (uchar)0;
      wd[cs][WDF_MIXED]			= (uchar)0;
      wd[cs][WDF_MIRRORING]		= (uchar)0;
      wd[cs][WDF_RESERVED08]		= (uchar)0;
      wd[cs][WDF_RESERVED09]		= (uchar)0;
      wd[cs][WDF_RESERVED10]		= (uchar)0;
      wd[cs][WDF_RESERVED11]		= (uchar)0;
      wd[cs][WDF_RESERVED12]		= (uchar)0;
      wd[cs][WDF_SUBWINDOW_LIST_MSB]	= (uchar)0;
      wd[cs][WDF_SUBWINDOW_LIST_LSB]	= (uchar)0;
      wd[cs][WDF_PAPER_SIZE]		= (uchar)0;

      wd[cs][WDF_X_MSB]		  = (uchar)0;
      wd[cs][WDF_X_MHB]		  = (uchar)0;
      wd[cs][WDF_X_MLB]		  = (uchar)0;
      wd[cs][WDF_X_LSB]		  = (uchar)0;

      wd[cs][WDF_Y_MSB]		  = (uchar)0;
      wd[cs][WDF_Y_MHB]		  = (uchar)0;
      wd[cs][WDF_Y_MLB]		  = (uchar)0;
      wd[cs][WDF_Y_LSB]		  = (uchar)0;

      wd[cs][WDF_RESERVED13]	  = (uchar)0;
      wd[cs][WDF_RESERVED14]	  = (uchar)0;
      break;

    case RICOH_IS410:
    case IBM_2456:
      wd[cs][WDR_RESERVED01]	    = (uchar)0;
      wd[cs][WDR_RESERVED02]	    = (uchar)0;
      wd[cs][WDR_MRIF_FILTERING_GAMMA_ID] = (uchar)0x80;
      wd[cs][WDR_RESERVED03]	    = (uchar)0;
      wd[cs][WDR_RESERVED04]	    = (uchar)0;
      wd[cs][WDR_BINARY_FILTER]	    = (uchar)0;
      wd[cs][WDR_RESERVED05]	    = (uchar)0;
      wd[cs][WDR_RESERVED06]	    = (uchar)0;
      wd[cs][WDR_RESERVED07]	    = (uchar)0;
      wd[cs][WDR_RESERVED08]	    = (uchar)0;
      wd[cs][WDR_RESERVED09]	    = (uchar)0;
      wd[cs][WDR_RESERVED10]	    = (uchar)0;
      wd[cs][WDR_RESERVED11]	    = (uchar)0;
      wd[cs][WDR_RESERVED12]	    = (uchar)0;
      wd[cs][WDR_RESERVED13]	    = (uchar)0;
      wd[cs][WDR_RESERVED14]	    = (uchar)0;
      wd[cs][WDR_RESERVED15]	    = (uchar)0;
      wd[cs][WDR_RESERVED16]	    = (uchar)0;
      wd[cs][WDR_RESERVED17]	    = (uchar)0;
      wd[cs][WDR_RESERVED18]	    = (uchar)0;
      wd[cs][WDR_RESERVED19]	    = (uchar)0;
      wd[cs][WDR_RESERVED20]	    = (uchar)0;
      wd[cs][WDR_RESERVED21]	    = (uchar)0;
      wd[cs][WDR_RESERVED22]	    = (uchar)0;
      break;

    case RICOH_IS50:
      bzero(&wd[cs][WDR_RESERVED01], 24);

      switch (scan_ptr->s_dds.image_mode) {
      case GRAYSCALE:
	wd[cs][WDR_MRIF_FILTERING_GAMMA_ID] = (uchar)0x90;
	break;
      case DITHERED_MONOCHROME:
	wd[cs][WDR_MRIF_FILTERING_GAMMA_ID] = (uchar)0x10;
	break;
      }

    case RICOH_FS1:
      bzero(&wd[cs][WDR_RESERVED01], 223);
      break;

    case UMAX_UC630:
    case UMAX_UG630:

      wd[cs][WDU_SPEED] = (uchar)0x01; /* fastest speed that doesn't smear */

      switch (scan_ptr->s_dds.image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	wd[cs][WDU_SELECT_COLOR] = (uchar)0;
	break;

      default:			/* color modes */
	switch (scan_ptr->s_dds.image_mode) {
	case RED:
	  wd[cs][WDU_SELECT_COLOR] = 0x80;
	  break;
	case GREEN:
	  wd[cs][WDU_SELECT_COLOR] = 0x40;
	  break;
	case BLUE:
	  wd[cs][WDU_SELECT_COLOR] = 0x20;
	  break;
#ifdef DEBUG_SCSI_CMD
	default:
	  printf("scan_scsi_cmd: problem setting colors\n");
#endif
	}
      }

      wd[cs][WDU_HIGHLIGHT] = (uchar)50;
      wd[cs][WDU_SHADOW] = (uchar)0;

      wd[cs][WDU_PAPER_LENGTH_MSB] = (uchar)0;
      wd[cs][WDU_PAPER_LENGTH_LSB] = (uchar)0;
      break;
    }

    /* customize the buf struct */
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_WRITE;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)wd[cs];
    switch (scan_ptr->s_dds.scanner_type) {

    case RICOH_FS1:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 0x0108; /* 256+8 */
    break;

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
    case FUJITSU:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 72;
      break;

    case UMAX_UC630:
    case UMAX_UG630:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 46 + 8;
      break;
    }
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 15;

    /* bytes 2-5 are reserved */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;

    /*
     * bytes 6-8 are for transfer length:
     *	  Ricoh IS410 wants 328 (0x148) bytes, but ignores extra data
     *	     and accepts less data as valid too
     *	  Fujitsu wants 72 (0x48) bytes
     *    Ricoh FS1 wants 264 (0x0108) bytes
     */

    /* shorthand #define for the sake of space (scb = scsi command byte) */
#define scb scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes

    switch (scan_ptr->s_dds.scanner_type) {

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
      scb[SCB6] = (uchar)0x00; /* MSB */
      scb[SCB7] = (uchar)0x00;
      scb[SCB8] = (uchar)0x48; /* LSB  (72 bytes) */
      break;

    case RICOH_FS1:
      scb[SCB6] = (uchar)0x00; /* MSB */
      scb[SCB7] = (uchar)0x01;
      scb[SCB8] = (uchar)0x08; /* LSB */
      break;

    case FUJITSU:
      scb[SCB6] = (uchar)0x00; /* MSB */
      scb[SCB7] = (uchar)0x00;
      scb[SCB8] = (uchar)0x48; /* LSB  (72 bytes) */
      break;

    case UMAX_UC630:
    case UMAX_UG630:
      scb[SCB6] = (uchar)0x00; /* MSB */
      scb[SCB7] = (uchar)0x00;
      scb[SCB8] = (uchar)0x36; /* LSB  (54 bytes) */
      break;
    }

    /* byte number 10 is called "vendor unique" or "control byte" */
    scb[SCB9] = (uchar)0;

    break;

  case SCSI_TRIGGER_SCAN:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_WRITE;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 90;

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
    case RICOH_FS1:
    case UMAX_UC630:
    case UMAX_UG630:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)window_zero;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)sizeof(window_zero);
      break;

    default:
      break;			/* no data out during start/stop */
    }

    /* bytes 2-3 & 5 are reserved */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
    case RICOH_FS1:
    case UMAX_UC630:
    case UMAX_UG630:
      /* transfer length */
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x01;
      break;
    }

    break;

  case SCSI_MODE_SELECT:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_WRITE;

    /* lun is really SCB1 (sigh) */
    if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
	scan_ptr->s_dds.scanner_type == IBM_2456    ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50  ||
	scan_ptr->s_dds.scanner_type == RICOH_FS1)
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.lun |= (uchar)PF_BIT;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_FS1:
      /* parameter list length */
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 12;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0;

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)&rmp;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)(12);

      /* mode select parameters... */
      /* header: */
      rmp.mode_data_length = 0;
      rmp.medium_type = 0;
      rmp.device_specific_parameter = 0;
      rmp.block_descriptor_length = 0;

      /* page descriptor: */
      rmp.page_code = SCANNING_MEASUREMENT_PARAMETERS;
      rmp.paramter_length = (uchar)6;
      rmp.base_measurement_unit = INCHES;
      rmp.reserved1 = 0;
      rmp.measurement_unit_devisor_msb = (1200 >> 8) & 0xff;
      rmp.measurement_unit_devisor_lsb = (1200 & 0xff);
      rmp.reserved2 = 0;
      rmp.reserved3 = 0;
      break;

    case RICOH_IS410:
    case IBM_2456:
    case RICOH_IS50:
      /* parameter list length */
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 20;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)&rmp;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)(sizeof(rmp));

      /* mode select parameters... */
      /* header: */
      rmp.mode_data_length = 0;
      rmp.medium_type = 0;
      rmp.device_specific_parameter = 0;
      rmp.block_descriptor_length = 0;

      /* page descriptor: */
      rmp.page_code = SCANNING_MEASUREMENT_PARAMETERS;
      rmp.paramter_length = (uchar)6;
      rmp.base_measurement_unit = INCHES;
      rmp.reserved1 = 0;
      rmp.measurement_unit_devisor_msb = (1200 >> 8) & 0xff;
      rmp.measurement_unit_devisor_lsb = (1200 & 0xff);
      rmp.reserved2 = 0;
      rmp.reserved3 = 0;

      /* ADF support */
      rmp.adf_page_code = 0x26;
      rmp.adf_parameter_length = 6;
      if (scan_ptr->adf_state == ADF_ARMED) {
#ifdef DEBUG_SCSI_CMD
	printf("scan_scsi_cmd: ADF is on\n");
#endif
	rmp.adf_control = '\01';
      } else {
#ifdef DEBUG_SCSI_CMD
	printf("scan_scsi_cmd: ADF is off\n");
#endif
	rmp.adf_control = '\0';
      }
      rmp.reserved4 = 0;
      rmp.reserved5 = 0;
      rmp.reserved6 = 0;
      rmp.reserved7 = 0;
      rmp.reserved8 = 0;
      break;
    }
    break;			/* end of mode select */

  case SCSI_GET_BUFFER_STATUS:
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr =
      (caddr_t)scan_ptr->buf_status;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount =
      (uint)sizeof(scan_ptr->buf_status);
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 15;

    /* reserved goo */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0x00;

    /* allocation length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] =
      (uchar)((sizeof(scan_ptr->buf_status) >> 8) & 0xff); /* MSB */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] =
      (uchar)(sizeof(scan_ptr->buf_status) & 0xff);        /* LSB */

    /* vendor unique goo */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0x00;

    break;			/* esac GET_BUFFER_STATUS */

  case SCSI_OBJECT_POSITION:
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 30;

    if (scan_ptr->do_unload) {
      /* the load bit has been cleared by the common code */
      scan_ptr->do_unload = FALSE;
    } else {
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.lun |= LOAD_BIT;
      scan_ptr->do_unload = TRUE;
    }

    /* count of pages to move */

    switch  (scan_ptr->s_dds.scanner_type) {

    case FUJITSU:		/* unsupported: must be set to zero */
    case RICOH_IS50:
    case UMAX_UG630:
    case UMAX_UC630:
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
      break;

    case RICOH_IS410:			/* one page */
    case IBM_2456:
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x01;
      break;
    }

    /* reserved goo */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0x00;

    break;			/* esac SCSI_OBJECT_POSITION */

  case SCSI_GET_IMAGE_STATUS:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr =
      (caddr_t)scan_ptr->buf_status;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 6;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 15;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= SC_NODISC;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    /* SCB4 is the transfer length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = '\06';
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    break;

  case SCSI_WRITE_EXTENDED:	/* so far only Sharp JX-600S */
    escape_codes[0] = '\0';

    strcpy(escape_codes, "T2");

    switch (scan_ptr->s_dds.x_resolution) {
    case 300:
      strcat(escape_codes, "D1");
      break;
    case 200:
      strcat(escape_codes, "D2");
      break;
    case 150:
      strcat(escape_codes, "D3");
      break;
    case 100:
      strcat(escape_codes, "D4");
      break;
    case 75:
      strcat(escape_codes, "D5");
      break;
    case 400:
      strcat(escape_codes, "D6");
      break;
    case 600:
      strcat(escape_codes, "D7");
      break;
    }

    switch (scan_ptr->s_dds.image_mode) {
    case COLOR:
      strcat(escape_codes, "C1");
      break;
    case RED:
      strcat(escape_codes, "C3L2");
      break;
    case GREEN:
      strcat(escape_codes, "C3L1");
      break;
    case BLUE:
      strcat(escape_codes, "C3L3");
      break;
    case GRAYSCALE:
      strcat(escape_codes, "C3L1"); /* man. says deflt drop out color green*/
      break;
    case BINARY_MONOCHROME:
      strcat(escape_codes, "C4L1");
      break;
    }

#if 0				/* warning: this code panics AIX 3.1 */
    if (scan_ptr->s_dds.image_mode != BINARY_MONOCHROME) {
      int i, j;

      strcat(escape_codes, "G0;");

      for (j = 0; j < 4; j++) {
	sprintf(escape_codes, "%s%d",
		escape_codes, scan_ptr->s_dds.brightness);
	for (i = 0; i < 255; i++)
	  sprintf(escape_codes, "%s,%d",
		  escape_codes, scan_ptr->s_dds.brightness);
	strcat(escape_codes, "/");
      }

      strcat(escape_codes, ";");
    } else
      strcat(escape_codes, "B1"); /* MONO threshold: normal */
#else
    strcat(escape_codes, "G2");
#endif /* 0 */

    strcat(escape_codes, "F5"); /* image format will be like PPM(5) */

    /* size & placement of window */

    sprintf(escape_codes, "%sA0;%d,%d,%d,%d;",
	    escape_codes,
	    scan_ptr->s_dds.x_origin / 48 + 3,
	    scan_ptr->s_dds.window_width / 48,
	    scan_ptr->s_dds.y_origin / 48 + 10,
	    scan_ptr->s_dds.window_length / 48);

    /* overhead time... */

    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_WRITE;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 15;

    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)escape_codes;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)strlen(escape_codes);

#ifdef DEBUG_READ
    printf("scanread: escape_codes is %s\n", escape_codes);
#endif

    /* transfer data type 0x80 is for setting mode parameters */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x80;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0x00;

    /* transfer length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] =
      (uchar)(strlen(escape_codes)) >> 8;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] =
      (uchar)(strlen(escape_codes)) & 0xff;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0x00;
    break;

  case SCSI_WRITE:		/* so far only HP ScanJet */
    switch (scan_ptr->hp_command) {
    case HP_CONFIG:
      sprintf(escape_codes, "%c*f%dP", '\033',
	      scan_ptr->s_dds.window_width / 4);
      sprintf(escape_codes, "%s%c*f%dQ", escape_codes, '\033',
	      scan_ptr->s_dds.window_length / 4);
      sprintf(escape_codes, "%s%c*f%dX", escape_codes, '\033',
	      scan_ptr->s_dds.x_origin / 4);
      sprintf(escape_codes, "%s%c*f%dY", escape_codes, '\033',
	      scan_ptr->s_dds.y_origin / 4);

      sprintf(escape_codes, "%s%c*a%dR", escape_codes, '\033',
	      scan_ptr->s_dds.x_resolution);
      sprintf(escape_codes, "%s%c*a%dS", escape_codes, '\033',
	      scan_ptr->s_dds.y_resolution);

      switch (scan_ptr->s_dds.image_mode) {
      case BINARY_MONOCHROME:
	sprintf(escape_codes, "%s%c*a%dT", escape_codes, '\033', 0);
	sprintf(escape_codes, "%s%c*a%dG", escape_codes, '\033', 1);
	/* make image data be "min-is-white ala PBM */
	sprintf(escape_codes, "%s%c*a%dI", escape_codes, '\033', 0);
	break;
      case DITHERED_MONOCHROME:
	sprintf(escape_codes, "%s%c*a%dT", escape_codes, '\033', 3);
	sprintf(escape_codes, "%s%c*a%dG", escape_codes, '\033', 1);
	/* make image data be "min-is-white ala PBM */
	sprintf(escape_codes, "%s%c*a%dI", escape_codes, '\033', 0);
	break;
      case GRAYSCALE:
	sprintf(escape_codes, "%s%c*a%dT", escape_codes, '\033', 4);
	sprintf(escape_codes, "%s%c*a%dG", escape_codes, '\033', 8);
	/* make image data be "min-is-black ala PGM */
	sprintf(escape_codes, "%s%c*a%dI", escape_codes, '\033', 1);
	break;
      case COLOR:
	sprintf(escape_codes, "%s%c*a%dT", escape_codes, '\033', 5);
	/* use 24-bit data */
	sprintf(escape_codes, "%s%c*a%dG", escape_codes, '\033', 24);
	/* make image data be "min-is-black ala PPM */
	sprintf(escape_codes, "%s%c*a%dI", escape_codes, '\033', 1);
	/* use pass-through matrix (disable NTSC) */
	sprintf(escape_codes, "%s%c*u%dT", escape_codes, '\033', 2);
      }

      sprintf(escape_codes, "%s%c*a%dL", escape_codes, '\033',
	      (int)(scan_ptr->s_dds.brightness) - 128);
      
      sprintf(escape_codes, "%s%c*a%dK", escape_codes, '\033',
	      (int)(scan_ptr->s_dds.contrast) - 128);

      break;

    case HP_TRIGGER:
      sprintf(escape_codes, "%c*f0S", '\033');
      break;

    case HP_PAPER_SENSE:
      sprintf(escape_codes, "%c*s1027E", '\033');
      break;

    case HP_UNLOAD_ADF:
      sprintf(escape_codes, "%c*u0U", '\033');
      break;
    }
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_flags = B_WRITE;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 15;

    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)escape_codes;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = (uint)strlen(escape_codes);

#ifdef DEBUG_READ
    printf("scanread: escape_codes are:\n");
    {
      int i = 0;
      while (escape_codes[i]) {
	if (escape_codes[i] == 033)
	  printf("[esc]");
	else
	  printf("%c", escape_codes[i]);
	++i;
      }
      printf("\n");
    }
#endif

    /* transfer length */
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x0;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] =
      (uchar)(strlen(escape_codes)) >> 8;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] =
      (uchar)(strlen(escape_codes)) & 0xff;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;

    break;

  case SCSI_INQUIRY:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)&scan_ptr->inq;

    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount =
      (uint)sizeof(struct scan_inq_data);

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x02;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] =
      (uchar)sizeof(struct scan_inq_data);
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    break;

  case SCSI_SHARP_READ_ABORT:
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x8F;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0x00;
    break;

  case SCSI_READ:
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = scan_ptr->hp_response;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 16;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x10;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;
    break;

#ifdef DEBUG_SET_WINDOW
  case SCSI_GET_WINDOW:
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.lun |= (uchar)0x01;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_un.b_addr = (caddr_t)wd;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_bcount = 256+8;
    scan_ptr->cmd_buf.scsi_buf.timeout_value = 5;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0x00;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0x00;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0x00;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] = 0x01;
    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] = 0x08;

    scan_ptr->cmd_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0x00;

    break;
#endif /* DEBUG_SET_WINDOW */
  }

  scan_ptr->cmd_buf.scan_ptr = scan_ptr;

  scan_ptr->cmd_state = cmd;

  /* more reserved crap courtesy of IBM */
  scan_ptr->cmd_buf.scsi_buf.resvd1 = (uint)0;
  scan_ptr->cmd_buf.scsi_buf.resvd2 = (uint)0;
  scan_ptr->cmd_buf.scsi_buf.resvd3 = (uint)0;
  scan_ptr->cmd_buf.scsi_buf.resvd4 = (uint)0;

  scan_ptr->cmd_buf.scsi_buf.resvd5 = (uchar)0;
  scan_ptr->cmd_buf.scsi_buf.resvd6 = (uchar)0;
  scan_ptr->cmd_buf.scsi_buf.resvd7 = (uchar)0;
  scan_ptr->cmd_buf.scsi_buf.resvd8 = (uchar)0;

  /****	 call the adapter driver's strategy routine via devstrat()  ***/

  (void) devstrat(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
  return;
}

/*
 *  scaniodone: the routine with the error recovery, restart, & cleanup code
 *
 * sc_strategy -> iodone -> scaniodone
 */

void
scaniodone(struct buf *bp)
{
  struct scan_buf *buf_ptr;
  struct buf	  *cur_bp, *next_bp;
  struct sc_buf	  *scsi_buf_ptr;
  struct scan_df  *scan_ptr;

#ifdef DEBUG_IODONE
  printf("scaniodone called\n");
#endif

  /*
   * The B_DONE flag must be cleared to force iowait to wait
   * on this buf struct.  The flag will be set when processing
   * is complete. (other invocations of scaniodone() may happen
   * concurrently)
   */
  bp->b_flags &= ~B_DONE;

  scsi_buf_ptr = (struct sc_buf *)bp; /* these 2 typecasts are made possible */
  buf_ptr = (struct scan_buf *)bp;    /* by the fact that the three structs */
  scan_ptr = buf_ptr->scan_ptr;	      /* are overlapping supersets */

#ifdef DEBUG_IODONE
  printf("scaniodone: bp->b_error/bp->b_resid is 0x%x/0x%x\n",
	 bp->b_error, bp->b_resid);
  printf("scaniodone: scan_ptr->cmd_state is 0x%x\n", scan_ptr->cmd_state);
#endif

  if (scan_ptr->cmd_state == SCAN_PASS_THRU) {
    /*
     * In the case of a pass-through command, the waiting process
     * is simply woken up.  No error handling is done.
     */
    bp->b_flags |= B_DONE;
    e_wakeup(&bp->b_event);

    goto iodone_exit;
  }


  if (!(bp->b_flags & B_ERROR)) {		/* if no error flag... */
    switch (scan_ptr->cmd_state) {

      /* handle trivial commands: */

    case SCSI_SET_WINDOW:
    case SCSI_GET_WINDOW:
    case SCSI_RESERVE_UNIT:
    case SCSI_RELEASE_UNIT:
    case SCSI_REQUEST_SENSE:
    case SCSI_TRIGGER_SCAN:
    case SCSI_MODE_SELECT:
    case SCSI_GET_BUFFER_STATUS:
    case SCSI_OBJECT_POSITION:
    case SCSI_WRITE_EXTENDED:
    case SCSI_WRITE:
    case SCSI_INQUIRY:
#ifdef DEBUG_IODONE
      printf("scaniodone: cmd 0x%x finished ok\n", scan_ptr->cmd_state);
#endif
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    case SCSI_GET_IMAGE_STATUS:
#ifdef DEBUG_IODONE
      printf("scaniodone: SCSI_GET_IMAGE_STATUS finished\n");
#endif
      scsi_buf_ptr->flags &= ~SC_NODISC;
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    case SCSI_READ_EXTENDED:
#ifdef DEBUG_IODONE
      printf("scaniodone: SCSI_READ_EXTENDED complete\n");
#endif
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    case SCSI_READ:
#ifdef DEBUG_IODONE
      printf("scaniodone: SCSI_READ complete\n");
#endif
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

#ifdef DEBUG_IODONE
    default:
      printf("%cscaniodone: unmapped command state: %d\n",
	     '\07', scan_ptr->cmd_state);
#endif
    }
  } else {   /* b_flags has the error bit set */

    bp->b_flags &= ~B_ERROR;	/* clear the error bit */

    if (scsi_buf_ptr->status_validity & SC_ADAPTER_ERROR) {
      scan_ptr->adapter_error = adapter_error(bp);
#if defined(DEBUG_IODONE) || defined(DEBUG_ADAPTER_ERROR)
      printf("scaniodone: adapter error; code = %d\n", scan_ptr->adapter_error);
#endif

      /* clear the adapter error bit */
      scsi_buf_ptr->status_validity &= ~SC_ADAPTER_ERROR;

      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
    }

    switch (scan_ptr->cmd_state) {

    case SCSI_REQUEST_SENSE:
#ifdef DEBUG_IODONE
      printf("scaniodone: REQUEST_SENSE failed!\n");
#endif
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    case SCSI_SET_WINDOW:
    case SCSI_GET_WINDOW:
    case SCSI_RESERVE_UNIT:
    case SCSI_RELEASE_UNIT:
    case SCSI_TRIGGER_SCAN:
    case SCSI_MODE_SELECT:
    case SCSI_GET_BUFFER_STATUS:
    case SCSI_OBJECT_POSITION:
    case SCSI_GET_IMAGE_STATUS:
    case SCSI_WRITE_EXTENDED:
    case SCSI_WRITE:
    case SCSI_INQUIRY:
      if (scsi_buf_ptr->scsi_status == SC_CHECK_CONDITION) {

	/* tell the top half to to a request sense */
	scan_ptr->check_condition = TRUE;
      }
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    case SCSI_READ_EXTENDED:
    case SCSI_READ:
      if (scsi_buf_ptr->scsi_status == SC_CHECK_CONDITION) {

#ifdef DEBUG_IODONE
	printf("scaniodone: SC_CHECK_CONDITION detected\n");
#endif
	/* tell the top half to to a request sense */
	scan_ptr->check_condition = TRUE;
      }
      bp->b_flags |= B_DONE;
      e_wakeup(&bp->b_event);
      break;

    default:
#ifdef DEBUG_IODONE
      printf("scaniodone: unrecognized command state: 0x%x\n",
	     scan_ptr->cmd_state);
#endif
      break;
    }
  }


 iodone_exit:

#ifdef IOSTAT
  scan_ptr->scan_dkstat.dk_status &= ~IOST_DK_BUSY;
#endif

#ifdef DEBUG_IODONE
  printf("scaniodone exiting\n");
#endif

#ifndef IOSTAT			/* xlc can't handle a label at the end of */
  return;			/* a function */
#endif
}

/*
 *   adapter_error: process errors sent by the host adapter
 *
 *   return ADAPTER_FATAL if the card is totally hosed
 *   or	    ADAPTER_RETRY if device might come around again
 */
uchar adapter_error(struct scan_buf *sbp)
{
#ifdef DEBUG_ADAPTER_ERROR
    printf("adapter_error:  general_card_status: %d\n",
	   sbp->scsi_buf.general_card_status);
#endif
  switch (sbp->scsi_buf.general_card_status) {
  case SC_HOST_IO_BUS_ERR:	/* indicates a Host I/O Bus error condition */
  case SC_SCSI_BUS_FAULT:	/* indicates a failure of the SCSI Bus */
  case SC_ADAPTER_HDW_FAILURE:	/* hardware failure */
  case SC_ADAPTER_SFW_FAILURE:	/* microcode failure */
  case SC_FUSE_OR_TERMINAL_PWR: /* blown fuse or bad termination */
    return (ADAPTER_FATAL);

  case SC_SCSI_BUS_RESET:	/* external SCSI bus reset */
  case SC_CMD_TIMEOUT:		/* timeout_value units expired */
  case SC_NO_DEVICE_RESPONSE:	/* target device did not respond to select */
    return (ADAPTER_RETRY);

#ifdef DEBUG_ADAPTER_ERROR
  default:
    printf("adapter_error: unmapped general_card_status: %d\n",
	   sbp->scsi_buf.general_card_status);
#endif
  }

  return (0);
}

/*
 * scanwait:  sleep on the b_event flag of a buf struct until the
 *	      B_DONE bit in the b_flags gets set.
 *
 * precautions:
 * 1. code executed during the disable operation must be pinned
 * 2. the event_word parameter must be initialized to EVENT_NULL
 *    before its first use
 */

void
scanwait(struct buf *bp)
{
  int old_ilevel;

  old_ilevel = i_disable(INTIODONE);

  while ((bp->b_flags & B_DONE) == 0) {
    e_sleep(&bp->b_event, EVENT_SHORT); /* sleep with signals disabled */
  }

  i_enable(old_ilevel);
}

#ifdef DEBUG_UIO
print_buf_struct(struct buf *bp)
{
  printf("bp->b_flags	  0x%x\t", bp->b_flags);
  printf("bp->b_forw	  0x%x\n", bp->b_forw);
  printf("bp->b_back	  0x%x\t", bp->b_back);
  printf("bp->av_forw	  0x%x\n", bp->av_forw);
  printf("bp->av_back	  0x%x\t", bp->av_back);
  printf("bp->b_iodone	  0x%x\n", bp->b_iodone);
  printf("bp->b_vp	  0x%x\t", bp->b_vp);
  printf("bp->b_dev	  0x%x\n", bp->b_dev);
  printf("bp->b_blkno	  0x%x\t", bp->b_blkno);
  printf("bp->b_baddr	  0x%x\n", bp->b_baddr);
  printf("bp->b_bcount	  0x%x\t", bp->b_bcount);
  printf("bp->b_error	  0x%x\n", bp->b_error);
  printf("bp->b_resid	  0x%x\t", bp->b_resid);
  printf("bp->b_work	  0x%x\n", bp->b_work);
  printf("bp->b_options	  0x%x\t", bp->b_options);
  printf("bp->b_event	  0x%x\n", bp->b_event);
}
#endif /* DEBUG */
