/*
 *  PINT Pint Is Not TWAIN - common scanner driver interface for UN*X
 *  Copyright (C) 1994 Kenneth Stailey ken@spacenet.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * scandd.t.c: top half of PINT
 *
 * This is the top half of the device driver.  Routines in it can be called
 * only by a process and can page fault (and get paged out).
 *
 * Entry point routines:
 *  scanconfig
 *  scanopen
 *  scanclose
 *  scanread
 *  scanioctl
 *
 */

#define _ALL_SOURCE	  /* implicit <fcntl.h> include needs this pair */
#define _KERNEL		  /* of #defines */

#include <sys/types.h>
#include <sys/device.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/scsi.h>
#include <sys/fp_io.h>
#include <sys/pin.h>
#ifdef IOSTAT
# include <sys/iostat.h>
#endif
#include <sys/ioctl.h>
#include <sys/devinfo.h>
#include <sys/priv.h>
#include <sys/watchdog.h>
#include <sys/sleep.h>
#include "scandd.h"
#include "scanio.h"

/*
 *	global data definitions
 */

extern struct scan_hash_table_df scanner_list[SCANNER_HASHSIZE];
extern int scanner_chain_lock;	 /* for exclusive access to scanner_list */
extern int scanner_config_count; /* # of devices configured	     */

/*
 * scanconfig:
 * This is the main entry point to the module. It is invoked via the
 * sysconfig system call from user mode.
 */

int
scanconfig(
 dev_t	 devno,		  /* major-minor device number	  */
 int	 cmd,		  /* command from sysconfig()	  */
 struct uio *uiop)	  /* uio structure containing device dependent data */
{
  struct scan_df  *scan_ptr, *back_ptr;
  struct devsw ds;
  char found;		  /* used for search through scanner_list */
  char scanner_offset;
  int undo = 0;		  /* used to keep track of what steps to reverse */
  int rc = 0;		  /* return code from this function */

#ifdef DEBUG_CONFIG
  printf("entering scanconfig ");
#endif

  /*
   * Check the list of defined scanner structures to see if this device
   * has already been defined.
   */
  scanner_offset = minor(devno) & 0x0f;

#ifdef DEBUG_CONFIG
  printf("(lock");
#endif
  (void) lockl(&scanner_chain_lock, LOCK_SHORT);
#ifdef DEBUG_CONFIG
  printf("ing) \n");
#endif

  found = FALSE;
  scan_ptr = scanner_list[scanner_offset].defined_head;
  back_ptr = scan_ptr;

#ifdef DEBUG_CONFIG
    printf("scanconfig: devno is 0x%x	 major/minor is %d/%d\n",
	   devno, major(devno), minor(devno));
    printf("scanconfig: scan_ptr is 0x%x before searching\n", scan_ptr);
#endif 

  while (!found && (scan_ptr != NULL)) {
    if (scan_ptr->devno == devno) {
      found = TRUE;
    } else {
      back_ptr = scan_ptr;
      scan_ptr = scan_ptr->next_defined;
    }
  }

#ifdef DEBUG_CONFIG
  printf("scanconfig: scan_ptr is 0x%x after searching\n", scan_ptr);
#endif

  switch (cmd) {

  case CFG_INIT:

    /**** make sure the device has not already been defined  ***/

    if (!found)	{
      /*
       * To define this device:
       * 1) allocate pinned storage for the disk structure
       * 2) fill in the disk structure (DDS, misc ptrs)
       *
       * if this is the first scanner to be defined
       * 3) add the device switch table entries for this device.
       *
       * 4) add this structure to the linked list of defined devices
       * 5) register the dkstat struct
       *
       */

      /****  allocate pinned memory for the scanner info structure  ***/

      if ((scan_ptr = (struct scan_df *)
	   xmalloc((int)sizeof(struct scan_df),
		   2, (caddr_t)pinned_heap)) == NULL) {

#ifdef DEBUG_CONFIG
	printf("scanconfig: xmalloc of scan_df failed\n");
#endif
	rc = ENOMEM;
	goto scanconf_exit;
      }
      ++undo;		  /* xmfree() is now necessary */
#ifdef DEBUG_CONFIG
	printf("scanconfig: xmalloc()ed scan_ptr is 0x%x\n", scan_ptr);
#endif

      /****  initialize the scan_df struct  ***/

      scan_ptr->next_defined       = NULL;
      scan_ptr->next_opened        = NULL;
      scan_ptr->devno	           = devno;
      scan_ptr->opened	           = (uchar)FALSE;
      scan_ptr->do_unload          = (uchar)TRUE;
      scan_ptr->diag_mode          = (uchar)FALSE;
      scan_ptr->retain_reservation = (uchar)FALSE;
      scan_ptr->adapter_error      = (uchar)FALSE;
      scan_ptr->check_condition    = (uchar)FALSE;
      scan_ptr->using_adf          = (uchar)FALSE;
      scan_ptr->adf_state          = ADF_UNUSED;

      /****  fetch the DDS from the config method	 ***/

      if (rc = uiomove((caddr_t)&scan_ptr->s_dds,
			sizeof(struct scan_dds), UIO_WRITE, uiop)) {
#ifdef DEBUG_CONFIG
	printf("scanconfig: uiomove of the DDS failed\n");
#endif
	goto scanconf_exit;
#ifdef DEBUG_CONFIG
      } else {
	printf("scanconfig: SCSI id/LUN is %d/%d\n",
	       scan_ptr->s_dds.scsi_id, scan_ptr->s_dds.lun);
	printf("scanconfig: parent major/minor is %d/%d\n",
	       major(scan_ptr->s_dds.parent_devno),
	       minor(scan_ptr->s_dds.parent_devno));
	printf("scanconfig: scanner type is %d	  logical name is %s\n",
	       scan_ptr->s_dds.scanner_type, scan_ptr->s_dds.logical_name);
	printf("scanconfig: window_width/window_length is %d/%d\n",
	       scan_ptr->s_dds.window_width, scan_ptr->s_dds.window_length);
	printf("scanconfig: x_resolution/y_resolution is %d/%d\n",
	       scan_ptr->s_dds.x_resolution, scan_ptr->s_dds.y_resolution);
#endif
      }

      if (++scanner_config_count == 1) {

	/**** add the device switch table entries for this device ***/

	ds.d_open     = (int(*)())scanopen;
	ds.d_close    = (int(*)())scanclose;
	ds.d_read     = (int(*)())scanread;
	ds.d_write    = nodev;
	ds.d_ioctl    = (int(*)())scanioctl;
	ds.d_strategy = nodev;
	ds.d_ttys     = NULL;
	ds.d_select   = nodev;
	ds.d_config   = (int(*)())scanconfig;
	ds.d_print    = nodev;
	ds.d_dump     = nodev;
	ds.d_mpx      = nodev;
	ds.d_revoke   = nodev;
	ds.d_dsdptr   = NULL;
	ds.d_selptr   = NULL;
	ds.d_opts     = 0L;

#ifdef DEBUG_CONFIG
	printf("scanconfig: calling devswadd\n");
#endif

	if (rc = devswadd(devno, &ds)) {
#ifdef DEBUG_CONFIG
	  printf("scanconfig: devswadd failed\n");
#endif
	  goto scanconf_exit;
	} else {
	  ++undo;	  /* devswdel is now necessary */
	}
      }

      /****  add this structure to the linked list of defined devices  ***/

#ifdef DEBUG_CONFIG
      printf("scanconfig: back_ptr is 0x%x\n", back_ptr);
#endif

      if (back_ptr == NULL) {
	scanner_list[scanner_offset].defined_head = scan_ptr;
#ifdef DEBUG_CONFIG
      printf("scanconfig: *scanner_list[scanner_offset].defined_head is 0x%x\n",
	     *scanner_list[scanner_offset].defined_head);
#endif
      } else {
	back_ptr->next_defined = scan_ptr;
      }

      /****  register the dkstat struct	 ***/
#ifdef IOSTAT
      strcpy(scan_ptr->scan_dkstat.diskname, scan_ptr->s_dds.logical_name);
      scan_ptr->scan_dkstat.dknextp   = NULL;
      scan_ptr->scan_dkstat.dk_status = (ushort)0;
      scan_ptr->scan_dkstat.dk_time   = (ulong)0;
      scan_ptr->scan_dkstat.dk_xrate  = (ulong)0;
      scan_ptr->scan_dkstat.dk_bsize  = (ulong)1; /* minimum you can ask for */
      scan_ptr->scan_dkstat.dk_xfers  = (ulong)0;
      scan_ptr->scan_dkstat.dk_rblks  = (ulong)0;
      scan_ptr->scan_dkstat.dk_wblks  = (ulong)0;
      scan_ptr->scan_dkstat.dk_seek   = (ulong)0;
      if (rc = iostadd(DD_DISK, &scan_ptr->scan_dkstat)) {
#ifdef DEBUG_CONFIG
	printf("scanconfig: iostadd failed rc = %d\n", rc);
#endif /* DEBUG */
      } else {
	++undo;		  /* iostdel is now necessary */
      }
#endif /* IOSTAT */

#ifdef DEBUG_CONFIG
      printf("scanconfig: scanner_list 0x%x, 0x%x, 0x%x\n",
	     scanner_list[0].defined_head,
	     scanner_list[1].defined_head,
	     scanner_list[2].defined_head);
#endif
    } else { /* this device has already been defined, so return an error */
      rc = EINVAL;
    }
    break;


  case CFG_TERM:
    if (found) {
      /*
       * If the scanner has been defined and is not currently opened, then
       * delete the device by freeing the disk structure and unregistering
       * the scan_dkstat structure.
       * If this is the last cdrom delete the switch table entry.
       */

#ifdef DEBUG_CONFIG
	    printf("scanconfig: removing scanner\n");
#endif

      if (!scan_ptr->opened) {
	if (--scanner_config_count == 0) {
#ifdef DEBUG_CONFIG
	    printf("scanconfig: performing devswdel\n");
#endif
	  if (rc = devswdel(devno)) {
#ifdef DEBUG_CONFIG
	    printf("scanconfig: devswdel failed\n");
#endif
	  }
	}

	if (rc == 0) {
	  if (back_ptr == scan_ptr) {
	    scanner_list[scanner_offset].defined_head = NULL;
	  } else {
	    back_ptr->next_defined = scan_ptr->next_defined;
	  }

#ifdef IOSTAT
	  /* unregister scan_dkstat structure */
#ifdef DEBUG_CONFIG
	  printf("scanconfig: unregistering scan_dkstat structure\n");
#endif /* DEBUG */
	  iostdel(&(scan_ptr->scan_dkstat));
#endif

	  (void)xmfree((caddr_t)scan_ptr, (caddr_t)pinned_heap);
	}
      } else {
#ifdef DEBUG_CONFIG
	printf("scanconfig: in CFGTERM the scanner is still open\n");
#endif
	rc = EINVAL;	  /* this device is flagged as open */
      }
    } else {
#ifdef DEBUG_CONFIG
	printf("scanconfig: no scanners found\n");
#endif
    }
    break;

  default:
#ifdef DEBUG_CONFIG
    printf("scanconfig: invalid command\n");
#endif
    return (EINVAL);
    break;
  }

  /*
   *  Get out of scan config performing any cleanup needed by crapping
   *  out in the middle.
   *
   *  If "rc" has any value other than zero, then something went wrong.
   */

 scanconf_exit:

#ifdef DEBUG_CONFIG
  printf("scanconfig: exiting ");
#endif

  if (rc == 0) {
#ifdef DEBUG_CONFIG
    printf("(unlock");
#endif
    unlockl(&scanner_chain_lock);
#ifdef DEBUG_CONFIG
    printf("ing) sucessfully\n");
#endif
    return (rc);
  }

#ifdef DEBUG_CONFIG
    printf("with level %d undo processing\n", undo);
#endif

  switch (undo) {
#ifdef IOSTAT
  case 3:		  /* iostdel() is now necessary */

    iostdel(&(scan_ptr->scan_dkstat)); /* fall through to level 2 */
#endif /* IOSTAT */

  case 2:		  /* devswdel() is now necessary */

    (void)devswdel(scan_ptr->devno); /* fall through to level 1 */

  case 1:		  /* xmfree() is now necessary */

    (void)xmfree((caddr_t)scan_ptr, (caddr_t)pinned_heap);
    
  case 0:		  /* release the lock */
    unlockl(&scanner_chain_lock);
    break;

  default:
#ifdef DEBUG_CONFIG
    printf("scanconf: scanconf_exit was handed a bad undo value: %d\n", undo);
#endif
    unlockl(&scanner_chain_lock); /* don't make things worse! */
  }

  return (rc);
}	/* end scanconfig */

/*
 * scanopen: open routine for scan device driver
 *
 * a) validate parameters
 * b) find the scanner in the defined list
 * c) determine if this scanner is being opened for the first time
 *
 * If this is the first open for a scanner:
 * 1) open the parent adapter device
 * 2) issue a START ioctl to the adapter device driver
 * 3) if this is a FORCED open, send a bus device reset to the scanner
 * 4) issue a SCSI_TEST_UNIT_READY ioctl to the adapter device driver
 * 5) pin the bottom half routines
 * 6) issue a SCSI_RESERVE_UNIT command to the scanner
 */

int
scanopen(dev_t devno, ulong devflag, chan_t chan, int ext)
{
  struct scan_df *scan_ptr;
  char scanner_offset;
  struct sc_ready scan_ready;
  struct devinfo adapter_devinfo;

  int undo = 0;	     /* holds level of undo used by exit routine */
  int rc = 0;	     /* return code from this function */

#ifdef DEBUG_OPEN
  printf("scanopen called ");
#endif

  /* parameter validation  */

  if ((devflag != FREAD)	&&	/* the scanner is a read only device */
      (devflag != FKERNEL)	&&	/* and it may be opened in diag mode */
      (devflag != (FREAD | FKERNEL))) {
#ifdef DEBUG_OPEN
    printf("with invalid parameters\n");
#endif
    return (EINVAL);
  }

  /****	 locate scanner struct	***/

#ifdef DEBUG_OPEN
  printf("(lock");
#endif
  (void)lockl(&scanner_chain_lock, LOCK_SHORT); /* unlockl() is now necessary*/
#ifdef DEBUG_OPEN
  printf("ing) ");
#endif

  scanner_offset = minor(devno) & 0x0f;
  scan_ptr = scanner_list[scanner_offset].defined_head;
  while (scan_ptr != NULL) {
    if (scan_ptr->devno == devno)
      break;
    else
      scan_ptr = scan_ptr->next_defined;
  }
  if (scan_ptr == NULL) {
    rc = ENODEV;	  /* can't find the scanner in the defined list */
#ifdef DEBUG_OPEN
    printf("but the scanner couldn't be found\n");
#endif
    goto scanopen_exit;
  }

  /****	 determine if the scanner is being opened for the first time  ***/

  if (scan_ptr->opened) {
#ifdef DEBUG_OPEN
    printf("again\n");
#endif

    rc = EBUSY;	  /* XXX ICK! En_US message catalog will say: */
		  /* "A filesystem cannot be unmounted" */
		  /* use C message catalog for "device busy" */
    goto scanopen_exit;

    /*
     * This is not the first open.  Check to make sure it was not already
     * in diag mode, or if it is being opened in diag mode (both conditions
     * are illegal).  Otherwise, nothing needs to be done for multiple opens.
     */
    if ((ext & SC_DIAGNOSTIC) || (scan_ptr->diag_mode)) {
      rc = EBUSY;
      goto scanopen_exit;
    }

    if (ext & SC_RETAIN_RESERVATION) {
      scan_ptr->retain_reservation = TRUE;
    }
    goto scanopen_exit;

  } else {
#ifdef DEBUG_OPEN
    printf("for the first time\n");
#endif

    /* openx stuff... */
#ifdef DEBUG_OPEN
    printf("scanopen: ext parameter is 0x%x\n", ext);
#endif

#if WhatIsPrivcheck
    /*
     * check that the user has CTL_DEV authority if ext paramaters are used.
     */
    if (ext && priv_chk(RAS_CONFIG)) {
      rc = EPERM;
      goto scanopen_exit;
    }
#endif

    if (ext & SC_DIAGNOSTIC) {
      scan_ptr->diag_mode = TRUE;
      /* XXX there's probably more to it that just that one line */
    }

    /*  open the host adapter device (/dev/scsi0) */

    if (rc = fp_opendev(scan_ptr->s_dds.parent_devno,
			FREAD, NULL, NULL, &scan_ptr->fp)) {
#ifdef DEBUG_OPEN
      printf("scanopen: fp_opendev of the adapter failed; rc = %d\n", rc);
#endif
      goto scanopen_exit;	/* open failed, get out of here */
    } else {
      ++undo;		/* closing the adapter will now be necessary */
    }

    /*  get the maximum transfer from the host adapter */

    if (rc = fp_ioctl(scan_ptr->fp, IOCINFO, &adapter_devinfo, NULL)) {
#ifdef DEBUG_OPEN
      printf("scanopen: IOCINFO ioctl failed\n");
#endif
    } else {
#ifdef DEBUG_OPEN
      printf("scanopen: max transfer is %d\n",
	     adapter_devinfo.un.scsi.max_transfer);
#endif
    }

    /*  use the host adapter device driver to send start unit command */

    if (rc = fp_ioctl(scan_ptr->fp, SCIOSTART,
		      IDLUN(scan_ptr->s_dds.scsi_id,
			    scan_ptr->s_dds.lun), NULL)) {
#ifdef DEBUG_OPEN
      printf("scanopen: fp_ioctl SCIOSTART failed; rc = %d\n", rc);
#endif
      goto scanopen_exit;	/* ioctl failed, get out of here */
    } else {
      ++undo;			/* ioctl SCIOSTOP is now necessary */
    }

    /* on a SC_FORCED_OPEN issue a bus device reset to the scanner */

    if (ext & SC_FORCED_OPEN) {
#ifdef DEBUG_OPEN
	printf("scanopen: SC_FORCED_OPEN detected--sending BDR\n");
#endif
      if (rc = fp_ioctl(scan_ptr->fp, SCIORESET,
			IDLUN(scan_ptr->s_dds.scsi_id,
			      scan_ptr->s_dds.lun), NULL)) {
#ifdef DEBUG_OPEN
	printf("scanopen: BDR failed\n");
#endif
	goto scanopen_exit;
      }
    }

    /*  pin the bottom half of the device driver  */

    if (rc = pincode(scaniodone)) {
      goto scanopen_exit;	/* pincode failed, get out of here */
    } else {
      ++undo;			/* unpincode is now necessary */
    }

    /*  use the host adapter device driver to send  Test Unit Ready command  */

    /* set the async flag if we know the scanner requires it */

#ifdef DEBUG_OPEN
	printf("scanopen: sending TEST UNIT READY\n", rc);
#endif

    if (scan_ptr->s_dds.scanner_type == RICOH_FS1  ||
	scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630 ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50 ||
	scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC)
      scan_ready.flags   = (uchar)SC_ASYNC;
    else
      scan_ready.flags   = (uchar)0;

    for (scan_ptr->tur_count = 0;
	 scan_ptr->tur_count < MAX_TUR_RETRIES; scan_ptr->tur_count++) {

      scan_ready.scsi_id = scan_ptr->s_dds.scsi_id;
      scan_ready.lun_id  = scan_ptr->s_dds.lun;
      scan_ready.status_validity = (uchar)0;
      scan_ready.scsi_status     = (uchar)0;
      scan_ready.rsv1    = (uchar)0;
      scan_ready.rsv2    = (uchar)0;
      scan_ready.rsv3    = (uchar)0;
      scan_ready.rsv4    = (uint)0;

      if (rc = fp_ioctl(scan_ptr->fp, SCIOTUR, &scan_ready, NULL)) {
	if (rc == ETIMEDOUT) {

	  /* retry with the async flag turned on */

#ifdef DEBUG_OPEN
	  printf("scanopen: setting SC_ASYNC flag\n", rc);
#endif
	  scan_ready.flags |= SC_ASYNC;
	  scan_ptr->retry = TRUE;
	  continue;
	}
#ifdef DEBUG_OPEN
	printf("scanopen: fp_ioctl SCIOTUR failed rc = %d\n", rc);
#endif

	if (scan_ready.scsi_status == SC_CHECK_CONDITION) {
#ifdef DEBUG_OPEN
	  printf("scanopen: check condition detected, sending request sense\n");
#endif
	  /* send request sense */
	  do_request_sense(scan_ptr);
	}
#ifdef DEBUG_OPEN
	printf("scanopen: retrying fp_ioctl SCIOTUR\n");
#endif
	delay(100);		/* wait for 1 second between tries */
	continue;
      } else {
	scan_ptr->retry = FALSE;
	break;
      }
    }
    if (scan_ptr->tur_count >= MAX_TUR_RETRIES) {
#ifdef DEBUG_OPEN
      printf("scanopen: Test Unit Ready retries exceeded\n");
#endif
      goto scanopen_exit;	/* ioctl failed, get out of here */
    }

    /****  send a Reserve Unit to the scanner  ***/

    if (scan_ptr->s_dds.scanner_type != HP_SCANJET_IIC) {
#ifdef DEBUG_OPEN
      printf("scanopen: sending SCSI_RESERVE_UNIT\n");
#endif

      scan_ptr->diag_mode = FALSE;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_RESERVE_UNIT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_OPEN
	printf("scanopen: SCSI_RESERVE_UNIT failed\n");
	printf("scanopen: sending SCSI_REQUEST_SENSE\n");
#endif
	/* send request sense */
	scan_ptr->check_condition = FALSE;
	scan_ptr->retry = TRUE;
	scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= SC_ASYNC;
	scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

	scan_scsi_cmd(SCSI_REQUEST_SENSE, scan_ptr);
	scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
#ifdef DEBUG_OPEN
	printf("scanopen: sense data is 0x%x/0x%x/0x%x\n",
	       scan_ptr->rs.sense_key,
	       scan_ptr->rs.additional_sense_code,
	       scan_ptr->rs.additional_sense_code_qualifier);
#endif
	scan_ptr->retry = FALSE;
	rc = EIO;
	goto scanopen_exit;
      }
#ifdef DEBUG_OPEN
      printf("scanopen: SCSI_RESERVE_UNIT ok\n");
#endif
    }
    ++undo;			/* release unit is now necessary */

#ifdef DEBUG_SET_WINDOW		/* results visable only on SCSI analyzer */
    printf("scanopen: sending SCSI_GET_WINDOW\n");

    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

    scan_scsi_cmd(SCSI_GET_WINDOW, scan_ptr);
    scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
#endif /* DEBUG_SET_WINDOW */

    scan_ptr->triggered = FALSE;
    scan_ptr->configured = FALSE;
  }

  /*   
   *  Get out of scanopen() performing any cleanup needed by crapping
   *  out in the middle.
   *
   *  If "rc" has any value other than zero, then something went wrong.
   */

scanopen_exit:

#ifdef DEBUG_OPEN
  printf("scanopen: exiting ");
#endif


  if (rc == 0) {
#ifdef DEBUG_OPEN
    printf("sucessfully\n");
#endif
    scan_ptr->opened = TRUE;
    scan_ptr->retain_reservation = (ext & SC_RETAIN_RESERVATION);
    scan_ptr->next_opened = scanner_list[scanner_offset].opened_head;
    scanner_list[scanner_offset].opened_head = scan_ptr;
    unlockl(&scanner_chain_lock);

    return (rc);
  }

#ifdef DEBUG_OPEN
  printf("with level %d undo processing\n", undo);
#endif

  switch (undo) {
  case 4:			/* release unit is now necessary */

    if (scan_ptr->s_dds.scanner_type != HP_SCANJET_IIC) {
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_RELEASE_UNIT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
    }
    
  case 3:			/* unpincode is now necessary */
    unpincode(scaniodone);

  case 2:			/* ioctl SCIOSTOP is now necessary */
    fp_ioctl(scan_ptr->fp, SCIOSTOP,
	     IDLUN(scan_ptr->s_dds.scsi_id, scan_ptr->s_dds.lun), NULL);

				/* fall through to 1 */

  case 1:			/* fp_close() on adapter device */
    (void) fp_close(scan_ptr->fp);
				/* fall through to 0 */

  case 0:			/* release the lock */
    unlockl(&scanner_chain_lock);
    break;

  default:			/* should not be here! */
#ifdef DEBUG_OPEN
    printf("scanopen: bad undo value: %d\n", undo);
#endif
    break;
  }

  return (rc);
}

/*
 * scanclose: close routine for scan device driver
 *
 * we get here on the last call to close() 
 * ) find scanner in the opened list
 * ) eject paper from ADF if appropriate
 * ) home the carriage on UMAX scanners
 * ) tell Sharp JX600 to stop if the window is not empty
 * ) send stop unit & release to the scanner
 * ) a STOP ioctl is issued to the adapter driver
 * ) the adapter device is closed
 * ) clear the "opened" flag
 * ) unpincode() the bottom half routines
 */

int
scanclose(dev_t devno, chan_t chan)
{
  struct scan_df *scan_ptr, *back_ptr;
  char scanner_offset;
  int undo = 0;
  int rc = 0;

#ifdef DEBUG_CLOSE
  printf("scanclose called ");
#endif

  scanner_offset = minor(devno) & 0x0f;

#ifdef DEBUG_CLOSE
  printf("(lock");
#endif
  (void)lockl(&scanner_chain_lock, LOCK_SHORT);
#ifdef DEBUG_CLOSE
  printf("ing) ");
#endif

  scan_ptr = scanner_list[scanner_offset].opened_head;
  back_ptr = scan_ptr;
  while (scan_ptr != NULL) {
    if (scan_ptr->devno == devno)
      break;
    else {
      back_ptr = scan_ptr;
      scan_ptr = scan_ptr->next_opened;
    }
  }
  if (scan_ptr == NULL) {	/* scanner not found */
#ifdef DEBUG_CLOSE
    printf("but the scanner wasn't on the opened list\n");
#endif
    unlockl(&scanner_chain_lock);
    return (ENXIO);
  } else {

#ifdef DEBUG_CLOSE
    printf("\n");
#endif

    /* remove the scanner from the open list
       (it will still be on the defined list) */

#ifdef DEBUG_CLOSE
    printf("scanclose: taking the scanner off of the open list\n");
#endif
    if (back_ptr == scan_ptr) {
      scanner_list[scanner_offset].opened_head = NULL;
    } else {
      back_ptr->next_opened = scan_ptr->next_opened;
    }

    if ((scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
	 scan_ptr->s_dds.scanner_type == IBM_2456    ||
	 scan_ptr->s_dds.scanner_type == RICOH_IS50)
	&& scan_ptr->adf_state == ADF_CLEANUP) {

      /* unload the paper */
      scan_ptr->do_unload = TRUE;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_OBJECT_POSITION, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_CLOSE
	printf("scanclose: SCSI_OBJECT_POSITION unload failed\n");
#endif
	do_request_sense(scan_ptr);
      }

      scan_ptr->adf_state = ADF_UNUSED;
    }

    /*
     * if there is no paper in the UMAX's ADF send an unload to home
     * the carriage.
     *
     * if there is paper in the UMAX's ADF send an unload to eject
     * the paper.
     */
    if (scan_ptr->triggered) {
      if (scan_ptr->s_dds.scanner_type == UMAX_UG630  ||
	  scan_ptr->s_dds.scanner_type == UMAX_UC630) {

#ifdef DEBUG_CLOSE
	printf("scanclose: sending SCSI_OBJECT_POSITION unload\n");
#endif
	scan_ptr->do_unload = TRUE;

	scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

	scan_scsi_cmd(SCSI_OBJECT_POSITION, scan_ptr);
	scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

	find_umax(scan_ptr);
      }

      /*
       * If there used to be paper in the ADF and now there's not then the
       * last sheet in the ADF just got ejected, so send an unload request
       * to home the carriage.
       */

      if (scan_ptr->using_adf && !umax_has_paper_in_adf(scan_ptr)) {

#ifdef DEBUG_CLOSE
	printf("scanclose: sending SCSI_OBJECT_POSITION unload\n");
#endif
	scan_ptr->do_unload = TRUE;

	scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

	scan_scsi_cmd(SCSI_OBJECT_POSITION, scan_ptr);
	scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

	find_umax(scan_ptr);
      }
    }

    /* if the Sharp has not read all of the data in the active area
     * issue a request to stop scanning
     */
    if (scan_ptr->s_dds.scanner_type == SHARP_JX600 && scan_ptr->window_size) {
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;
      scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= SC_ASYNC;
      scan_scsi_cmd(SCSI_SHARP_READ_ABORT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
    }

    /* if the HP ScanJet IIc has not read all of it's data and the ADF
     * was the paper source, send an unload request
     */
    if (scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC &&
	scan_ptr->triggered && scan_ptr->using_adf) {
      scan_ptr->hp_command = HP_UNLOAD_ADF;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_CLOSE
	printf("scanioctl: SCSI_WRITE of HP_UNLOAD_ADF failed\n");
#endif
	do_request_sense(scan_ptr);
      }
    }

    /****  do a release on the scanner	***/

    if (scan_ptr->s_dds.scanner_type != HP_SCANJET_IIC) {
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_RELEASE_UNIT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
    }
  }

  /****	 do a SCIOSTOP on the adapter device  ***/

  rc = fp_ioctl(scan_ptr->fp, SCIOSTOP,
		IDLUN(scan_ptr->s_dds.scsi_id,
		      scan_ptr->s_dds.lun), NULL);
#ifdef DEBUG_CLOSE
  if (rc)
    printf("scanclose: fp_ioctl SCIOSTOP failed; rc = %d\n", rc);
#endif

  /**** close the adapter device driver	 ***/

  (void)fp_close(scan_ptr->fp);

  /****	 clear the opened flag	***/

  scan_ptr->opened = FALSE;

  rc = unpincode(scaniodone);

  unlockl(&scanner_chain_lock);

  if (rc) {

#ifdef DEBUG_CLOSE
    printf("scanclose exiting with return value: %d\n", rc);
#endif

    return(rc);
  } else {

#ifdef DEBUG_CLOSE
    printf("scanclose exiting successfully\n");
#endif

    return (0);
  }
}

/*
 *   scanread: read routine for scan device driver
 *
 * (not really a raw interface to a block device at all, but that saves on
 *  pinned memory)
 *
 * basic workflow:
 *  1 time configure (ususally SET WINDOW operation)
 *  1 time calculation of how much data there is to move (window size)
 *  1 time trigger to start scanning (most scanners)
 *  on all calls convert the uio struct to a scsi buf struct and
 *     pass the result to the SCSI I/O controller device driver's strategy
 */

int
scanread(dev_t devno, struct uio *uiop, chan_t chan, int ext)
{
  struct iovec	  *iov;
  struct xmem	  *xmemp;
  struct scan_buf *sbp;
  struct scan_df  *scan_ptr;
  int old_ilevel;
  int tries;
  ushort lines;

  int rc = 0;
  int undo = 0;

#ifdef DEBUG_READ
  printf("scanread called\n");
#endif

  /* find the scanner in the opened list */

  old_ilevel = i_disable(INTIODONE);

  scan_ptr = scanner_list[minor(devno) & 0x0F].opened_head;
  while (scan_ptr != NULL) {
    if (scan_ptr->devno == devno)
      break;			/* found it! */
    else
      scan_ptr = scan_ptr->next_opened;
  }
  if (scan_ptr == NULL) {
#ifdef DEBUG_READ
    printf("scanread: can't find the scanner on the opened list\n");
#endif
    i_enable(old_ilevel);
    return(ENXIO);
  }

  i_enable(old_ilevel);

  /* clear the resid out of the uio pointer */
  uiop->uio_resid = 0;

  if (scan_ptr->configured == FALSE) {

    /* ************************************** */
    /* initialize the scanner's "active area" */
    /* ************************************** */

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_IS410:
    case IBM_2456:
    case RICOH_FS1:
    case RICOH_IS50:

      /****  send a SCSI_MODE_SELECT to scanners that need it ***/

#ifdef DEBUG_READ
      printf("scanread: sending SCSI_MODE_SELECT\n");
#endif
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_MODE_SELECT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_MODE_SELECT failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EIO;
	goto scanread_exit;
      }

      if (scan_ptr->adf_state == ADF_ARMED)
	scan_ptr->adf_state = ADF_CLEANUP; /* send unload on scanclose() */

#ifdef DEBUG_READ
      printf("scanread: SCSI_MODE_SELECT ok\n");
#endif

      /* fall through to SET_WINDOW... */

    case FUJITSU:
    case UMAX_UC630:
    case UMAX_UG630:

#ifdef DEBUG_READ
      printf("scanread: sending SCSI_SET_WINDOW\n");
#endif

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_SET_WINDOW, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_SET_WINDOW failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EINVAL;
	goto scanread_exit;
      }
#ifdef DEBUG_READ
      printf("scanread: SCSI_SET_WINDOW ok\n");
#endif
      scan_ptr->configured = TRUE;
      rc = 0;
      break;

    case SHARP_JX600:
#ifdef DEBUG_READ
      printf("scanread: sending SCSI_WRITE_EXTENDED\n");
#endif

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE_EXTENDED, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_WRITE_EXTENDED failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EINVAL;
	goto scanread_exit;
      }
      scan_ptr->configured = TRUE;
      rc = 0;
      break;

    case HP_SCANJET_IIC:
#ifdef DEBUG_READ
      printf("scanread: sending SCSI_WRITE of set_window data\n");
#endif

      scan_ptr->hp_command = HP_CONFIG;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_WRITE failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EINVAL;
	goto scanread_exit;
      }

      scan_ptr->configured = TRUE;
      rc = 0;
      break;
    }
  }

  /* this section does the */
  /* 1 time calculation of how much data there is to move (window size) */
  /* and the */
  /* 1 time trigger to start scanning (for most scanners) */

  if ((scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
       scan_ptr->s_dds.scanner_type == IBM_2456    ||
       scan_ptr->s_dds.scanner_type == RICOH_FS1   ||
       scan_ptr->s_dds.scanner_type == RICOH_IS50  ||
       scan_ptr->s_dds.scanner_type == SHARP_JX600 ||
       scan_ptr->s_dds.scanner_type == UMAX_UC630  ||
       scan_ptr->s_dds.scanner_type == UMAX_UG630  ||
       scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC)
      && scan_ptr->triggered == FALSE) {

    /*
     * deal with the fact that the Sharp JX600S uses 1/25" not 1/1200"
     * as its base unit of measurement
     */

    if (scan_ptr->s_dds.scanner_type == SHARP_JX600) {
      scan_ptr->s_dds.window_length /= 48;
      scan_ptr->s_dds.window_length *= 48;

      scan_ptr->s_dds.window_width /= 48;
      scan_ptr->s_dds.window_width *= 48;
    }

    /*
     * deal with the fact that the HP ScanJet IIc uses 1/300" not 1/1200"
     * as its base unit of measurement
     */

    if (scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC) {
      scan_ptr->s_dds.window_length /= 4;
      scan_ptr->s_dds.window_length *= 4;

      scan_ptr->s_dds.window_width /= 4;
      scan_ptr->s_dds.window_width *= 4;
    }

    if ((scan_ptr->s_dds.scanner_type == UMAX_UC630  ||
	 scan_ptr->s_dds.scanner_type == UMAX_UG630) &&
	scan_ptr->s_dds.image_mode >= GRAYSCALE) {
      int row_len;

      row_len = calc_umax_row_len(scan_ptr->s_dds.x_resolution,
				  scan_ptr->s_dds.window_width);

      scan_ptr->window_size = row_len *
	((scan_ptr->s_dds.window_length * scan_ptr->s_dds.y_resolution) / 1200);

    } else {
      scan_ptr->window_size =
	window_size(scan_ptr->s_dds.window_width, scan_ptr->s_dds.x_resolution,
		    scan_ptr->s_dds.window_length, scan_ptr->s_dds.y_resolution)
	  * (scan_ptr->s_dds.image_mode == COLOR ? 3 : 1)
	    / (scan_ptr->s_dds.image_mode < GRAYSCALE ? 8 : 1);
    }
#ifdef DEBUG_READ
    printf("scanread: window size = %d\n", scan_ptr->window_size);
#endif

    /*
     * since Ricoh & UMAX scanners in monochrome mode throw away bits at
     * the end of a row that exceed the byte boundary, you need to shrink
     * the window to reflect the data that are lost by this step
     *
     * <window size> -= <number of rows> * (<pix per row> - <pix per row> % 8)
     */

    if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
	scan_ptr->s_dds.scanner_type == IBM_2456    ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50  ||
	scan_ptr->s_dds.scanner_type == UMAX_UC630  ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630)
      if (scan_ptr->s_dds.image_mode == BINARY_MONOCHROME ||
	  scan_ptr->s_dds.image_mode == DITHERED_MONOCHROME) {

	int num_rows =
	  scan_ptr->s_dds.window_length * scan_ptr->s_dds.y_resolution / 1200;

	int bytes_per_row =
	  scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution / 9600;

	scan_ptr->window_size = num_rows * bytes_per_row;
      }

#ifdef DEBUG_READ
    printf("scanread: window is  %d bytes total\n", scan_ptr->window_size);
#endif

    /********************************************************************/
    /*** end of truncate-to-byte-boundary window size adjustment step ***/
    /********************************************************************/


    /* do what it takes to trigger the scanning operation */

    switch (scan_ptr->s_dds.scanner_type) {
    case SHARP_JX600:
      /* nothing */
      break;

    case HP_SCANJET_IIC:
      /*
       * remember if the HP ScanJet IIc had paper in the ADF when the read
       * started because if the read is aborted the scanner will not unload
       * the ADF
       */
      if (scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC)
	scan_ptr->using_adf = hp_has_paper_in_adf(scan_ptr);

      scan_ptr->hp_command = HP_TRIGGER;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: HP_TRIGGER failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EIO;
	goto scanread_exit;
      }
#ifdef DEBUG_READ
      printf("HP_TRIGGER_SCAN successful\n");
#endif
      break;

    default:

#ifdef DEBUG_READ
      printf("scanread: sending SCSI_TRIGGER_SCAN\n");
#endif
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_TRIGGER_SCAN, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_TRIGGER_SCAN failed\n");
#endif
	do_request_sense(scan_ptr);

	rc = EIO;
	goto scanread_exit;
      }
#ifdef DEBUG_READ
      printf("SCSI_TRIGGER_SCAN successful\n");
#endif
    }

    /*
     * remember if the UMAX had paper in the ADF when the read started
     * because the ADF paper sensor is released at the end of the read
     * and we need to know how many unloads to send in order to eject
     * the last sheet and home the carriage
     */
    if (scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630) {
      find_umax(scan_ptr);
      scan_ptr->using_adf = umax_has_paper_in_adf(scan_ptr);
    }

    scan_ptr->triggered = TRUE;
  }

  /**********************************/
  /*** end of one-time init stuff ***/
  /**********************************/

  /* ************************************************ */
  /* EOF detection for scanners that don't do EOM bit */
  /* ************************************************ */

  switch (scan_ptr->s_dds.scanner_type) {
  case RICOH_IS410:
  case RICOH_FS1:
  case SHARP_JX600:
  case RICOH_IS50:
  case IBM_2456:
  case UMAX_UC630:
  case UMAX_UG630:
  case HP_SCANJET_IIC:
    if (scan_ptr->window_size <= 0) {
      uiop->uio_resid = uiop->uio_iov->iov_len;	/* Sorry, */
      uiop->uio_iov++;				/* no data */
      uiop->uio_iovcnt--;			/* for you */
      uiop->uio_iovdcnt++;			/* today. */

      rc = 0;			/* EOF is indicated by no data & no error */
      goto scanread_exit;	/* at the same time. */
    }
  }

  /* ***************************************************** */
  /* end of EOF section for scanners that don't do EOM bit */
  /* ***************************************************** */

  /* ******************************* */
  /* start of RICOH specific polling */
  /* ******************************* */

  if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
      scan_ptr->s_dds.scanner_type == IBM_2456    ||
      scan_ptr->s_dds.scanner_type == RICOH_IS50  ||
      scan_ptr->s_dds.scanner_type == RICOH_FS1) {

    tries = 10;		/* don't get stuck forever! */

    do {

      /* if you wait a bit slowpoke scanners perform better */
      if (scan_ptr->s_dds.scanner_type == RICOH_IS50)
	delay(50);

#ifdef DEBUG_READ
      printf("scanread: sending SCSI_GET_BUFFER_STATUS\n");
#endif

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_GET_BUFFER_STATUS, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_GET_BUFFER_STATUS failed\n");
#endif
	do_request_sense(scan_ptr);

	if (scan_ptr->rs.sense_key == 0x05 &&
	    scan_ptr->rs.additional_sense_code == 0x2c) {
#ifdef DEBUG_READ
	  printf("scanread: command out of sequence\n");
#endif
	  goto scanread_exit;
	}
      }

      scan_ptr->bytes_ready = 
	(scan_ptr->cmd_buf.scsi_buf.bufstruct.b_baddr[9] * 65536)
	  + (scan_ptr->cmd_buf.scsi_buf.bufstruct.b_baddr[10] * 256)
	    + (scan_ptr->cmd_buf.scsi_buf.bufstruct.b_baddr[11]);
#ifdef DEBUG_PERFORMANCE
      printf("scanread: %d bytes wanted\n", uiop->uio_iov->iov_len);
      printf("scanread: %d bytes ready\n", scan_ptr->bytes_ready);
#endif

    } while (scan_ptr->bytes_ready == 0 && --tries);

    /* ******************************************** */
    /* trim transfer down to the data that is ready */
    /* ******************************************** */

    if (scan_ptr->bytes_ready < uiop->uio_iov->iov_len) {
      uiop->uio_resid = uiop->uio_iov->iov_len - scan_ptr->bytes_ready;
      uiop->uio_iov->iov_len = scan_ptr->bytes_ready;
#ifdef DEBUG_READ
      printf("scanread: bytes wanted exceeds bytes ready\n");
      printf("scanread: xfer reduced to %d\n", uiop->uio_iov->iov_len);
#endif
    }
  }
  /* *********************************** */
  /* end of RICOH-specific polling stuff */
  /* *********************************** */

  /* ***************************************** */
  /* trim the transfer down to the window size */
  /* ***************************************** */

  if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
      scan_ptr->s_dds.scanner_type == IBM_2456    ||
      scan_ptr->s_dds.scanner_type == RICOH_IS50  ||
      scan_ptr->s_dds.scanner_type == UMAX_UC630  ||
      scan_ptr->s_dds.scanner_type == UMAX_UG630  ||
      scan_ptr->s_dds.scanner_type == RICOH_FS1   ||
      scan_ptr->s_dds.scanner_type == SHARP_JX600 ||
      scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC) {
    if (scan_ptr->window_size < uiop->uio_iov->iov_len) {
      uiop->uio_resid += uiop->uio_iov->iov_len - scan_ptr->window_size;
      uiop->uio_iov->iov_len = scan_ptr->window_size;
#ifdef DEBUG_READ
      printf("scanread: bytes wanted exceeds window size\n");
      printf("scanread: xfer reduced to %d\n", uiop->uio_iov->iov_len);
#endif
    }
  }

  /* ****************************************************** */
  /*  trim FS1 transfer down to an even scan line boundary  */
  /* ****************************************************** */

  /* note: the number 400 is from (1/2000") / (3 bytes-per-pixel) */

  if (scan_ptr->s_dds.scanner_type == RICOH_FS1) {
    if (uiop->uio_iov->iov_len %
	(scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution / 400)) {
      uiop->uio_resid += uiop->uio_iov->iov_len %
	(scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution / 400);
      uiop->uio_iov->iov_len -= uiop->uio_iov->iov_len %
	(scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution / 400);
#ifdef DEBUG_READ
      printf("scanread: a row is %d bytes\n",
	     (scan_ptr->s_dds.window_width *
	      scan_ptr->s_dds.x_resolution / 400));
      printf("scanread: transfer reduced to %d\n", uiop->uio_iov->iov_len);
#endif
      if (uiop->uio_iov->iov_len <= 0) {
#ifdef DEBUG_READ
	printf("scanread: the read request was smaller than one row\n");
#endif
	rc = EINVAL;
	goto scanread_exit;
      }
    }
  }


  /* ******************************************************** */
  /* trim Sharp JX600 read down to an even scan line boundary */
  /* ******************************************************** */

  if (scan_ptr->s_dds.scanner_type == SHARP_JX600) {

    int row_len = scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution;

    switch (scan_ptr->s_dds.image_mode) {
    case COLOR:
      row_len /= 400;
      break;
    case GRAYSCALE:
    case RED:
    case GREEN:
    case BLUE:
      row_len /= 1200;
      break;
    case BINARY_MONOCHROME:
      row_len /= 9600;
      break;
    }
    lines = uiop->uio_iov->iov_len / row_len;

    if (lines < 1) {
#ifdef DEBUG_READ
      printf("scanread: the read request was smaller than one row\n");
#endif
      rc = EINVAL;
      goto scanread_exit;     
    }

    if (uiop->uio_iov->iov_len % row_len &&
	uiop->uio_iov->iov_len > uiop->uio_iov->iov_len % row_len) {
      uiop->uio_resid += uiop->uio_iov->iov_len % row_len;
      uiop->uio_iov->iov_len -= uiop->uio_iov->iov_len % row_len;
#ifdef DEBUG_READ
      printf("scanread: a row is %d bytes (# of rows %d)\n", row_len, lines);
      printf("scanread: transfer reduced to %d\n", uiop->uio_iov->iov_len);
#endif
    }
  }

  /* ************************************************** */
  /* trim UMAX reads down to an even scan line boundary */
  /* ************************************************** */

  if (scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
      scan_ptr->s_dds.scanner_type == UMAX_UG630) {
    int row_len;

    if (scan_ptr->s_dds.image_mode >= GRAYSCALE)
      row_len = calc_umax_row_len(scan_ptr->s_dds.x_resolution,
				  scan_ptr->s_dds.window_width);
    else
      row_len =
	scan_ptr->s_dds.window_width * scan_ptr->s_dds.x_resolution / 9600;

    if (uiop->uio_iov->iov_len < row_len) {
#ifdef DEBUG_READ
      printf("scanread: the read request was smaller than one row\n");
#endif
      rc = EINVAL;
      goto scanread_exit;
    }

    if (uiop->uio_iov->iov_len % row_len) {
      uiop->uio_resid += uiop->uio_iov->iov_len % row_len;
      uiop->uio_iov->iov_len -= uiop->uio_iov->iov_len % row_len;
#ifdef DEBUG_READ
      printf("scanread: a row is %d bytes\n", row_len);
      printf("scanread: transfer reduced to %d bytes (%d rows)\n",
	     uiop->uio_iov->iov_len, uiop->uio_iov->iov_len / row_len);
#endif
    }
  }

  /* ***************************************** */
  /* end of transfer size re-adjusting section */
  /* ***************************************** */

  /* **************************************************************** */
  /* loop through the iovec's turning them into scsi_buf structs and  */
  /* passing the buf structs to the SCSI I/O controller device driver */
  /* **************************************************************** */

  sbp = &scan_ptr->data_buf;

  while (uiop->uio_iovcnt > 0) {

    iov = uiop->uio_iov;
    xmemp = uiop->uio_xmem;

    /* fill out the data_buf struct */

    sbp->scsi_buf.bufstruct.b_flags = B_READ;
    sbp->scsi_buf.bufstruct.b_forw = (struct buf *)sbp;
    sbp->scsi_buf.bufstruct.av_forw = NULL;
    sbp->scsi_buf.bufstruct.av_back = NULL;
    sbp->scsi_buf.bufstruct.b_iodone = scaniodone;
    sbp->scsi_buf.bufstruct.b_back = (struct buf *)sbp;
    sbp->scsi_buf.bufstruct.b_vp = NULL;
    sbp->scsi_buf.bufstruct.b_dev = scan_ptr->s_dds.parent_devno;
    sbp->scsi_buf.bufstruct.b_blkno = 0;
    sbp->scsi_buf.bufstruct.b_un.b_addr = iov->iov_base;
    sbp->scsi_buf.bufstruct.b_bcount = iov->iov_len;
    sbp->scsi_buf.bufstruct.b_error = 0;
    sbp->scsi_buf.bufstruct.b_resid = 0;
    sbp->scsi_buf.bufstruct.b_work  = 0;
    sbp->scsi_buf.bufstruct.b_options = 0;
    sbp->scsi_buf.bufstruct.b_event = EVENT_NULL;

    curtime(&sbp->scsi_buf.bufstruct.b_start);

    sbp->scan_ptr = scan_ptr;

    if (scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC) {
      scan_ptr->cmd_state = SCSI_READ;
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_length = (uchar)6;
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_op_code =
	SCSI_READ;
    } else {
      scan_ptr->cmd_state = SCSI_READ_EXTENDED;
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_length = (uchar)10;
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_op_code =
	SCSI_READ_EXTENDED;
    }

    scan_ptr->data_buf.scsi_buf.scsi_command.scsi_id =
      (uchar)scan_ptr->s_dds.scsi_id;
    scan_ptr->data_buf.scsi_buf.scsi_command.resvd = (uchar)0;

    if (scan_ptr->s_dds.scanner_type == RICOH_FS1  ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50 ||
	scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630 ||
	scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC)
      scan_ptr->data_buf.scsi_buf.scsi_command.flags = (uchar)SC_ASYNC;
    else
      scan_ptr->data_buf.scsi_buf.scsi_command.flags = (uchar)0;

    scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.lun =
      scan_ptr->s_dds.lun << 5;

    /* byte 2 is the data type: 0 for image data, 0x80 for pixel size */
    /* except for the Sharp, which uses 0x80 for image data */
    if (scan_ptr->s_dds.scanner_type == SHARP_JX600)
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0x80;
    else
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] = 0;

    /* byte 3 is reserved */
    scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] = 0;

    /* bytes 4-5 make up the data type qualifier */

    /* Fujitsu: 4 must be set to zero; 5 must be the window identifier */
    /* RICOH: 4-5 are data type qualifier -- 0 is image data*/
    /* Sharp: 4-5 are reserved and must be set to zero */
    scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] = 0;
    scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0;

    switch (scan_ptr->s_dds.scanner_type) {
    case HP_SCANJET_IIC:
      /* only 6 byte read request */
      /* bytes 2-4 are for transfer length */
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB2] =
	(uchar)((scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount >> 16) & 0xff);
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB3] =
	(uchar)((scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount >> 8) & 0xff);
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB4] =
	(uchar)(scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount & 0xff);

      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB5] = 0;
      break;

    case SHARP_JX600:
      /* reserved */
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] = 0;

      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] =
	(uchar)(lines >> 8);
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] =
	(uchar)(lines & 0xff);
      break;

    default:
      /* bytes 6-8 are for transfer length */
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB6] =
	(uchar)((scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount >> 16) & 0xff);
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB7] =
	(uchar)((scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount >> 8) & 0xff);
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB8] =
	(uchar)(scan_ptr->data_buf.scsi_buf.bufstruct.b_bcount & 0xff);

      /* the "control byte" aka vendor unique */
      scan_ptr->data_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[SCB9] = 0;
    }

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_IS410:
    case IBM_2456:
    case RICOH_FS1:
    case RICOH_IS50:
    case FUJITSU:
    case HP_SCANJET_IIC:
      scan_ptr->data_buf.scsi_buf.timeout_value = 15;
      break;

    case SHARP_JX600:
    case UMAX_UC630:
    case UMAX_UG630:
      scan_ptr->data_buf.scsi_buf.timeout_value = 90;
      break;

#ifdef DEBUG_READ
    default:
      printf("scanread: unknown scanner type\n");
#endif
    }

#ifdef IOSTAT
    scan_ptr->scan_dkstat.dk_status |= IOST_DK_BUSY;
#endif

    /* first pin the user space memory buffer and then xmattach to it */

    if ((rc = pinu(sbp->scsi_buf.bufstruct.b_un.b_addr,
	 sbp->scsi_buf.bufstruct.b_bcount,
	 uiop->uio_segflg)) != 0) {
#ifdef DEBUG_READ
      printf("scanread: pinu(0x%x, %d, %d) failed\n",
	     sbp->scsi_buf.bufstruct.b_un.b_addr,
	     sbp->scsi_buf.bufstruct.b_bcount,
	     uiop->uio_segflg);
#endif
      goto scanread_exit;
    }
    ++undo;			/* unpinu is now necessary */


    /* xmattach to the user proc's buffer */

    sbp->scsi_buf.bufstruct.b_xmemd.aspace_id = XMEM_INVAL;
    if (xmattach(iov->iov_base, iov->iov_len, &sbp->scsi_buf.bufstruct.b_xmemd,
	     uiop->uio_segflg) != XMEM_SUCC) {
#ifdef DEBUG_READ
      printf("scanread: xmattach(%d, %d, %d, %d) failed\n",
	     iov->iov_base, iov->iov_len,
	     &sbp->scsi_buf.bufstruct.b_xmemd, uiop->uio_segflg);
#endif
      rc = EFAULT;
      goto scanread_exit;
    }
    ++undo;			/* xmdetach is now necessary */

    devstrat((struct buf *)sbp);
    scanwait((struct buf *)sbp);

/*
 *  post processing: convert the buf struct back into a uio struct
 */

    uiop->uio_resid += sbp->scsi_buf.bufstruct.b_resid;

/*
 * reduce the byte count in "window_size" (check for underflows)
 */

    if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
	scan_ptr->s_dds.scanner_type == IBM_2456) {
      if (iov->iov_len - uiop->uio_resid <= 0)
	if (scan_ptr->bytes_ready <= scan_ptr->window_size)
	  scan_ptr->window_size -= scan_ptr->bytes_ready;
	else
	  scan_ptr->window_size = 0;
      else
	scan_ptr->window_size -= (iov->iov_len - uiop->uio_resid);
#ifdef DEBUG_READ
      printf("scanread: window now has %d bytes\n", scan_ptr->window_size);
#endif
    }

    if (scan_ptr->s_dds.scanner_type == RICOH_FS1  ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50 ||
	scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630 ||
	scan_ptr->s_dds.scanner_type == SHARP_JX600) {
      scan_ptr->window_size -= iov->iov_len;
#ifdef DEBUG_READ
      printf("scanread: iov_len:%d  resid:%d\n", iov->iov_len, uiop->uio_resid);
      printf("scanread: window now has %d bytes\n", scan_ptr->window_size);
#endif
    }

#ifdef IOSTAT
    scan_ptr->scan_dkstat.dk_rblks +=
	(iov->iov_len - sbp->scsi_buf.bufstruct.b_resid);

    ++scan_ptr->scan_dkstat.dk_xfers;
#endif

    xmdetach(&sbp->scsi_buf.bufstruct.b_xmemd);
    --undo;

    unpinu(sbp->scsi_buf.bufstruct.b_un.b_addr,
	   sbp->scsi_buf.bufstruct.b_bcount,
	   uiop->uio_segflg);
    --undo;

#ifdef DEBUG_PIN
    printf("scanread: unpinu(0x%x, %d, %d)\n",
	   sbp->scsi_buf.bufstruct.b_un.b_addr,
	   sbp->scsi_buf.bufstruct.b_bcount,
	   uiop->uio_segflg);
#endif

    if (scan_ptr->check_condition) {	     /* test for EOM condition */
      scan_ptr->check_condition = FALSE;
      scan_ptr->retry = TRUE;		     /* error recovery is happening */

      scan_scsi_cmd(SCSI_REQUEST_SENSE, scan_ptr); 
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

#ifdef DEBUG_READ
      printf("scanread: EOM test. error code 0x%x; sense key is 0x%x\n",
	     scan_ptr->rs.err_code, scan_ptr->rs.sense_key);
#endif

      if (scan_ptr->rs.sense_key & EOM_BIT) { /* EOM detected, no more data */
	rc = 0;
	goto scanread_exit;
      }

      if (scan_ptr->rs.sense_key & ILI_BIT) { /* EOM UMAX style */
#ifdef DEBUG_READ
	printf("scanread: oops! ILI detected\n");
	if (scan_ptr->rs.err_code & 0x80)
	  printf("scanread: you tried to read %d too much\n",
		 (scan_ptr->rs.info1 * 65536 +
		  scan_ptr->rs.info2 * 256 +
		  scan_ptr->rs.info3));
#endif
	rc = EIO;
	goto scanread_exit;
      }

      if (scan_ptr->rs.sense_key == 0) { /* no error? */
	rc = 0;
	goto scanread_exit;
      }

      rc = EIO;
      goto scanread_exit;
    }

    if (scan_ptr->adapter_error) {
#ifdef DEBUG_READ
      printf("scanread: adapter error -- aborting\n");
#endif
      scan_ptr->retry = TRUE;		     /* error recovery is happening */
      scan_ptr->adapter_error = FALSE;
      rc = EIO;
      goto scanread_exit;
    }

    uiop->uio_iov++;
    uiop->uio_iovcnt--;
    uiop->uio_iovdcnt++;
  }

scanread_exit:

  if (rc == 0) {
#ifdef DEBUG_READ
    printf("scanread: exiting ok\n");
#endif
    return (0);
  }

  switch (undo) {
  case 2:			/* xmdetach is now necessary */
    xmdetach(&sbp->scsi_buf.bufstruct.b_xmemd);

    /* fall through... */

  case 1:			/* unpinu is now necessary */

    unpinu(sbp->scsi_buf.bufstruct.b_un.b_addr,
	   sbp->scsi_buf.bufstruct.b_bcount,
	   uiop->uio_segflg);
#ifdef DEBUG_PIN
    printf("scanread unpinu(0x%x, %d, %d)\n",
	   sbp->scsi_buf.bufstruct.b_un.b_addr,
	   sbp->scsi_buf.bufstruct.b_bcount,
	   uiop->uio_segflg);
#endif
    break;

  default:
#ifdef DEBUG_READ
    printf("scanread: exiting with level %d undo processing\n", undo);
    printf("scanread: returning %d\n", rc);
#endif
    break;
  }

  return(rc);

}


/*
 *   scanioctl: ioctl routine for scan device driver
 */

int
scanioctl(dev_t devno, int cmd, int arg, ulong devflag, chan_t chan, int ext)
{
  struct scan_df *scan_ptr;
  struct scan_io s_io;
  struct scan_pixel_sizes pix_size;
  struct sc_readblk rdblk;
  char scanner_offset;
  uchar undo = 0;
  int rc = 0;

#ifdef DEBUG_IOCTL
  printf("scanioctl called, cmd = %d ", cmd);
#endif

  /****	 locate scanner struct	***/

#ifdef DEBUG_IOCTL
  printf("(lock");
#endif
  (void)lockl(&scanner_chain_lock, LOCK_SHORT); /* unlockl() is now necessary*/
#ifdef DEBUG_IOCTL
  printf("ing)\n");
#endif

  scanner_offset = minor(devno) & 0x0f;
  scan_ptr = scanner_list[scanner_offset].defined_head;
  while (scan_ptr != NULL) {
    if (scan_ptr->devno == devno)
      break;
    else
      scan_ptr = scan_ptr->next_defined;
  }
  if (scan_ptr == NULL) {
#ifdef DEBUG_IOCTL
    printf("but the scanner couldn't be found\n");
#endif
    unlockl(&scanner_chain_lock);
    return (ENODEV);	      /* can't find the scanner in the defined list */
  }

  unlockl(&scanner_chain_lock);
#ifdef DEBUG_IOCTL
    printf("(unlocked)\n");
#endif

  switch (cmd) {
  case SCAN_GET:
    s_io.scan_window_width  = scan_ptr->s_dds.window_width;
    s_io.scan_window_length = scan_ptr->s_dds.window_length;
    s_io.scan_x_resolution  = scan_ptr->s_dds.x_resolution;
    s_io.scan_y_resolution  = scan_ptr->s_dds.y_resolution;
    s_io.scan_x_origin	    = scan_ptr->s_dds.x_origin;
    s_io.scan_y_origin	    = scan_ptr->s_dds.y_origin;
    s_io.scan_image_mode    = scan_ptr->s_dds.image_mode;
    s_io.scan_brightness    = scan_ptr->s_dds.brightness;
    s_io.scan_contrast      = scan_ptr->s_dds.contrast;
    s_io.scan_scanner_type  = scan_ptr->s_dds.scanner_type;

    if (devflag & FKERNEL) {			 /* kernel process */
      bcopy(&s_io, (caddr_t)arg,  sizeof(s_io));
      return (0);
    } else {					 /* user process */
      return
	(copyout(&s_io, (caddr_t)arg, sizeof(s_io)));
    }

  case SCAN_SET:
    if (devflag & FKERNEL) {			 /* kernel process */
      bcopy((caddr_t)arg, &s_io, sizeof(s_io));
    } else {					 /* user process */
      rc = copyin((caddr_t)arg, &s_io, sizeof(s_io));
      if (rc)
	return (rc);
    }

#ifdef DEBUG_IOCTL
    printf("scanioctl: w/l/x/y are %d/%d/%d/%d\n",
	   s_io.scan_window_width,
	   s_io.scan_window_length,
	   s_io.scan_x_origin,
	   s_io.scan_y_origin);
    printf("scanioctl: data size is %d\n",
	   (s_io.scan_window_width / 1200) * s_io.scan_x_resolution *
	   (s_io.scan_window_length / 1200) * s_io.scan_y_resolution);
#endif


    /****  Validate parameters  ***/

    switch (scan_ptr->s_dds.scanner_type) {
    case FUJITSU:

      /* size constraints... (these values are taken from the OEM manual) */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 14592	|| /* 12.16" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 20736)	   /* 17.28" */
	return (EINVAL);

      /* Fujitsu can only handle 4 Meg windows max */

      if ((s_io.scan_window_width / 1200) * s_io.scan_x_resolution *
	  (s_io.scan_window_length / 1200) * s_io.scan_y_resolution
	  > (s_io.scan_image_mode == GRAYSCALE ? 4063232 : 32505856))
	return (EINVAL);

      switch (s_io.scan_x_resolution) {
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io.scan_y_resolution) {
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac FUJITSU */

    case RICOH_IS410:
    case IBM_2456:

      /* size constraints... */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 13800	|| /* 11.5" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 20400)    /* 17" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io.scan_x_resolution < 100 || s_io.scan_x_resolution > 400 ||
	  s_io.scan_y_resolution < 100 || s_io.scan_y_resolution > 400)
	return (EINVAL);

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac RICOH_IS410 (IBM_2456) */

    case RICOH_FS1:

      /* size constraints... */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 10500	|| /* 8.75" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 14400)   /* 12" */
	return (EINVAL);

      /* resolution (dpi)... */

      switch (s_io.scan_x_resolution) {
      case 120:
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io.scan_y_resolution) {
      case 120:
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      if (s_io.scan_image_mode != COLOR)
	return (EINVAL);

      break;			/* esac RICOH_FS1 */

    case SHARP_JX600:

      /* size constraints... */

      if (s_io.scan_window_width <= 0		||
	  s_io.scan_x_origin
	  + s_io.scan_window_width > 14208	|| /* 11.84" (296 25ths) */
	  s_io.scan_window_length <= 0		||
	  s_io.scan_y_origin
	  + s_io.scan_window_length > 20880) /* 17.40" (435 25ths) */
	return (EINVAL);

      switch (s_io.scan_x_resolution) {
      case 75:
      case 100:
      case 150:
      case 200:
      case 300:
      case 400:
      case 600:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io.scan_y_resolution) {
      case 75:
      case 100:
      case 150:
      case 200:
      case 300:
      case 400:
      case 600:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case GRAYSCALE:
      case COLOR:
      case RED:
      case GREEN:
      case BLUE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac SHARP_JX600 */


    case RICOH_IS50:

      /* size constraints... */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 10200	|| /* 8.5" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io.scan_x_resolution < 100 || s_io.scan_x_resolution > 400 ||
	  s_io.scan_y_resolution < 100 || s_io.scan_y_resolution > 400)
	return (EINVAL);

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac RICOH_IS50 */

    case UMAX_UC630:
      
      /* size constraints... */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 10200	|| /* 8.5" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin +  s_io.scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io.scan_x_resolution < 100 || s_io.scan_x_resolution > 300 ||
	  s_io.scan_y_resolution < 100 || s_io.scan_y_resolution > 600)
	return (EINVAL);

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
      case RED:
      case GREEN:
      case BLUE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac UMAX UC630 */

    case UMAX_UG630:
      
      /* size constraints... */

      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 10200	|| /* 8.5" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io.scan_x_resolution < 100 || s_io.scan_x_resolution > 300 ||
	  s_io.scan_y_resolution < 100 || s_io.scan_y_resolution > 300)
	return (EINVAL);

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac UMAX UG630 */

    case HP_SCANJET_IIC:
      if (s_io.scan_window_width <= 0				||
	  s_io.scan_x_origin + s_io.scan_window_width > 10200	|| /* 8.5" */
	  s_io.scan_window_length <= 0				||
	  s_io.scan_y_origin + s_io.scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io.scan_x_resolution < 100 || s_io.scan_x_resolution > 400 ||
	  s_io.scan_y_resolution < 100 || s_io.scan_y_resolution > 400)
	return (EINVAL);

      switch (s_io.scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
      case COLOR:
	break;
      default:
	return (EINVAL);
      }

      break;

#ifdef DEBUG_IOCTL
    default:
      printf("scanioctl: unsupported scanner type: %d\n",
	     scan_ptr->s_dds.scanner_type);
#endif
    }

    /****  change the DDS to the new values  ***/

    scan_ptr->s_dds.window_width  = s_io.scan_window_width;
    scan_ptr->s_dds.window_length = s_io.scan_window_length;
    scan_ptr->s_dds.x_resolution  = s_io.scan_x_resolution;
    scan_ptr->s_dds.y_resolution  = s_io.scan_y_resolution;
    scan_ptr->s_dds.x_origin	  = s_io.scan_x_origin;
    scan_ptr->s_dds.y_origin	  = s_io.scan_y_origin;
    scan_ptr->s_dds.image_mode	  = s_io.scan_image_mode;
    scan_ptr->s_dds.brightness    = s_io.scan_brightness;
    scan_ptr->s_dds.contrast      = s_io.scan_contrast;

    /****  do what it takes to actualize the new values  ***/

    /* if the scanner is currently reading data don't change the
     * parameters since many scanners will break out of read-mode
     */
    if (scan_ptr->configured == TRUE) /* configured flag gets set when */
      break;                          /* scanner enters read-mode */

    switch (scan_ptr->s_dds.scanner_type) {
    case RICOH_IS410:		/* support for scanners that need */
    case IBM_2456:		/* ADF mode to be set in firmware */
    case RICOH_IS50:
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;
      scan_scsi_cmd(SCSI_MODE_SELECT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_MODE_SELECT failed\n");
#endif
	do_request_sense(scan_ptr);

	return (EIO);
      }
      /* fall through */

    case RICOH_FS1:
    case FUJITSU:
    case UMAX_UC630:
    case UMAX_UG630:

      /****  send a SCSI_SET_WINDOW ***/

#ifdef DEBUG_IOCTL
      printf("scanioctl: sending SCSI_SET_WINDOW\n");
#endif

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_SET_WINDOW, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_READ
	printf("scanread: SCSI_SET_WINDOW failed\n");
#endif
	do_request_sense(scan_ptr);

	return (EIO);
      }
#ifdef DEBUG_IOCTL
      printf("scanioctl: SCSI_SET_WINDOW ok\n");
#endif
      break;

    case SHARP_JX600:
      
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE_EXTENDED, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_IOCTL
	printf("scanioctl: SCSI_WRITE_EXTENDED failed\n");
#endif
	do_request_sense(scan_ptr);

	return (EIO);
      }

      break;

    case HP_SCANJET_IIC:
      scan_ptr->hp_command = HP_CONFIG;
      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

      scan_scsi_cmd(SCSI_WRITE, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

      if (scan_ptr->check_condition) {
#ifdef DEBUG_IOCTL
	printf("scanioctl: SCSI_WRITE failed\n");
#endif
	do_request_sense(scan_ptr);

	return (EIO);
      }

      break;

#ifdef DEBUG_IOCTL
    default:
      printf("scanioctl: unsupported scanner type: %d\n",
	     scan_ptr->s_dds.scanner_type);
#endif
    }

    break;

  case SCAN_USE_ADF:
    /* Do what it takes to make the next read come from the ADF.
     * If the ADF is not ready (i.e. empty or jammed) return EIO.
     * If the scanner is in read-mode return EINVAL.
     */

    /* if the scanner is currently reading data don't change the
     * paper since many scanners will break out of read-mode
     */
    if (scan_ptr->configured == TRUE &&
	scan_ptr->window_size > 0)
      return (EINVAL);

#ifdef DEBUG_IOCTL
    printf("scanioctl: sending SCSI_OBJECT_POSITION unload\n");
#endif

    /* UMAX UC630/UG630 just tests ADF for paper */
    if (scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630) {
      if (umax_has_paper_in_adf(scan_ptr))
	return (0);
      else
	return (EIO);
    }

    /* HP ScanJet IIc just tests ADF to see if it is ready */
    if (scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC) {
      if (hp_has_paper_in_adf(scan_ptr))
	return (0);
      else
	return (EIO);
    }

    if (scan_ptr->s_dds.scanner_type == RICOH_IS410 ||
	scan_ptr->s_dds.scanner_type == IBM_2456 ||
	scan_ptr->s_dds.scanner_type == RICOH_IS50) {

      scan_ptr->adf_state = ADF_ARMED;

      scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;
      scan_scsi_cmd(SCSI_MODE_SELECT, scan_ptr);
      scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);
      if (scan_ptr->check_condition) {
#ifdef DEBUG_IOCTL
	printf("scanioctl: SCSI_MODE_SELECT failed\n");
#endif
	do_request_sense(scan_ptr);

	return (EIO);
      }
    }

    /*
     * Note that the SCSI_OBJECT_POSITION command will automatically toggle
     * between the "unload" and "load" states, so just call it twice to
     * get the whole job done.
     */

    /* unload the paper */
    scan_ptr->do_unload = TRUE;
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

    scan_scsi_cmd(SCSI_OBJECT_POSITION, scan_ptr);
    scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

    if (scan_ptr->check_condition) {
#ifdef DEBUG_IOCTL
      printf("scanioctl: SCSI_OBJECT_POSITION unload failed\n");
#endif
      do_request_sense(scan_ptr);
    }

#ifdef DEBUG_IOCTL
    printf("scanioctl: sending SCSI_OBJECT_POSITION load\n");
#endif

    /* load the paper */
    scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

    scan_scsi_cmd(SCSI_OBJECT_POSITION, scan_ptr);
    scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

    if (scan_ptr->check_condition) {
#ifdef DEBUG_IOCTL
      printf("scanioctl: SCSI_OBJECT_POSITION failed\n");
#endif
      do_request_sense(scan_ptr);

      scan_ptr->adf_state = ADF_UNUSED;
      return (EIO);
    }
    scan_ptr->adf_state = ADF_ARMED;
    break;

  case SCAN_GET_PIXEL_SIZE:
    if (scan_ptr->s_dds.scanner_type == SHARP_JX600) {
      scan_ptr->s_dds.window_length /= 48;
      scan_ptr->s_dds.window_length *= 48;

      scan_ptr->s_dds.window_width /= 48;
      scan_ptr->s_dds.window_width *= 48;
    }
    if ((scan_ptr->s_dds.scanner_type == UMAX_UC630 ||
	scan_ptr->s_dds.scanner_type == UMAX_UG630) &&
	scan_ptr->s_dds.image_mode >= GRAYSCALE)
      pix_size.pix_width = calc_umax_row_len(scan_ptr->s_dds.x_resolution,
					     scan_ptr->s_dds.window_width);
    else {
      pix_size.pix_width = (scan_ptr->s_dds.window_width
			    * scan_ptr->s_dds.x_resolution
			    / 1200);

      if (scan_ptr->s_dds.image_mode < GRAYSCALE) {
	if (scan_ptr->s_dds.scanner_type == FUJITSU ||/* pad to byte boundary */
	    scan_ptr->s_dds.scanner_type == HP_SCANJET_IIC)
	  pix_size.pix_width = (pix_size.pix_width + 7) / 8;
	else			                /* truncate to byte boundary */
	  pix_size.pix_width /= 8;
	pix_size.pix_width *= 8;
      }
    }
    pix_size.pix_height = (scan_ptr->s_dds.window_length
			   * scan_ptr->s_dds.y_resolution
			   / 1200);

    if (devflag & FKERNEL) {			 /* kernel process */
      bcopy(&pix_size, (caddr_t)arg,  sizeof(pix_size));
      return (0);
    } else {					 /* user process */
      return (copyout(&pix_size, (caddr_t)arg, sizeof(pix_size)));
    }
    break;

  default:
    return (EINVAL);
  }

#ifdef DEBUG_IOCTL
  printf("scanioctl exiting successfully\n");
#endif

  return (0);
}

#ifdef DEBUG_UIO
print_uio(struct uio *uiop)
{
  printf("uiop->uio_iov->iov_base = 0x%x\n", uiop->uio_iov->iov_base);
  printf("uiop->uio_iov->iov_len  = %d\n",   uiop->uio_iov->iov_len);
  printf("uiop->uio_iovcnt  = %d\n", uiop->uio_iovcnt);
  printf("uiop->uio_iovdcnt = %d\n", uiop->uio_iovdcnt);
  printf("uiop->uio_offset  = %d\n", uiop->uio_offset);
  printf("uiop->uio_resid   = %d\n", uiop->uio_resid);
}
#endif

int calc_umax_row_len(int dpi, int ww)
{
  int st[301];
  int i;
  int rowB = 0;

  for (i = 1; i <= 300; i++)
    st[i] = 1;

  for (i = 1; i <= 300 - dpi; i++)
    st[i * 300 / (300 - dpi)] = 0;

  for (i = 1; i <= (ww % 1200) / 4; i++) {
    if (st[i])
      rowB++;
  }

  return ((ww / 1200) * dpi + rowB);
}

int umax_has_paper_in_adf(struct scan_df *scan_ptr)
{
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

  scan_scsi_cmd(SCSI_INQUIRY, scan_ptr);
  scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

  if (scan_ptr->inq.device_type_qualifier == 1)
    return (1);			/* paper in ADF */
  else
    return (0);			/* no paper in ADF */
}


/*
 * since UMAX scanners return an "operation complete" message even before
 * finishing an unload or trigger operation, poll the scanner with TEST
 * UNIT READY until it's really done and comes back online.
 */

void find_umax(struct scan_df *scan_ptr)
{
  struct sc_ready scan_ready;
  int rc;

  do {
    scan_ready.flags   = (uchar)SC_ASYNC;
    scan_ready.scsi_id = scan_ptr->s_dds.scsi_id;
    scan_ready.lun_id  = scan_ptr->s_dds.lun;
    scan_ready.status_validity = (uchar)0;
    scan_ready.scsi_status     = (uchar)0;
    scan_ready.rsv1    = (uchar)0;
    scan_ready.rsv2    = (uchar)0;
    scan_ready.rsv3    = (uchar)0;
    scan_ready.rsv4    = (uint)0;
    if (rc = fp_ioctl(scan_ptr->fp, SCIOTUR, &scan_ready, NULL)) {
      scan_ptr->retry = TRUE;
      delay(100);
    } else
      scan_ptr->retry = FALSE;
  } while (rc);
}

void do_request_sense(struct scan_df *scan_ptr)
{
#ifdef DEBUG_REQUEST_SENSE
  printf("do_request_sense: sending SCSI_REQUEST_SENSE\n");
#endif
  scan_ptr->check_condition = FALSE;
  scan_ptr->retry = TRUE;
  scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= SC_ASYNC;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

  scan_scsi_cmd(SCSI_REQUEST_SENSE, scan_ptr); 
  scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

#ifdef DEBUG_REQUEST_SENSE
  printf("do_request_sense: sense data is 0x%x/0x%x/0x%x\n",
	 scan_ptr->rs.sense_key,
	 scan_ptr->rs.additional_sense_code,
	 scan_ptr->rs.additional_sense_code_qualifier);
#endif
  scan_ptr->retry = FALSE;
}

int hp_has_paper_in_adf(struct scan_df *scan_ptr)
{
  scan_ptr->hp_command = HP_PAPER_SENSE;
  scan_ptr->cmd_buf.scsi_buf.scsi_command.flags |= SC_ASYNC;
  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

  scan_scsi_cmd(SCSI_WRITE, scan_ptr);
  scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

  if (scan_ptr->check_condition) {
    do_request_sense(scan_ptr);
  }

  scan_ptr->cmd_buf.scsi_buf.bufstruct.b_event = EVENT_NULL;

  scan_scsi_cmd(SCSI_READ, scan_ptr);
  scanwait(&scan_ptr->cmd_buf.scsi_buf.bufstruct);

  if (scan_ptr->check_condition) {
    do_request_sense(scan_ptr);
  }

#ifdef DEBUG_IOCTL
  printf("scanioctl: hp_response is:  ");
  {
    int i = 0;
    while (scan_ptr->hp_response[i]) {
      if (scan_ptr->hp_response[i] == 033)
	printf("[esc]");
      else
	printf("%c", scan_ptr->hp_response[i]);
      ++i;
      if (i == 10)
	break;
    }
    printf("\n");
  }
#endif
  if (scan_ptr->hp_response[8] == '1')
    return (1);			/* paper present & ready */
  else
    return (0);			/* no paper, or ADF not ready */
}
