/*
 *  PINT Pint Is Not TWAIN - common scanner driver interface for UN*X
 *  Copyright (C) 1994 Kenneth Stailey ken@spacenet.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  scan.c: PINT scanner device driver for SunOS 4.1 using SCSI scanners
 *  third release version 0.3
 */

#include "scan.h"
#if NSCAN > 0

#define FALSE 0
#define TRUE 1

/*
 * NOTES
 *
 * XXX	Code of interest and places where you might need to make changes
 *	are marked with 'XXX'.
 *
 * XXX	Sequential Access Device Block Size can be fixed or variable,
 *	depending on the device. The read/write command CDB has a bit,
 *	'fixed' that's set to 1 for fixed block size when the transfer
 *	count specifies the # of blocks, or to 0 for variable block
 *	size when the transfer count specifies the # of bytes. See the
 *	#ifdef VAR_BLK_SZ section of code.
 *
 * XXX Two methods of sending commands to the target are provided: scancmd
 *	and scansimple. Scansimple is for simple non-DVMA commands; it sends
 *	a polled (i.e. non-interrupting, synchronous) command via scsi_poll.
 *	Scancmd is more general; it uses scanstrategy and scan_make_cmd to
 *	send an 'in band' command, i.e. queued with read/write requests.
 *	Scancmd can do DVMA; this is handled by DVMA-ing to a local buffer
 *	(un_tmpbuf) and then copying the data into the user's buffer. Both
 *	scancmd and scansimple are called from scanioctl. If you want to add
 *	a new ioctl, you can use either scancmd or scansimple. Since scancmd
 *	calls scan_make_cmd, any new ioctls must have the appropriate entry
 *	added to scan_make_cmd's case statement.
 */

#include <scsi/scsi.h>
#include <sys/scanio.h>
#include "scandd.h"
#include <vm/hat.h>
#include <vm/seg.h>
#include <vm/as.h>

#define DNAME		devp->sd_dev->devi_name
#define DUNIT		devp->sd_dev->devi_unit
#define CNAME		devp->sd_dev->devi_parent->devi_name
#define CUNIT		devp->sd_dev->devi_parent->devi_unit

/*
 * the buf struct's back pointer is used for the scsi packet
 */
#define BP_PKT(bp)      ((struct scsi_pkt *)bp->av_back)
#define SCBP(pkt)	((struct scsi_status *)(pkt)->pkt_scbp)
#define SCBP_C(pkt)	((*(pkt)->pkt_scbp) & STATUS_MASK)
#define CDBP(pkt)	((union scsi_cdb *)(pkt)->pkt_cdbp)
#define TGT(devp)	(devp->sd_address.a_target)
#define LUN(devp)	(devp->sd_address.a_lun)
#define SCANPDS		((struct scsi_scan *)(devp)->sd_private)
#define SCANUNIT(d)	(minor((d))>>3)

#define newstate(s) SCANPDS->un_last_state=SCANPDS->un_state, SCANPDS->un_state=(s)

#define SCANTIMEOUT	  2*hz	/* 2sec wait on a busy status return */
#define DFLT_SCAN_TIMEOUT 20	/* 20sec should be enough for most operations */
#define SCAN_RETRY_COUNT  15	/* max # retries on an operation */

#define VIDSZ		8
#define PIDSZ		16

/*
 * Maximum single DMA transfer size
 * 63k is the maximum allowed by SunOS.
 */
#define SCAN_MAXREC	(63 * 1024)

/*
 * Maximum number of scan peripherals
 * This is set by the structure of the minor device number. E.g. for
 * a tape device, 1 bit might be used to signal no rewind and 2 or 3
 * bits for different densities, leaving 4 or 5 bits for the device
 * number. Theoretically you could have 256 devices, but 8 seems like
 * a more reasonable number. This defines the length of the scanunits
 * array for the devices' private data structures.
 */
#define SCAN_MAXUNITS	4

/*
 * Adb-able debug flag/level: 0=off, 1=errors, 2 & 3 more verbose
 * To set: # adb -kw /vmunix /dev/mem
 *	   physmem <some value>		<- msg from adb
 *	   scan_debug/X			<- type this to display scan_debug
 *	   scan_debug/W <newval>	<- type this to change scan_debug
 *	   $q				<- type this to exit adb
 */
int scan_debug = 0;

#define DPRINTF(L)	if (scan_debug >= L) printf

  /*
   * Driver private data struct
   */
  struct scsi_device *scanunits[SCAN_MAXUNITS];

int		nscans = NSCAN;
static int	scanpri;
static int	scan_error_reporting = SCANERR_RETRYABLE;

/*
 * other externs
 */
extern char *scan_cmds[];	/* acutally in this file */

/*
 * Configuration Data
 */

/*
 * Device driver ops vector
 */
int scanslave(), scanattach(), scanopen(), scanclose(), scanread();
int scanstrategy(), scanioctl(), scansize();
extern int nulldev(), nodev();
struct dev_ops scan_ops = {
  1,
  scanslave,
  scanattach,
  scanopen,
  scanclose,
  scanread,
  nodev,
  scanstrategy,
  nodev,
  nulldev,
  scanioctl,
};

static void scanintr();
static void scandone(), scan_make_cmd(), scanerrmsg();
static void clean_print(), find_umax();
static int do_object_position();
static int calc_umax_row_len();
static int umax_has_paper_in_adf();
static long window_size();

static u_char model_to_type_code();
/*
 *********************************
 *
 * Autoconfiguration Routines
 *
 *********************************
 */

/*
 * scanslave
 * Calls scsi_slave() to see if there's a device at our target id. If
 * there is, scsi_slave will fill in the sd_inq struct (in the devp
 * scsi device struct) with the inquiry data. Validate the data here,
 * and fill in the private data structure.
 */
int scanslave(devp)
     register struct scsi_device *devp;
{
  char			vpid[VIDSZ+PIDSZ+1];
  int			status;
  struct scsi_pkt	*rsp, *get_pktiopb();

  DPRINTF(4)("scanslave: called\n");

  if (DUNIT >= nscans) {
    DPRINTF(1)("scanslave: DUNIT (%d) > nscans (%d)\n", DUNIT, nscans);
    return(0);
  }

  scanunits[DUNIT] = devp;	/* fill in our array */

  scanpri = MAX(scanpri, ipltospl(devp->sd_dev->devi_intr->int_pri));

  scsi_options &= ~(SCSI_OPTIONS_SYNC);

#if 0
  if (scsi_ifsetcap(&devp->sd_address, "synchronous", FALSE, 1) != 1)
    DPRINTF(1)("scanslave: scsi_ifsetcap failed\n");
#endif

  /*
   * scsi_slave() will send test unit ready and inquiry commands,
   * with reqest sense if necessary.
   */
  DPRINTF(4)("scanslave: calling scsi_slave\n");
  status = scsi_slave(devp, NULL_FUNC);
  switch(status) {
  case SCSIPROBE_EXISTS:
    /*
     * inquiry succeeded, devp->sd_inq is now filled in
     * XXX	Check inq_dtype, inq_vid, inq_pid and any other
     *	fields to make sure the target/unit is what's
     *	expected (sd_inq is a struct scsi_inquiry,
     *	defined in scsi/generic/inquiry.h).
     * XXX	Put device-specific checking into the appropriate
     *	case statement, and change the default: to
     *	an error message & return(0).
     */
    switch(devp->sd_inq->inq_dtype) {
    case DTYPE_PROCESSOR:
      if ((strncmp(devp->sd_inq->inq_pid, "C1750A", 6) == 0) ||
	  (strncmp(devp->sd_inq->inq_pid, "C2500A", 6) == 0))
	devp->sd_inq->inq_dtype = DTYPE_SCANNER;
      else
	goto inq_default;
      /* fall through if HP scanner */
    case DTYPE_SCANNER:
      DPRINTF(4)("%s%d: found %s device (tgt%d, lun%d) on %s%d\n",
		 DNAME, DUNIT, scsi_dname((int)devp->sd_inq->inq_dtype),
		 TGT(devp), LUN(devp), CNAME, CUNIT);
      break;
    default:
    inq_default:
      DPRINTF(1)("%s%d: target is not a scanner\n", DNAME, DUNIT);
      return(0);
    }
    break;
  case SCSIPROBE_NONCCS:    /* device doesn't support the inquiry command */
    /*
     * XXX If you have a non-ccs device, you must fill in the
     *     sd_inq structure manually
     */
    printf("%s%d: found non-CCS device (tgt%d, lun%d) on %s%d\n",
	   DNAME, DUNIT, TGT(devp), LUN(devp), CNAME, CUNIT);
    break;
  case SCSIPROBE_NOMEM:
  case SCSIPROBE_FAILURE:
  case SCSIPROBE_NORESP:
  default:
    DPRINTF(1)("scan_findslave: failed, scsi_slave returned 0x%x\n", status);
    devp->sd_present = 0;	/* no unit present */
    return(0);
    break;
  }

  /*
   * The expected device is indeed at the probed target id.
   * The rest of this routine fills in as much of the private
   * data structure as possible. The attach routine will
   * validate the unit (eg read a disk label).
   */

  /*
   * Allocate a request sense packet
   */
  if (!(rsp = get_pktiopb(&devp->sd_address, (caddr_t *)&devp->sd_sense,
			  CDB_GROUP0, 1, SENSE_LENGTH, B_READ, NULL_FUNC))) {
    return(0);
  }
  makecom_g0(rsp, devp, 0, SCMD_REQUEST_SENSE, 0, SENSE_LENGTH);
  rsp->pkt_pmon = -1;
  rsp->pkt_comp = scanintr;
  rsp->pkt_time = DFLT_SCAN_TIMEOUT;

  /*
   * Allocate and initialize private data
   */

  SCANPDS = (struct scsi_scan *)kmem_zalloc(sizeof(struct scsi_scan));
  if (!SCANPDS) {
    printf("%s%d: No memory for device structure\n", DNAME, DUNIT);
    free_pktiopb(rsp, (caddr_t)devp->sd_sense, SENSE_LENGTH);
    return(0);
  }
  SCANPDS->un_rbufp = (struct buf *)kmem_zalloc(sizeof (struct buf));
  SCANPDS->un_sbufp = (struct buf *)kmem_zalloc(sizeof (struct buf));
  if (!SCANPDS->un_rbufp || !SCANPDS->un_sbufp) {
    printf("%s%d: can't allocate raw/special buffer\n", DNAME, DUNIT);
    if (SCANPDS->un_rbufp) {
      (void) kmem_free((caddr_t)SCANPDS->un_rbufp,sizeof (struct buf));
    }
    free_pktiopb(rsp, (caddr_t)devp->sd_sense, SENSE_LENGTH);
    (void) kmem_free((caddr_t)(SCANPDS), sizeof (struct scsi_scan));
    return(0);
  }
  SCANPDS->un_tmpbuf = (caddr_t)kmem_zalloc(UN_TMPBUF_SIZE);
  if (!SCANPDS->un_tmpbuf) {
    printf("%s%d: can't allocate un_tmpbuf\n", DNAME, DUNIT);
    (void) kmem_free((caddr_t)SCANPDS->un_sbufp,sizeof (struct buf));
    (void) kmem_free((caddr_t)SCANPDS->un_rbufp,sizeof (struct buf));
    (void) kmem_free((caddr_t)(SCANPDS), sizeof (struct scsi_scan));
    free_pktiopb(rsp, (caddr_t)devp->sd_sense, SENSE_LENGTH);
    return(0);
  }
  /*
   * Fill in private data
   */
  SCANPDS->sio.scan_window_width  = 10200L;
  SCANPDS->sio.scan_window_length = 13200L;
  SCANPDS->sio.scan_x_resolution  = 200;
  SCANPDS->sio.scan_y_resolution  = 200;
  SCANPDS->sio.scan_x_origin      = 0;
  SCANPDS->sio.scan_y_origin      = 0;
  SCANPDS->sio.scan_image_mode    = (u_char)BINARY_MONOCHROME;
  SCANPDS->sio.scan_brightness    = (u_char)128;
  SCANPDS->sio.scan_contrast      = (u_char)128;
  SCANPDS->sio.scan_scanner_type  = model_to_type_code(devp->sd_inq->inq_pid);

  SCANPDS->un_rsp = rsp;
  SCANPDS->un_sd = devp;

  SCANPDS->adf_state = ADF_UNUSED;

  devp->sd_present = 1;	/* mark unit present */

  devp->sd_dev->devi_driver = &scan_ops;
  return(1);
}

/*
 * scanattach
 * Device-dependent initialization. Scanslave checks that the target
 * controller is there and responding. This routine initializes the
 * device attached to the controller (if needed). E.g for a removable
 * disk you might check that the disk you expect is there. If the
 * device can go away after boot (e,g, it's removable) you might
 * want to call scanattach from scanopen.
 */
int scanattach(devp)
     struct scsi_device *devp;
{
  char scan_id[29], *p, *q;
  int count;

  if (!SCANPDS) {
    DPRINTF(1)("scanattach: no private data struct\n");
    return;
  }
  DPRINTF(4)("scanattach: called\n");

  /*
   * print vendor, model & version info
   */

  p = devp->sd_inq->inq_vid;
  q = scan_id;
  count = 8;
  while (*p != ' ' && count--)
    *q++ = *p++;

  *q++ = ' ';
  p = devp->sd_inq->inq_pid;
  count = 16;
  while (*p != ' ' && count--)
    *q++ = *p++;

  *q++ = ' ';
  p = devp->sd_inq->inq_revision;
  count = 4;
  while (*p != ' ' && count--)
    *q++ = *p++;

  *q++ = '\0';

  printf("%s%d: <%s>\n", DNAME, DUNIT, scan_id);

  /*
   * If this is the first time through, mark closed.
   */
  if (SCANPDS->un_state == SCAN_STATE_NIL) {
    newstate(SCAN_STATE_CLOSED);
  }
  SCANPDS->un_attached = 1;
  return;
}

/*
 *********************************
 *
 * Un!x Entry Points
 *
 *********************************
 */


/*
 * open the device
 * Make sure the device is present and correct. Do any initialization that's
 * needed (start it up, load media, etc).
 */
int scanopen(dev, flag)
     dev_t	dev;
     int	flag;
{
  register struct scsi_device *devp;
  register int unit, s;
  int rc = 0;

  DPRINTF(4)("scanopen: SCANUNIT(dev) %d, nscans %d\n", SCANUNIT(dev), nscans);
  DPRINTF(4)("scanopen: scanunits[unit] %d\n", scanunits[unit]);

  if (((unit = SCANUNIT(dev)) >= nscans) || (!(devp = scanunits[unit]))) {
    DPRINTF(1)("scanopen: nodev, unit=%d, devp=0x%x\n", unit, devp);
    return(ENODEV);
  }
  DPRINTF(2)("scanopen: scan%d, devp=0x%x\n", unit, devp);

  /*
   * If the device isn't attached/online, attach it
   */
  if (!SCANPDS) {
    DPRINTF(2)("scanopen: no PDS\n");
    return(ENXIO);
  }
  if (!SCANPDS->un_attached) {
    DPRINTF(2)("scanopen: attach failed\n");
    return(ENXIO);
  }

  /*
   * If the device was closed, open and initialize it
   */

  if (SCANPDS->un_state == SCAN_STATE_CLOSED) {
    newstate(SCAN_STATE_OPENING);

    /* Test Unit Ready */
    while (rc = scansimple(devp, SCMD_TEST_UNIT_READY, 0)) {
      int tries = 5;

      if (--tries == 0)
	goto scanopen_exit;

      DPRINTF(2)("scanopen: re-sending TEST_UNIT_READY\n");
    }

    /* Reserve Unit */
    if (SCANPDS->sio.scan_scanner_type != HP_SCANJET_IIC)
      if (rc = scansimple(devp, SCMD_RESERVE, 0))
	goto scanopen_exit;
  } else {
    if (SCANPDS->un_state == SCAN_STATE_OPENING ||
	SCANPDS->un_state == SCAN_STATE_OPEN) {
      DPRINTF(2)("scanopen: device is already open\n");
      rc = EBUSY;
      goto scanopen_exit;
    } else {
      DPRINTF(3)("scanopen: SCANPDS->un_state %d\n", SCANPDS->un_state);
      rc = EIO;			/* XXX->what should errno be here? */
      goto scanopen_exit;
    }
  }
  SCANPDS->un_triggered = FALSE;
  newstate(SCAN_STATE_OPEN);

scanopen_exit:
  if (rc) {
    newstate(SCAN_STATE_CLOSED);
    DPRINTF(1)("scanopen: returning %d\n", rc);
  }
  return(rc);
}

static int do_object_position(devp, move_type)
     struct scsi_device *devp;
     int move_type;
{
  struct scsi_pkt *posp, *rqpkt = SCANPDS->un_rsp;
  int err;

  /*
   * Get a pkt with no dma resources
   */
  if (!(posp = scsi_pktalloc(&devp->sd_address, CDB_GROUP1, 1, SLEEP_FUNC))) {
    return(ENOMEM);
  }

  if (SCANPDS->sio.scan_scanner_type == RICOH_IS410 ||
      SCANPDS->sio.scan_scanner_type == IBM_2456    ||
      SCANPDS->sio.scan_scanner_type == RICOH_IS50)
    makecom_g1(posp, devp, FLAG_NOINTR, SCMD_OBJECT_POSITION, 0x0100 /* 1 page */, 0);
  else
    makecom_g1(posp, devp, FLAG_NOINTR, SCMD_OBJECT_POSITION, 0, 0);

  if (move_type == LOAD_ADF)
    posp->pkt_cdbp[1] = (u_char)1;

  if (scsi_poll(posp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_OBJECT_POSITION %s\n",
	       DNAME, DUNIT, (move_type == LOAD_ADF ? "load" : "unload"));
    scsi_resfree(posp);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(posp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(posp)->scc_cmd, scan_cmds));
    } else {
      if (scan_debug > 1)	/* XXX this should determine error type */
	scanerrmsg(devp, posp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(posp)->scc_cmd, scan_cmds),
	       SCBP_C(posp));
    err = EIO;
    break;
  }
  scsi_resfree(posp);
  return(err);
}

static void do_hp_unload_adf(devp)
     struct scsi_device *devp;
{
  char escape_codes[20];

  sprintf(escape_codes, "%c*u0U", '\033');

  hp_send_cmd(devp, escape_codes);
}

/*
 * close the device
 */
int scanclose(dev)
     dev_t	dev;
{
  register struct scsi_device *devp;
  int	unit;

  if (((unit = SCANUNIT(dev)) >= nscans) || (!(devp = scanunits[unit]))) {
    DPRINTF(1)("scanclose: nodev, unit=%d, devp=0x%x\n", unit, devp);
    return(ENODEV);
  }

  DPRINTF(2)("scanclose: scan%d, devp=0x%x\n", unit, devp);

  if ((!devp->sd_present) || (!SCANPDS) ||
      (SCANPDS->un_state < SCAN_STATE_OPEN)) {
    DPRINTF(1)("scanclose: sd_present=0x%x, SCANPDS=0x%x, state=%d\n",
	       devp->sd_present, SCANPDS, SCANPDS->un_state);
    return(ENXIO);
  }

  newstate(SCAN_STATE_CLOSED);

  if (SCANPDS->un_triggered) {

    /* if the HP ScanJet IIc was using the ADF as
     * the paper source, send an unload request
     */
    if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC && SCANPDS->using_adf)
      do_hp_unload_adf(devp);

    /*
     * if there is no paper in the UMAX's ADF send an unload to home
     * the carriage.
     *
     * if there is paper in the UMAX's ADF send an unload to eject
     * the paper.
     */
    if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	SCANPDS->sio.scan_scanner_type == UMAX_UG630) {
      do_object_position(devp, UNLOAD_ADF);
      find_umax(devp);

      /*
       * If there used to be paper in the ADF and now there's not then the
       * last sheet in the ADF just got ejected, so send an unload request
       * to home the carriage.
       */
      if (SCANPDS->using_adf && !umax_has_paper_in_adf(devp)) {
	do_object_position(devp, UNLOAD_ADF);
	find_umax(devp);
      }
    }
  }

  /* if the Ricoh ADF was being used eject the sheet */

  if ((SCANPDS->sio.scan_scanner_type == RICOH_IS410 ||
       SCANPDS->sio.scan_scanner_type == IBM_2456    ||
       SCANPDS->sio.scan_scanner_type == RICOH_IS50)
      && SCANPDS->adf_state == ADF_CLEANUP) {

    do_object_position(devp, UNLOAD_ADF);
    SCANPDS->adf_state = ADF_UNUSED;
  }

  /* if the Sharp has not read all of the data in the active area
   * issue a request to stop scanning
   */

  /****  do a SCSI release unit on the scanner	***/
  if (SCANPDS->sio.scan_scanner_type != HP_SCANJET_IIC)
    scansimple(devp, SCMD_RELEASE, 0);

  return(0);
}

/*
 * Raw i/o operations
 */

/*
 * Allow a maximum of SCAN_MAXREC bytes per transfer
 */
static void scanminphys(bp)
     struct buf	*bp;
{
}


static int do_set_window(devp)
     struct scsi_device *devp;
{
  switch (SCANPDS->sio.scan_scanner_type) {
  case UMAX_UC630:
  case UMAX_UG630:
    return (do_umax_set_window(devp));

  case FUJITSU:
    return (do_fujitsu_set_window(devp));

  case HP_SCANJET_IIC:
    return (do_hp_set_window(devp));

  case RICOH_IS410:
  case IBM_2456:
    return (do_ricoh_is410_set_window(devp));

  default:
    return (EINVAL);
  }
}

#define RICOH_IS410_WD_LEN 72	/* 8+64 */

static int do_ricoh_is410_set_window(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *swp, *rqpkt = SCANPDS->un_rsp;
  unsigned char *wd;
  int err;

  if (!(swp = get_pktiopb(devp->sd_address, (caddr_t)&wd, CDB_GROUP1, 1,
			  RICOH_IS410_WD_LEN, B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  /*
   * fill out header & window descriptor block
   */
  bzero(wd, RICOH_IS410_WD_LEN);	/* all fields default to zero */

  /* header... */
  wd[6] = 0x01;			/* window descriptor block   */
  wd[7] = 0x40;			/* must be 0x140 (320) bytes */

  /* window descriptor block... */

  /* ...resolution... */
  wd[8+2] = (uchar)(SCANPDS->sio.scan_x_resolution >> 8 & 0xff);
  wd[8+3] = (uchar)(SCANPDS->sio.scan_x_resolution & 0xff);

  wd[8+4] = (uchar)(SCANPDS->sio.scan_y_resolution >> 8 & 0xff);
  wd[8+5] = (uchar)(SCANPDS->sio.scan_y_resolution & 0xff);

  /* ...X origin... */
  wd[8+6] = (uchar)(SCANPDS->sio.scan_x_origin >> 24 & 0xff);
  wd[8+7] = (uchar)(SCANPDS->sio.scan_x_origin >> 16 & 0xff);
  wd[8+8] = (uchar)(SCANPDS->sio.scan_x_origin >>  8 & 0xff);
  wd[8+9] = (uchar)(SCANPDS->sio.scan_x_origin & 0xff);

  /* ...Y origin... */
  wd[8+10] = (uchar)(SCANPDS->sio.scan_y_origin >> 24 & 0xff);
  wd[8+11] = (uchar)(SCANPDS->sio.scan_y_origin >> 16 & 0xff);
  wd[8+12] = (uchar)(SCANPDS->sio.scan_y_origin >>  8 & 0xff);
  wd[8+13] = (uchar)(SCANPDS->sio.scan_y_origin & 0xff);

  /* ...width... */
  wd[8+14] = (uchar)(SCANPDS->sio.scan_window_width >> 24 & 0xff);
  wd[8+15] = (uchar)(SCANPDS->sio.scan_window_width >> 16 & 0xff);
  wd[8+16] = (uchar)(SCANPDS->sio.scan_window_width >>  8 & 0xff);
  wd[8+17] = (uchar)(SCANPDS->sio.scan_window_width & 0xff);

  /* ...length... */
  wd[8+18] = (uchar)(SCANPDS->sio.scan_window_length >> 24 & 0xff);
  wd[8+19] = (uchar)(SCANPDS->sio.scan_window_length >> 16 & 0xff);
  wd[8+20] = (uchar)(SCANPDS->sio.scan_window_length >>  8 & 0xff);
  wd[8+21] = (uchar)(SCANPDS->sio.scan_window_length & 0xff);

  wd[8+22] = SCANPDS->sio.scan_brightness; /* brightness */
  wd[8+23] = SCANPDS->sio.scan_brightness; /* threshold */
  wd[8+24] = SCANPDS->sio.scan_contrast;

  wd[8+25] = SCANPDS->sio.scan_image_mode;

  if (SCANPDS->sio.scan_image_mode < GRAYSCALE)
    wd[8+26] = 1;		/* bits-per-pixel */
  else
    wd[8+26] = 8;		/* bits-per-pixel */

  wd[8+27] = 2;			/* halftone code to turn on dithering */
  wd[8+28] = 0x0a;		/* 8x8 bayer pattern */

  wd[8+29] = 3;			/* padding type, truncate to byte boundary */

  wd[8+31] = 7;			/* bit ordering LSB, (set to big-endian) */

  wd[8+42] = 0x80;		/* MRIF bit set on, grayscale will be min-is-black */

  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g1(swp, devp, FLAG_NOINTR, SCMD_SET_WINDOW, 0, RICOH_IS410_WD_LEN);

  if (scsi_poll(swp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_SET_WINDOW\n", DNAME, DUNIT);
    free_pktiopb(swp, wd, RICOH_IS410_WD_LEN);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(swp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n",
		 DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, swp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds),
	       SCBP_C(swp));
    err = EIO;
    break;
  }
  free_pktiopb(swp, wd, RICOH_IS410_WD_LEN);
  return(err);
}

#define UMAX_WD_LEN 54

static int do_umax_set_window(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *swp, *rqpkt = SCANPDS->un_rsp;
  unsigned char *wd;
  int err;

  if (!(swp = get_pktiopb(devp->sd_address, (caddr_t)&wd,
			  CDB_GROUP1, 1, UMAX_WD_LEN, B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  /*
   * fill out header & window descriptor block
   */
  bzero(wd, UMAX_WD_LEN);		/* all fields default to zero */

  /* header... */
  wd[7] = 0x2e;			/* window descriptor block len */

  /* window descriptor block... */

  /* ...resolution... */
  wd[8+2] = (uchar)(SCANPDS->sio.scan_x_resolution >> 8 & 0xff);
  wd[8+3] = (uchar)(SCANPDS->sio.scan_x_resolution & 0xff);
  wd[8+4] = (uchar)(SCANPDS->sio.scan_y_resolution >> 8 & 0xff);
  wd[8+5] = (uchar)(SCANPDS->sio.scan_y_resolution & 0xff);

  /* ...X origin... */
  wd[8+0x6] = (uchar)(SCANPDS->sio.scan_x_origin >> 24 & 0xff);
  wd[8+0x7] = (uchar)(SCANPDS->sio.scan_x_origin >> 16 & 0xff);
  wd[8+0x8] = (uchar)(SCANPDS->sio.scan_x_origin >>  8 & 0xff);
  wd[8+0x9] = (uchar)(SCANPDS->sio.scan_x_origin & 0xff);

  /* ...Y origin... */
  wd[8+0xA] = (uchar)(SCANPDS->sio.scan_y_origin >> 24 & 0xff);
  wd[8+0xB] = (uchar)(SCANPDS->sio.scan_y_origin >> 16 & 0xff);
  wd[8+0xC] = (uchar)(SCANPDS->sio.scan_y_origin >>  8 & 0xff);
  wd[8+0xD] = (uchar)(SCANPDS->sio.scan_y_origin & 0xff);

  /* ...width... */
  wd[8+0x0E] = (uchar)(SCANPDS->sio.scan_window_width >> 24 & 0xff);
  wd[8+0x0F] = (uchar)(SCANPDS->sio.scan_window_width >> 16 & 0xff);
  wd[8+0x10] = (uchar)(SCANPDS->sio.scan_window_width >>  8 & 0xff);
  wd[8+0x11] = (uchar)(SCANPDS->sio.scan_window_width & 0xff);

  /* ...length... */
  wd[8+0x12] = (uchar)(SCANPDS->sio.scan_window_length >> 24 & 0xff);
  wd[8+0x13] = (uchar)(SCANPDS->sio.scan_window_length >> 16 & 0xff);
  wd[8+0x14] = (uchar)(SCANPDS->sio.scan_window_length >>  8 & 0xff);
  wd[8+0x15] = (uchar)(SCANPDS->sio.scan_window_length & 0xff);

  wd[8+0x16] = SCANPDS->sio.scan_brightness; /* brightness */
  wd[8+0x17] = SCANPDS->sio.scan_brightness; /* threshold */
  wd[8+0x18] = SCANPDS->sio.scan_contrast;

  switch (SCANPDS->sio.scan_image_mode) {
  case RED:
  case BLUE:
  case GREEN:
    wd[8+0x19] = GRAYSCALE;
    break;

  default:
    wd[8+0x19] = SCANPDS->sio.scan_image_mode;
    break;
  }

  if (SCANPDS->sio.scan_image_mode < GRAYSCALE)
    wd[8+0x1A] = 1;		/* bits-per-pixel */
  else
    wd[8+0x1A] = 8;		/* bits-per-pixel */

  wd[8+0x1C] = 1;		/* halftone pattern */

  wd[8+0x1D] = 3;		/* padding type */

  wd[8+0x28] = 1;		/* speed: fastest speed that doesn't smear */

  switch (SCANPDS->sio.scan_image_mode) {
  case BINARY_MONOCHROME:
  case DITHERED_MONOCHROME:
  case GRAYSCALE:
    wd[8+0x29] = (uchar)0;
    break;
  case RED:
    wd[8+0x29] = (uchar)0x80;
    break;
  case GREEN:
    wd[8+0x29] = (uchar)0x40;
    break;
  case BLUE:
    wd[8+0x29] = (uchar)0x20;
    break;
  }

  wd[8+0x2A] = 50;		/* highlight */

  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g1(swp, devp, FLAG_NOINTR, SCMD_SET_WINDOW, 0, UMAX_WD_LEN);

  if (scsi_poll(swp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_SET_WINDOW\n",
	       DNAME, DUNIT);
    free_pktiopb(swp, wd, UMAX_WD_LEN);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(swp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n",
		 DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, swp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds),
	       SCBP_C(swp));
    err = EIO;
    break;
  }
  free_pktiopb(swp, wd, UMAX_WD_LEN);
  return(err);
}

#define FUJITSU_WD_LEN 48

static int do_fujitsu_set_window(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *swp, *rqpkt = SCANPDS->un_rsp;
  unsigned char *wd;
  int err;

  if (!(swp = get_pktiopb(devp->sd_address, (caddr_t)&wd, CDB_GROUP1,
			  1, FUJITSU_WD_LEN, B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  /*
   * fill out header & window descriptor block
   */
  bzero(wd, FUJITSU_WD_LEN);		/* all fields default to zero */

  /* header... */
  wd[7] = (uchar)40;			/* window descriptor block len */

  /* window descriptor block... */

  /* ...resolution... */
  wd[8+2] = (uchar)(SCANPDS->sio.scan_x_resolution >> 8 & 0xff);
  wd[8+3] = (uchar)(SCANPDS->sio.scan_x_resolution & 0xff);
  wd[8+4] = (uchar)(SCANPDS->sio.scan_y_resolution >> 8 & 0xff);
  wd[8+5] = (uchar)(SCANPDS->sio.scan_y_resolution & 0xff);

  /* ...X origin... */
  wd[8+0x6] = (uchar)(SCANPDS->sio.scan_x_origin >> 24 & 0xff);
  wd[8+0x7] = (uchar)(SCANPDS->sio.scan_x_origin >> 16 & 0xff);
  wd[8+0x8] = (uchar)(SCANPDS->sio.scan_x_origin >>  8 & 0xff);
  wd[8+0x9] = (uchar)(SCANPDS->sio.scan_x_origin & 0xff);

  /* ...Y origin... */
  wd[8+0xA] = (uchar)(SCANPDS->sio.scan_y_origin >> 24 & 0xff);
  wd[8+0xB] = (uchar)(SCANPDS->sio.scan_y_origin >> 16 & 0xff);
  wd[8+0xC] = (uchar)(SCANPDS->sio.scan_y_origin >>  8 & 0xff);
  wd[8+0xD] = (uchar)(SCANPDS->sio.scan_y_origin & 0xff);

  /* ...width... */
  wd[8+0x0E] = (uchar)(SCANPDS->sio.scan_window_width >> 24 & 0xff);
  wd[8+0x0F] = (uchar)(SCANPDS->sio.scan_window_width >> 16 & 0xff);
  wd[8+0x10] = (uchar)(SCANPDS->sio.scan_window_width >>  8 & 0xff);
  wd[8+0x11] = (uchar)(SCANPDS->sio.scan_window_width & 0xff);

  /* ...length... */
  wd[8+0x12] = (uchar)(SCANPDS->sio.scan_window_length >> 24 & 0xff);
  wd[8+0x13] = (uchar)(SCANPDS->sio.scan_window_length >> 16 & 0xff);
  wd[8+0x14] = (uchar)(SCANPDS->sio.scan_window_length >>  8 & 0xff);
  wd[8+0x15] = (uchar)(SCANPDS->sio.scan_window_length & 0xff);

#if 0
  wd[8+0x16] = 0;  /*SCANPDS->sio.scan_brightness; /* brightness */
  wd[8+0x17] = SCANPDS->sio.scan_brightness; /* threshold */
  wd[8+0x18] = SCANPDS->sio.scan_contrast;
#endif

  wd[8+0x19] = SCANPDS->sio.scan_image_mode;

  if (SCANPDS->sio.scan_image_mode < GRAYSCALE)
    wd[8+0x1A] = 1;		/* bits-per-pixel */
  else
    wd[8+0x1A] = 8;		/* bits-per-pixel */

  wd[8+0x1C] = 1;		/* halftone pattern */

  wd[8+0x1D] = 0;		/* padding type */

  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g1(swp, devp, FLAG_NOINTR, SCMD_SET_WINDOW, 0, FUJITSU_WD_LEN);

  if (scsi_poll(swp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_SET_WINDOW\n",
	       DNAME, DUNIT);
    free_pktiopb(swp, wd, FUJITSU_WD_LEN);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(swp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n",
		 DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, swp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(swp)->scc_cmd, scan_cmds),
	       SCBP_C(swp));
    err = EIO;
    break;
  }
  free_pktiopb(swp, wd, FUJITSU_WD_LEN);
  return(err);
}


static int hp_send_cmd(devp, escape_codes)
     struct scsi_device *devp;
     char *escape_codes;
{
  struct scsi_pkt *spkt, *rqpkt = SCANPDS->un_rsp;
  int code_count, copied;
  caddr_t datap;
  int err;

  code_count = strlen(escape_codes);

  if (scan_debug) {
    printf("hp_send_cmd: %d escape_codes:\n", code_count);
    {
      int i = 0;
      while (escape_codes[i]) {
	if (escape_codes[i] == 033)
	  printf("[esc]");
	else
	  printf("%c", escape_codes[i]);
	++i;
      }
      printf("\n");
    }
  }

  if (!(spkt = get_pktiopb(&devp->sd_address, &datap,
			   CDB_GROUP0, 1, code_count,
			   B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  makecom_g0_s(spkt, devp, FLAG_NOINTR|FLAG_NODISCON, SCMD_WRITE, code_count, 0);

  copystr(escape_codes, datap, code_count, &copied);

  if (scsi_poll(spkt)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_WRITE\n", DNAME, DUNIT);
    goto hp_send_cmd_out;
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(spkt)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(spkt)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, spkt, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(spkt)->scc_cmd, scan_cmds),
	       SCBP_C(spkt));
    err = EIO;
    break;
  }
 hp_send_cmd_out:
  free_pktiopb(spkt, datap, code_count);
  return(err);
}

static int hp_recv_response(devp, escape_codes, code_count)
     struct scsi_device *devp;
     char *escape_codes;
     int code_count;
{
  struct scsi_pkt *rdpkt, *rqpkt = SCANPDS->un_rsp;
  int copied;
  caddr_t datap;
  int err;

  if (!(rdpkt = get_pktiopb(&devp->sd_address, &datap,
			   CDB_GROUP0, 1, code_count,
			   B_READ, SLEEP_FUNC))) {
    return (EIO);
  }

  makecom_g0_s(rdpkt, devp, FLAG_NOINTR|FLAG_NODISCON, SCMD_READ, code_count, 0);

  if (scsi_poll(rdpkt)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_READ\n", DNAME, DUNIT);
    goto hp_send_cmd_out;
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(rdpkt)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    copystr(datap, escape_codes, code_count, &copied);
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(rdpkt)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, rdpkt, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(rdpkt)->scc_cmd, scan_cmds),
	       SCBP_C(rdpkt));
    err = EIO;
    break;
  }

 hp_send_cmd_out:
  free_pktiopb(rdpkt, datap, code_count);

  return(err);
}

static int do_get_buffer_status(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *rdpkt, *rqpkt = SCANPDS->un_rsp;
  int copied;
  caddr_t datap;
  int err;

  if (!(rdpkt = get_pktiopb(&devp->sd_address, &datap,
			   CDB_GROUP1, 1, 12, B_READ, SLEEP_FUNC))) {
    return (EIO);
  }

  makecom_g1(rdpkt, devp, FLAG_NOINTR, SCMD_GET_BUFFER_STATUS, 0, 12);

  if (scsi_poll(rdpkt)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_GET_BUFFER_STATUS\n",
		DNAME, DUNIT);
    goto do_get_buffer_status_out;
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(rdpkt)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    copystr(&datap[9], SCANPDS->buf_status, 3, &copied);
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(rdpkt)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, rdpkt, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(rdpkt)->scc_cmd, scan_cmds),
	       SCBP_C(rdpkt));
    err = EIO;
    break;
  }

do_get_buffer_status_out:
  free_pktiopb(rdpkt, datap, 12);

  return(err);
}

static int do_hp_set_window(devp)
     struct scsi_device *devp;
{
  static char escape_codes[128];
  char tmp[15];
  int copied;

  /*
   * deal with the fact that the HP ScanJet IIc uses 1/300" not 1/1200"
   * as its base unit of measurement
   */

  if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC) {
    SCANPDS->sio.scan_window_length /= 4;
    SCANPDS->sio.scan_window_length *= 4;

    SCANPDS->sio.scan_window_width /= 4;
    SCANPDS->sio.scan_window_width *= 4;
  }

  sprintf(escape_codes, "%c*f%dP", '\033', SCANPDS->sio.scan_window_width / 4);

  sprintf(tmp, "%c*f%dQ", '\033', SCANPDS->sio.scan_window_length / 4);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  sprintf(tmp, "%c*f%dX", '\033', SCANPDS->sio.scan_x_origin / 4);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  sprintf(tmp, "%c*f%dY", '\033', SCANPDS->sio.scan_y_origin / 4);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  sprintf(tmp, "%c*a%dR", '\033', SCANPDS->sio.scan_x_resolution);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  sprintf(tmp, "%c*a%dS", '\033', SCANPDS->sio.scan_y_resolution);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  switch (SCANPDS->sio.scan_image_mode) {
  case BINARY_MONOCHROME:
    sprintf(tmp, "%c*a%dT", '\033', 0);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    sprintf(tmp, "%c*a%dG", '\033', 1);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* make image data be "min-is-white ala PBM */
    sprintf(tmp, "%c*a%dI", '\033', 0);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    break;
  case DITHERED_MONOCHROME:
    sprintf(tmp, "%c*a%dT", '\033', 3);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    sprintf(tmp, "%c*a%dG", '\033', 1);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* make image data be "min-is-white ala PBM */
    sprintf(tmp, "%c*a%dI", '\033', 0);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    break;
  case GRAYSCALE:
    sprintf(tmp, "%c*a%dT", '\033', 4);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    sprintf(tmp, "%c*a%dG", '\033', 8);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* make image data be "min-is-black ala PGM */
    sprintf(tmp, "%c*a%dI", '\033', 1);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    break;
  case COLOR:
    sprintf(tmp, "%c*a%dT", '\033', 5);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* use 24-bit data */
    sprintf(tmp, "%c*a%dG", '\033', 24);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* make image data be "min-is-black ala PPM */
    sprintf(tmp, "%c*a%dI", '\033', 1);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
    /* use pass-through matrix (disable NTSC) */
    sprintf(tmp, "%c*u%dT", '\033', 2);
    copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
  }

  sprintf(tmp, "%c*a%dL", '\033', (int)(SCANPDS->sio.scan_brightness) - 128);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);
     
  sprintf(tmp, "%c*a%dK", '\033', (int)(SCANPDS->sio.scan_contrast) - 128);
  copystr(tmp, &escape_codes[strlen(escape_codes)], strlen(tmp) + 1, &copied);

  return (hp_send_cmd(devp, escape_codes));
}

static int do_umax_trigger(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *trgp, *rqpkt = SCANPDS->un_rsp;
  caddr_t window_id;
  int err;

  if (!(trgp = get_pktiopb(devp->sd_address, &window_id,
			   CDB_GROUP0, 1, 1, B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  *window_id = '\0';

  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g0_s(trgp, devp, FLAG_NOINTR, SCMD_TRIGGER_SCAN, 1, 0);

  if (scsi_poll(trgp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_TRIGGER_SCAN\n",
		DNAME, DUNIT);
    free_pktiopb(trgp, window_id, 1);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(trgp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(trgp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, trgp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(trgp)->scc_cmd, scan_cmds),
	       SCBP_C(trgp));
    err = EIO;
    break;
  }
  free_pktiopb(trgp, window_id, 1);
  return(err);
}

#define MODE_SELECT_LEN sizeof(struct ricoh_mode_params)

static int do_mode_select(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt *msp, *rqpkt = SCANPDS->un_rsp;
  struct ricoh_mode_params *rmp;
  int err;

  if (!(msp = get_pktiopb(devp->sd_address, (caddr_t)&rmp,
			  CDB_GROUP0, 1, MODE_SELECT_LEN, B_WRITE, SLEEP_FUNC))) {
    return (EIO);
  }

  /* mode select parameters... */
  bzero(rmp, MODE_SELECT_LEN);

  /* measurement page descriptor: */

  rmp->page_code = SCANNING_MEASUREMENT_PARAMETERS;
  rmp->paramter_length = (uchar)6;
  rmp->base_measurement_unit = INCHES;
  rmp->measurement_unit_devisor_msb = (uchar)((1200 >> 8) & 0xff);
  rmp->measurement_unit_devisor_lsb = (uchar)((1200 & 0xff));

  /* ADF support page descriptor */

  rmp->adf_page_code = ADF_CONTROL;
  rmp->adf_parameter_length = (uchar)6;
  if (SCANPDS->adf_state == ADF_ARMED) {
    DPRINTF(3)("do_mode_select: ADF is on\n");
    rmp->adf_control = '\01';
  } else {
    DPRINTF(3)("do_mode_select: ADF is off\n");
    rmp->adf_control = '\0';
  }

  /* build CDB */

  makecom_g0_s(msp, devp, FLAG_NOINTR, SCMD_MODE_SELECT, MODE_SELECT_LEN, 0x10);

  if (scsi_poll(msp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_MODE_SELECT\n",
		DNAME, DUNIT);
    free_pktiopb(msp, rmp, MODE_SELECT_LEN);
    return (EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(msp)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n", DNAME, DUNIT,
		 scsi_cmd_decode(CDBP(msp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, msp, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(msp)->scc_cmd, scan_cmds),
	       SCBP_C(msp));
    err = EIO;
    break;
  }
  free_pktiopb(msp, rmp, MODE_SELECT_LEN);
  return(err);
}

static int do_hp_trigger(devp)
     struct scsi_device *devp;
{
  char escape_codes[20];

  sprintf(escape_codes, "%c*f0S", '\033');

  return (hp_send_cmd(devp, escape_codes));
}

static int do_trigger(devp)
     struct scsi_device *devp;
{
  switch (SCANPDS->sio.scan_scanner_type) {
  case UMAX_UC630:
  case UMAX_UG630:
  case RICOH_IS410:
  case IBM_2456:
    return (do_umax_trigger(devp));

  case HP_SCANJET_IIC:
    return (do_hp_trigger(devp));

  default:
    return (EINVAL);
  }
}

int scanread(dev, uio)
      dev_t	  dev;
      struct uio *uio;
{
  register struct scsi_device *devp;
  register int unit, rc;

  if (((unit = SCANUNIT(dev)) >= nscans) || (!(devp = scanunits[unit]))) {
    DPRINTF(1)("scanread: nodev, unit=%d, devp=0x%x\n", unit, devp);
    return(ENODEV);
  }
  DPRINTF(3)("scanread: scan%d, devp=0x%x\n", unit, devp);


  if (!SCANPDS->un_triggered) {

    /* mode select command */

    if (SCANPDS->sio.scan_scanner_type == RICOH_IS410 ||
	SCANPDS->sio.scan_scanner_type == IBM_2456) {
      DPRINTF(4)("scanread: sending SCMD_MODE_SELECT\n");
      if (rc = do_mode_select(devp)) {
	DPRINTF(2)("scanread: SCMD_MODE_SELECT failed\n");
	goto scanread_exit;
      }
      DPRINTF(4)("scanread: SCMD_MODE_SELECT ok\n");

      /* remember to send an "unload" on scanclose() */

      if (SCANPDS->adf_state == ADF_ARMED)
	SCANPDS->adf_state = ADF_CLEANUP;
    }

    /* set window command */

    DPRINTF(4)("scanread: sending SCMD_SET_WINDOW\n");
    if (rc = do_set_window(devp)) {
      DPRINTF(2)("scanread: SCMD_SET_WINDOW failed\n");
      goto scanread_exit;
    }
    DPRINTF(4)("scanread: SCMD_SET_WINDOW ok\n");

    /* remember if HP is using ADF so that close routine will eject paper */

    if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC)
      SCANPDS->using_adf = hp_has_paper_in_adf(devp);

    /* read trigger command */

    if (SCANPDS->sio.scan_scanner_type != FUJITSU) {
      DPRINTF(4)("scanread: sending SCMD_TRIGGER_SCAN\n");
      if (rc = do_trigger(devp)) {
	DPRINTF(2)("scanread: SCMD_TRIGGER_SCAN failed\n");
	goto scanread_exit;
      }

      /* deal with the UMAX sending "command complete" too soon */

      if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	  SCANPDS->sio.scan_scanner_type == UMAX_UG630)
	find_umax(devp);

      DPRINTF(4)("scanread: SCMD_TRIGGER_SCAN ok\n");
    }

    /* calculate the window size */

    SCANPDS->window_size = calc_window_size(devp);

    /* remember if UMAX is using ADF so that close routine will eject paper */

    if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	SCANPDS->sio.scan_scanner_type == UMAX_UG630)
      SCANPDS->using_adf = umax_has_paper_in_adf(devp);

    SCANPDS->un_triggered = TRUE;
  }

  DPRINTF(3)("scanread: window size is %d\n", SCANPDS->window_size);

  /* *********************************************** */
  /* trim the transfer down to Sun's measly 63K DVMA */
  /* *********************************************** */

  if (uio->uio_iov->iov_len > SCAN_MAXREC)
    uio->uio_iov->iov_len = SCAN_MAXREC;

  /* ***************************************** */
  /* trim the transfer down to the window size */
  /* ***************************************** */

  if (SCANPDS->sio.scan_scanner_type != FUJITSU)
    if (uio->uio_iov->iov_len > SCANPDS->window_size) {
      uio->uio_iov->iov_len = SCANPDS->window_size;
      DPRINTF(3)("scanread: bytes wanted exceeds window size\n");
      DPRINTF(3)("scanread: xfer reduced to %d\n", uio->uio_iov->iov_len);
    }

  /* ************************************************** */
  /* trim UMAX reads down to an even scan line boundary */
  /* ************************************************** */

  if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
      SCANPDS->sio.scan_scanner_type == UMAX_UG630) {
    int row_len;

    if (SCANPDS->sio.scan_image_mode >= GRAYSCALE)
      row_len = calc_umax_row_len(SCANPDS->sio.scan_x_resolution,
				  SCANPDS->sio.scan_window_width);
    else
      row_len =
	SCANPDS->sio.scan_window_width * SCANPDS->sio.scan_x_resolution / 9600;

    if (uio->uio_iov->iov_len < row_len) {
      uprintf("%s%d: read requests must be at least one row of image data\n",
	      DNAME, DUNIT);
      uio->uio_iov->iov_len = 0;
      return;
    }

    if (uio->uio_iov->iov_len % row_len) {
      uio->uio_iov->iov_len -= uio->uio_iov->iov_len % row_len;
      DPRINTF(3)("scanread: a row is %d bytes\n", row_len);
      DPRINTF(3)("scanread: transfer reduced to %d bytes (%d rows)\n",
		 uio->uio_iov->iov_len, uio->uio_iov->iov_len / row_len);
    }
  }

  /* *************************************** */
  /* trim Ricoh reads down to available data */
  /* *************************************** */

  if ((SCANPDS->sio.scan_scanner_type == RICOH_IS410 ||
       SCANPDS->sio.scan_scanner_type == IBM_2456)
      && SCANPDS->window_size) {
    int bytes_ready;

    do {
      do_get_buffer_status(devp);
      bytes_ready =
	(SCANPDS->buf_status[0] * 65536) +
	  (SCANPDS->buf_status[1] * 256) +
	    (SCANPDS->buf_status[2]);
      DPRINTF(4)("scanread: %d bytes ready\n", bytes_ready);
    } while (bytes_ready == 0);
      
    if (uio->uio_iov->iov_len > bytes_ready) {
      uio->uio_iov->iov_len = bytes_ready;
      DPRINTF(3)("scanread: bytes wanted exceeds bytes ready\n");
      DPRINTF(3)("scanread: xfer reduced to %d\n", uio->uio_iov->iov_len);
    }
  }

  /* ************* */
  /* EOF detection */
  /* ************* */

  if (SCANPDS->window_size <= 0) {
    uio->uio_iov++;
    uio->uio_iovcnt--;
    rc = 0;
    goto scanread_exit;
  }


  rc = scanrw(dev, uio, B_READ);

scanread_exit:
  return(rc);
}

/*
 * Raw read/write routine
 * Check uio values, call scanstrategy via physio for the actual transfer
 */
static int scanrw(dev, uio, flag)
     dev_t dev;
     struct uio *uio;
{
  struct scsi_device	*devp;
  register int		unit;

  if (((unit = SCANUNIT(dev)) >= nscans) || (!(devp = scanunits[unit]))) {
    DPRINTF(1)("scanrw: nodev, unit=%d, devp=0x%x\n", unit, devp);
    return(ENODEV);
  }
  DPRINTF(2)("scanrw: (%s) scan%d, devp=0x%x\n",
	     (flag==B_READ) ? "read" : "write", unit, devp);

  if ((!devp->sd_present) || (!SCANPDS)
  || (SCANPDS->un_state < SCAN_STATE_OPEN)) {
    DPRINTF(1)("scanrw: sd_present=0x%x, SCANPDS=0x%x, state=%d\n",
	       devp->sd_present, SCANPDS, SCANPDS->un_state);
    return(ENXIO);
  }

#if 0
  /*
   * Check that block offset and length are modulo the block size
   */
  if ((uio->uio_offset & (DEV_BSIZE-1)) ||
      (uio->uio_iov->iov_len & (DEV_BSIZE-1))) {
    DPRINTF(1)("scanrw: invalid block length=%d, offset=%d\n",
	       uio->uio_iov->iov_len, uio->uio_offset);
    return(EINVAL);
  }
#endif

  DPRINTF(3)("scanrw: offset = %d, length = %d\n",
	     uio->uio_offset, uio->uio_iov->iov_len);

  return(physio(scanstrategy, SCANPDS->un_rbufp, dev, flag, scanminphys, uio));
}

/*
 * scanstrategy
 * Main routine for commands to the device.
 * Put the buf on the pending queue and kick the start routine. The
 * start routine (scanstart) will allocate scsi command and any dma
 * resources needed, and send the command off to the host adapter controller.
 */
int scanstrategy(bp)
     register struct buf *bp;
{
  register struct scsi_device *devp = scanunits[SCANUNIT(bp->b_dev)];
  register int s;
  struct buf	*ap;

  DPRINTF(2)("scanstrategy:\n");

  /*
   * Check that the unit's really present
   */
  if (!devp || !devp->sd_present || !SCANPDS) {
    DPRINTF(1)("scanstrategy: error, unit not present\n");
    bp->b_flags = B_ERROR;
    bp->b_resid = bp->b_bcount;
    bp->b_error = ENXIO;
    iodone(bp);
    return(ENXIO);
  }

  s = splr(scanpri);

  /*
   * Put the buf on the queue
   */
  if (SCANPDS->un_quehd) {
    SCANPDS->un_quetl->av_forw = bp;
  } else {
    SCANPDS->un_quehd = bp;
  }
  SCANPDS->un_quetl = bp;
  SCANPDS->un_bufcnt++;

  bp->av_forw = 0;
  bp->b_flags &= ~(B_DONE|B_ERROR);
  bp->b_resid = 0;

  /*
   * If it's not a special command, scanstart must allocate the packet,
   * so zero bp->av_back to warn scanstart.
   */
  if (bp != SCANPDS->un_sbufp) {
    BP_PKT(bp) = 0;
  }

  /*
   * kick the start routine if it's inactive
   */
  if (SCANPDS->un_active == NULL)
    scanstart(devp);
  (void)splx(s);

  return(0);
}

/*
 * ioctl calls.
 * Called from the device switch table
 * Ioctls are device specific. Only a generic ioctl is supplied. This
 * serves as an example, and allows a user program to send an arbitrary
 * SCSI command to the target.
 */
int scanioctl(dev, cmd, data, flag)
     dev_t dev;
     register int cmd;
     register caddr_t data;
     int flag;
{
  register struct scsi_device *devp = scanunits[SCANUNIT(dev)];
  struct scan_io *s_io;
  struct scan_pixel_sizes *pix_size;
  int rc;

  if (!devp || !devp->sd_present || !SCANPDS) {
    DPRINTF(1)("scanioctl: %s%d: target not present\n", DNAME, DUNIT);
    return(ENXIO);
  }

  DPRINTF(2)("scanioctl: %s%d, devp=0x%x\n", DNAME, DUNIT, devp);

  switch(cmd) {
  case SCANIOC_GETERRC:		/* Get detailed error code */
    bcopy((caddr_t)devp->sd_sense, data, sizeof(struct scsi_extended_sense));
    break;

  case SCANIOC_CMD:		/* run a generic command */
    return(scancmd(dev, SCMD_UCMD, data));
    break;

  case SCAN_GET:
    bcopy(&(SCANPDS->sio), data, sizeof(struct scan_io));
    return (0);

  case SCAN_SET:
    s_io = (struct scan_io *)data;

    /****  Validate parameters  ***/
    switch (SCANPDS->sio.scan_scanner_type) {

    case HP_SCANJET_IIC:
      
      /* size constraints... */

      if (s_io->scan_window_width <= 0				||
	  s_io->scan_x_origin + s_io->scan_window_width > 10200	|| /* 8.5" */
	  s_io->scan_window_length <= 0				||
	  s_io->scan_y_origin + s_io->scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io->scan_x_resolution < 100 || s_io->scan_x_resolution > 400 ||
	  s_io->scan_y_resolution < 100 || s_io->scan_y_resolution > 400)
	return (EINVAL);

      switch (s_io->scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
      case COLOR:
	break;
      default:
	return (EINVAL);
      }
      break;

    case UMAX_UG630:
    case UMAX_UC630:
      
      /* size constraints... */

      if (s_io->scan_window_width <= 0				||
	  s_io->scan_x_origin + s_io->scan_window_width > 10200	|| /* 8.5" */
	  s_io->scan_window_length <= 0				||
	  s_io->scan_y_origin + s_io->scan_window_length > 16800)   /* 14" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io->scan_x_resolution < 100 || s_io->scan_x_resolution > 300 ||
	  s_io->scan_y_resolution < 100 || s_io->scan_y_resolution > 600)
	return (EINVAL);

      if (SCANPDS->sio.scan_scanner_type == UMAX_UC630) {
	switch (s_io->scan_image_mode) {
	case BINARY_MONOCHROME:
	case DITHERED_MONOCHROME:
	case GRAYSCALE:
	case RED:
	case GREEN:
	case BLUE:
	  break;
	default:
	  return (EINVAL);
	}
      } else {			/* scanner is a UMAX_UG630 */
	switch (s_io->scan_image_mode) {
	case BINARY_MONOCHROME:
	case DITHERED_MONOCHROME:
	case GRAYSCALE:
	  break;
	default:
	  return (EINVAL);
	}
      }

      break;			/* esac UMAX UC630 */

    case FUJITSU:
      /* size constraints... (these values are taken from the OEM manual) */

      if (s_io->scan_window_width <= 0				||
	  s_io->scan_x_origin + s_io->scan_window_width > 14592	|| /* 12.16" */
	  s_io->scan_window_length <= 0				||
	  s_io->scan_y_origin + s_io->scan_window_length > 20736)  /* 17.28" */
	return (EINVAL);

      /* Fujitsu can only handle 4 Meg windows max */

      if ((s_io->scan_window_width / 1200) * s_io->scan_x_resolution *
	  (s_io->scan_window_length / 1200) * s_io->scan_y_resolution
	  > (s_io->scan_image_mode == GRAYSCALE ? 4063232 : 32505856))
	return (EINVAL);

      switch (s_io->scan_x_resolution) {
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io->scan_y_resolution) {
      case 200:
      case 240:
      case 300:
      case 400:
	break;
      default:
	return (EINVAL);
      }

      switch (s_io->scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac FUJITSU */

    case RICOH_IS410:
    case IBM_2456:

      /* size constraints... */

      if (s_io->scan_window_width <= 0				||
	  s_io->scan_x_origin + s_io->scan_window_width > 13800	|| /* 11.5" */
	  s_io->scan_window_length <= 0				||
	  s_io->scan_y_origin + s_io->scan_window_length > 20400)    /* 17" */
	return (EINVAL);

      /* resolution (dpi)... */

      if (s_io->scan_x_resolution < 100 || s_io->scan_x_resolution > 400 ||
	  s_io->scan_y_resolution < 100 || s_io->scan_y_resolution > 400)
	return (EINVAL);

      switch (s_io->scan_image_mode) {
      case BINARY_MONOCHROME:
      case DITHERED_MONOCHROME:
      case GRAYSCALE:
	break;
      default:
	return (EINVAL);
      }

      break;			/* esac RICOH_IS410 (IBM_2456) */
    }

    /****  change the PDS to the new values  ***/

    bcopy(s_io, &SCANPDS->sio, sizeof(struct scan_io));

    /****  do what it takes to actualize the new values  ***/

    switch (SCANPDS->sio.scan_scanner_type) {
    case RICOH_IS410:
    case IBM_2456:
    case UMAX_UC630:
    case UMAX_UG630:
    case FUJITSU:
    case HP_SCANJET_IIC:

      /****  send a SCSI_SET_WINDOW ***/
      return (do_set_window(devp));
    }
    DPRINTF(3)("scanioctl: should not be here\n");
    return(ENOTTY);

  case SCAN_GET_PIXEL_SIZE:
    pix_size = (struct scan_pixel_sizes *)data;

    if ((SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	 SCANPDS->sio.scan_scanner_type == UMAX_UG630) &&
	SCANPDS->sio.scan_image_mode >= GRAYSCALE)
      pix_size->pix_width = calc_umax_row_len(SCANPDS->sio.scan_x_resolution,
					     SCANPDS->sio.scan_window_width);
    else {
      pix_size->pix_width = (SCANPDS->sio.scan_window_width
			    * SCANPDS->sio.scan_x_resolution
			    / 1200);

      if (SCANPDS->sio.scan_image_mode < GRAYSCALE) {
	if (SCANPDS->sio.scan_scanner_type == FUJITSU ||/* pad to byte boundary */
	    SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC)
	  pix_size->pix_width = (pix_size->pix_width + 7) / 8;
	else			                /* truncate to byte boundary */
	  pix_size->pix_width /= 8;
	pix_size->pix_width *= 8;
      }
    }
    pix_size->pix_height = (SCANPDS->sio.scan_window_length
			   * SCANPDS->sio.scan_y_resolution
			   / 1200);
    return (0);
    break;

  case SCAN_USE_ADF:

    if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	SCANPDS->sio.scan_scanner_type == UMAX_UG630) {
      if (umax_has_paper_in_adf(devp))
	return (0);
      else
	return(EIO);
    }

    if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC) {
      if (hp_has_paper_in_adf(devp))
	return (0);
      else
	return(EIO);
    }

    if (SCANPDS->sio.scan_scanner_type == RICOH_IS410 ||
	SCANPDS->sio.scan_scanner_type == IBM_2456    ||
	SCANPDS->sio.scan_scanner_type == RICOH_IS50) {

      SCANPDS->adf_state = ADF_ARMED;

      if (rc = do_mode_select(devp)) {
	SCANPDS->adf_state = ADF_UNUSED;
	return (rc);
      }
    }

    do_object_position(devp, UNLOAD_ADF); /* "scanner never fails to unload" */

    if (rc = do_object_position(devp, LOAD_ADF))
      SCANPDS->adf_state = ADF_UNUSED;
    else
      SCANPDS->adf_state = ADF_ARMED;

    return (rc);

  default:
    /*
     * Unknown ioctl
     */
    DPRINTF(1)("scanioctl: %s%d: unknown ioctl 0x%x\n", DNAME, DUNIT, cmd);
    return(ENOTTY);
  }
}

/*
 * Send a command to the device.
 * Called by scanioctl, only handles simple commands with no dma.
 * XXX To add a new command, add a new ioctl to scanioctl and
 *     call this routine from it.
 */
static int scansimple(devp, cmd, arg)
     register struct scsi_device *devp;
     int cmd;	/* Actual SCSI command */
     int arg;	/* command argument, put into byte 4 of the CDB */
{
  struct scsi_pkt *pkt, *rqpkt = SCANPDS->un_rsp;
  int err;

  DPRINTF(2)("scansimple: cmd = 0x%x, arg = 0x%x\n", cmd, arg);

  /*
   * Get a pkt with no dma resources
   */
  if (!(pkt = scsi_pktalloc(&devp->sd_address, CDB_GROUP0, 1, SLEEP_FUNC))) {
    return(ENOMEM);
  }
  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g0(pkt, devp, FLAG_NOINTR, cmd, 0, arg);

  if (scsi_poll(pkt)) {
    DPRINTF(1)("%s%d: Target not responding to <%s> cmd\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(pkt)->scc_cmd, scan_cmds));
    scsi_resfree(pkt);
    return(EIO);
  }

  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(pkt)) {
  case STATUS_GOOD:	/* command succeeded */
    err = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n",
		 DNAME, DUNIT, scsi_cmd_decode(CDBP(pkt)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, pkt, SCANERR_RETRYABLE);
    }
    err = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(pkt)->scc_cmd, scan_cmds),
	       SCBP_C(pkt));
    err = EIO;
    break;
  }
  scsi_resfree(pkt);
  return(err);
}


/*
 * General command handler. Do any command, with or without DMA.
 *
 * XXX	The thread is that this routine is called by scanioctl; it
 *	sets up the struct buf and calls make_scan_cmd and scanstrategy to
 *	make and queue the command. Scanstart then dequeues the command
 *	and calls pkt_transport to send the command. Eventually scanintr
 *	is called. It figures out whether the command worked OK and what
 *	to do. If the command worked, scanintr calls scandone which wakes
 *	us up. If the user wanted data we DVMA it into a local buffer
 *	(un_tmpbuf) and copy it into the users buffer.
 *
 * XXX	For the generic ioctl SCANIOC_CMD, the command data (a struct
 *	scancmd) is saved in our private data structure. Scan_make_cmd
 *	spots this and uses the cdb from that instead of creating one.
 *	Note that this means that if the if the request is for a data
 *	transfer larger than SCAN_MAXREC, this routine won't work because
 *	physio will split the request and call scanstrategy more than once,
 *	and after the first time the CDB will be wrong. Actually the DVMA
 *	data size is limited to the size of un_tmpbuf. Use scanread
 *	for large transfers.
 *
 * XXX	To add a new command, add a new ioclt to scanioctl calling this
 *	routine, and add the new SCSI command to scan_make_cmd to fill in
 *	the CDB.
 */
static int scancmd(dev, cmd, data)
     dev_t	dev;
     int	cmd;
     caddr_t	data;
{
  register struct scsi_device *devp = scanunits[SCANUNIT(dev)];
  register struct buf *bp = SCANPDS->un_sbufp;
  register struct scsi_pkt *pkt;
  struct scancmd	*args;
  register int	s, err;

  /*
   * Sanity check of arguments
   */
  args = (struct scancmd *)data;

  DPRINTF(2)("scancmd: %s%d Command = 0x%x\n", DNAME, DUNIT, cmd);
  if (scan_debug > 2) {
    printf("scancmd: args = ");
    printf("\tCDB: ");
    clean_print(args->sc_cdb, sizeof(union scsi_cdb));
    printf("\n");
    printf("\tDVMA Address: 0x%x\n", args->sc_bufaddr);
    printf("\tDVMA Length: %d\n", args->sc_buflen);
    printf("\tTimeout: %d sec\n", args->sc_timeout);
    printf("\tDVMA Direction: (%d) %s\n", args->sc_rw,
	   (args->sc_rw == B_READ) ? "read" :"write");
  }

  if ((args->sc_bufaddr && !args->sc_buflen) ||
      (!args->sc_bufaddr && args->sc_buflen)) {
    DPRINTF(1)("scancmd: %s%d bad buf addr/len\n", DNAME, DUNIT);
    return(EINVAL);
  }

  if (args->sc_buflen > UN_TMPBUF_SIZE) {
    DPRINTF(1)("scancmd: %s%d transfer (%d bytes) too large\n",
	       DNAME, DUNIT, args->sc_buflen);
    return(EINVAL);
  }

  /*
   * Get the special buf
   */
  s = splr(scanpri);
  while (bp->b_flags & B_BUSY) {
    bp->b_flags |= B_WANTED;
    (void) sleep((caddr_t) bp, PRIBIO);
  }
  bzero((caddr_t) bp, sizeof (struct buf));
  bp->b_flags = B_BUSY;
  (void) splx(s);

  bp->b_dev = dev;
  if (args->sc_buflen) {	/* user wants data */
    bp->b_bcount = args->sc_buflen;
    bp->b_un.b_addr = SCANPDS->un_tmpbuf;
    bp->b_flags |= args->sc_rw;
  }
  /*
   * scan_make_cmd needs to know what command this is to set up the
   * CDB, and it also needs the dma parameters.
   */
  bp->b_forw = (struct buf *)cmd;
  bcopy((caddr_t)args, (caddr_t)&SCANPDS->un_ucmd, sizeof(struct scancmd));

  scan_make_cmd(devp, bp, SLEEP_FUNC);

  scanstrategy(bp);
  /*
   * Wait for command to complete
   */
  s = splr(scanpri);
  while (!(bp->b_flags & B_DONE)) {
    sleep((caddr_t)bp, PRIBIO);
  }
  (void)splx(s);

  err = geterror(bp);

  /*
   * release resources
   */
  scsi_resfree(BP_PKT(bp));	/* allocated by scan_make_cmd */

  s = splr(scanpri);
  if (bp->b_flags & B_WANTED)
    wakeup((caddr_t)bp);
  bp->b_flags &= ~(B_BUSY|B_WANTED);
  (void)splx(s);

  /*
   * If there was DVMA, copyout to the user's buffer
   */
  if (args->sc_buflen) {
    copyout(SCANPDS->un_tmpbuf, args->sc_bufaddr, args->sc_buflen);
  }

  return(err);
}

/*
 * scanstart - send the next command to the host adapter for transmission.
 * Call scan_make_cmd (if this is not a special cmd from scancmd) to
 * create the SCSI CDB and allocate any needed DMA resources.
 */
int scanstart(devp)
     register struct scsi_device *devp;
{
  register struct buf *bp;
  int	scan_retry();
  int   err;

  DPRINTF(2)("scanstart: %d requests queued\n", SCANPDS->un_bufcnt);

  if (SCANPDS->un_active) {
    printf("scanstart: called when already busy\n");
    return;
  }

  if ((bp = SCANPDS->un_quehd) == NULL) {
    DPRINTF(1)("scanstart: nothing to do\n");
    return;
  }

  /*
   * If we need resources, call scan_make_cmd to get them. If there
   * are none available, scanretry will be called (automagically)
   * when they become available.
   */
  if (!BP_PKT(bp)) {
    DPRINTF(3)("scanstart: calling scan_make_cmd\n");
    scan_make_cmd(devp, bp, scan_retry);
    if (!BP_PKT(bp)) {
      newstate(SCAN_STATE_RWAIT);
      return;
    }
  }

  /*
   * Adjust the queues: move this command from pending to active.
   */
  SCANPDS->un_active = bp;
  SCANPDS->un_quehd = bp->av_forw;
  bp->av_forw = 0;
  SCANPDS->un_bufcnt--;

  /*
   * XXX Linked command support would go here
   */

  /*
   * Send the command
   */
  DPRINTF(3)("scanstart: calling pkt_transport\n");
  err = pkt_transport(BP_PKT(bp));
  DPRINTF(3)("scanstart: pkt_transport returned %d\n", err);
  if (err == 0) {
    printf("%s%d: transport rejected\n", DNAME, DUNIT);
    bp->b_resid = bp->b_bcount;
    bp->b_flags |= B_ERROR;
    bp->b_error = EIO;
    scandone(devp);
  }
  if (err == -1) {
    DPRINTF(1)("scanstart: bad packet\n");
  }

  return;
}

/*
 * Retry a command that's been waiting for resources.
 * Note the difference between this and scanrestart.
 */
static int scan_retry()
{
  register int	i, s;
  register struct scsi_device *devp;

  DPRINTF(2)("scanretry: %s%d\n", DNAME, DUNIT);

  s = splr(scanpri);
  for (i = 0; i < nscans; i++) {
    devp = scanunits[i];
    if (devp && devp->sd_present && SCANPDS && SCANPDS->un_attached) {
      DPRINTF(1)("scan_retry: retrying %s%d\n", DNAME, DUNIT);
      scanstart(devp);
      if (SCANPDS->un_state == SCAN_STATE_RWAIT) {
	(void)splx(s);
	DPRINTF(1)("scan_retry: %s%d retry failed\n",
		   DNAME, DUNIT);
	return(0);
      }
      DPRINTF(1)("scan_retry: %s%d succeeded\n", DNAME, DUNIT);
    }
  }
  (void)splx(s);
  return(1);
}

/*
 * Clean up after previous command and start the next one
 */
static void scandone(devp)
     register struct scsi_device *devp;
{
  register struct buf *bp;

  DPRINTF(2)("scandone: %s%d\n", DNAME, DUNIT);
  bp = SCANPDS->un_active;
  SCANPDS->un_active = NULL;	/* this command is no longer active */

  if (SCANPDS->sio.scan_scanner_type != FUJITSU)
    SCANPDS->window_size -= bp->b_bcount;

  if (SCANPDS->un_quehd) {
    DPRINTF(2)("scandone: calling scanstart for %s%d\n", DNAME, DUNIT);
    scanstart(devp);
  }

  if (bp != SCANPDS->un_sbufp) {
    scsi_resfree(BP_PKT(bp));
    iodone(bp);
  } else {
    bp->b_flags |= B_DONE;
    wakeup((caddr_t)bp);
  }
  return;
}

/*
 * scan_make_cmd - create the scsi_pkt to be sent to the host adapter
 * Scsi_resalloc allocates a comand buffer and if needed also sets up
 * for DMA (mapping and locking memory).
 * XXX If you are adding a special command, point allocbp at
 *     the passed bp if there will be dma, otherwise leave it
 *     as zero.
 */
static void scan_make_cmd(devp, bp, func)
     register struct scsi_device *devp;
     register struct buf *bp;
     int (*func)();
{
  register struct scsi_pkt *pkt;
  struct buf *allocbp = 0;	/* buf for dma resources, used by
				 * scsi_resalloc for special cmd */
  long	tval = 30;
  int	cmd, flags, fixbit, count, blkno;

  flags = (scsi_options & SCSI_OPTIONS_DR) ? 0 : FLAG_NODISCON;

  if (bp != SCANPDS->un_sbufp) {	/* not a special command */
    if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC) {
      pkt = scsi_resalloc(&devp->sd_address, CDB_GROUP0, 1, (opaque_t)bp, func);
      if (pkt == (struct scsi_pkt *) 0) {
	return;
      }
      makecom_g0_s(pkt, devp, flags, SCMD_READ, bp->b_bcount, 0);
    } else {
      pkt = scsi_resalloc(&devp->sd_address, CDB_GROUP1, 1, (opaque_t)bp, func);
      if (pkt == (struct scsi_pkt *) 0) {
	return;
      }
      makecom_g1(pkt, devp, flags, SCMD_READ_G1, 0, bp->b_bcount);
    }
    DPRINTF(2)("%s%d Read, count=%d\n", DNAME, DUNIT, bp->b_bcount);

  } else {			/* internal command from scancmd */
    cmd = (int) bp->b_forw;

    switch(cmd) {
    case SCMD_UCMD:	/* User specified command, from scancmd */
      tval = SCANPDS->un_ucmd.sc_timeout;
      if (SCANPDS->un_ucmd.sc_buflen > 0) {
	allocbp = bp;	/* Have DVMA */
      }
      break;
    default:
      /*
       * Naughty naughty!
       * scan_make_cmd only works for commands it knows
       * about. For special commands using un_sbufp it
       * may only be called from scancmd. Can't recover
       * from here.
       */
      printf("%s%d: unkown cmd 0x%x\n", DNAME, DUNIT, cmd);
      panic("scan_make_cmd, bad command\n");
    }
    /*
     * Allocate a packet. This also sets up the
     * DMA mapping
     */
    pkt = scsi_resalloc(&devp->sd_address, CDB_GROUP1, 1,
			(opaque_t)allocbp, SLEEP_FUNC);

    /*
     * If this is a user specified command, use the cdb the
     * user gave us, otherwise make one.
     */
    if (cmd == SCMD_UCMD) {
      bcopy((caddr_t)&SCANPDS->un_ucmd.sc_cdb,
	    (caddr_t)pkt->pkt_cdbp, sizeof(union scsi_cdb));
    } else {
      if (devp->sd_inq->inq_dtype == DTYPE_DIRECT) {
	makecom_g0(pkt, devp, flags, cmd, (int)blkno, count);
      } else {
	makecom_g0_s(pkt, devp, flags, cmd, count, fixbit);
      }
    }
  }
  pkt->pkt_time = tval;
  pkt->pkt_comp = scanintr;	/* completion routine */
  pkt->pkt_private = (opaque_t) bp;
  BP_PKT(bp) = pkt;
  pkt->pkt_pmon = -1;		/* no performance measurement */
}


/*
 * Command completion routine. Figure out what to do next based on
 * the returned status of the scsi command
 * It's declared static because it is not called directly, but via the
 * pkt_comp field of the scsi packet.
 */
static void scanintr(pkt)
     struct scsi_pkt *pkt;
{
  struct buf *bp;
  struct scsi_device *devp;
  int	status;

  DPRINTF(3)("scanintr: called\n");

  if (!pkt) {
    printf("scanintr: Null pkt!\n");
    return;
  }

  if (pkt->pkt_flags & FLAG_NOINTR) {
    printf("scanintr: Non-interrupting cmd!\n");
    return;
  }

  /*
   * If the buf is not in the packet, things are very wrong.
   * strictly, we should panic here, but a scan failing isn't
   * necessarily reason to crash.
   */
  if (!(bp = (struct buf *)pkt->pkt_private)) {
    printf("scanintr: Null buffer!\n");
    return;
  }

  if (!(devp = scanunits[SCANUNIT(bp->b_dev)])) {
    printf("scanintr: Null devp!\n");
    return;
  }

  /*
   * Only one command active at a time. If the completed command
   * isn't the command we're waiting for, we have a problem! The
   * above comment about panicking applies.
   */
  if (bp != SCANPDS->un_active) {
    printf("scanintr: %s%d: buf not on queue!\n", DNAME, DUNIT);
    return;
  }

  if (pkt->pkt_reason != CMD_CMPLT) {
    status = scan_incomplete(devp);
  } else {
    /*
     * Command completed successfully.
     * What were we doing?
     */
    if (SCANPDS->un_state == SCAN_STATE_SENSING) {
      /*
       * We were runnning a request sense command.
       * Restore previous state and find what to
       * do next.
       */
      SCANPDS->un_state = SCANPDS->un_last_state;
      SCANPDS->un_last_state = SCAN_STATE_SENSING;
      status = scan_handle_sense(devp);
    } else {
      /*
       * Normal command completion. Check for errors.
       */
      status = scan_check_error(devp);
    }
  }

  switch(status) {
  case COMMAND_DONE_ERROR:
    SCANPDS->un_err_severe = SCANERR_FATAL;
    SCANPDS->un_err_resid = bp->b_resid = bp->b_bcount;
    bp->b_flags |= B_ERROR;
    /* **FALL THROUGH** */
  case COMMAND_DONE:
    scandone(devp);
    break;
  case QUE_SENSE:			/* need to do a request sense */
    DPRINTF(2)("scanintr: requesting sense\n");
    newstate(SCAN_STATE_SENSING);
    SCANPDS->un_rsp->pkt_private = (opaque_t)bp;
    bzero((caddr_t)devp->sd_sense, SENSE_LENGTH);
    if (pkt_transport(SCANPDS->un_rsp) == 0) {
      /*
       * request sense transport failed. Big problem.
       * Should panic, but it's nicer
       * not to, so just close the unit.
       */
      printf("%s%d: Couldn't send Request Sense, closing unit\n", DNAME, DUNIT);
      (void)scanclose(bp->b_dev);
    }
    break;
  case QUE_COMMAND:		/* retry command */
    DPRINTF(2)("scanintr: retrying command\n");
    if (pkt_transport(BP_PKT(bp)) == 0) {
      printf("scanintr: command requeue failed\n");
      SCANPDS->un_err_resid = bp->b_resid = bp->b_bcount;
      bp->b_flags |= B_ERROR;
      scandone(devp);
    }
    break;
  case JUST_RETURN:
    break;
  }
  return;
}


/*
 * Clean up after an incomplete command
 */
static int scan_incomplete(devp)
     struct scsi_device *devp;
{
  struct scsi_pkt	*pkt;
  struct buf	*bp;
  int		rval;

  bp = SCANPDS->un_active;
  rval = COMMAND_DONE_ERROR;

  if (SCANPDS->un_state == SCAN_STATE_SENSING) {
    /*
     * We're sensing, retry the request sense
     */
    DPRINTF(2)("scan_incomplete: %s%d retry request sense\n",
	       DNAME, DUNIT);
    pkt = SCANPDS->un_rsp;
    if (SCANPDS->un_retries++ < SCAN_RETRY_COUNT) {
      rval = QUE_SENSE;
    }
  } else {
    /*
     * Not sensing, retry the command
     */
    pkt = BP_PKT(bp);
    if (SCANPDS->un_retries++ < SCAN_RETRY_COUNT) {
      DPRINTF(2)("scan_incomplete: %s%d retry command\n",
		 DNAME, DUNIT);
      rval = QUE_COMMAND;
    }
  }

  if (pkt->pkt_state == STATE_GOT_BUS && rval == COMMAND_DONE_ERROR) {
    /*
     * We've exceeded the retry count, and the host adapter
     * got the bus but couldn't get the target.
     */
    printf("%s%d: target not responding to selection\n",
	   DNAME,DUNIT);
  }

  return(rval);
}


/*
 * check the status of the request sense command
 */
static int scan_handle_sense(devp)
     register struct scsi_device *devp;
{
  struct scsi_pkt	*pkt = BP_PKT(SCANPDS->un_active),
  *rsp = SCANPDS->un_rsp;
  struct buf	*bp = SCANPDS->un_active;
  int		rval, sdl, severity;
  void		scanrestart();

  DPRINTF(2)("scan_handle_sense: %s%d\n", DNAME, DUNIT);
  if (SCBP(rsp)->sts_busy) {
    DPRINTF(1)("%s%d%: busy unit on request sense\n", DNAME, DUNIT);
    if (SCANPDS->un_retries++ < SCAN_RETRY_COUNT) {
      timeout(scanrestart, (caddr_t)devp, SCANTIMEOUT);
      rval = JUST_RETURN;
    }
    return(rval);
  }

  if (SCBP(rsp)->sts_chk) {
    /*
     * Got a 'Check Condition' from a Request Sense command
     * - bad news!
     */
    DPRINTF(1)("%s%d%: check condition on request sense\n", DNAME, DUNIT);
    return(COMMAND_DONE_ERROR);
  }

  sdl = SENSE_LENGTH - rsp->pkt_resid;

  if ((rsp->pkt_state & STATE_XFERRED_DATA) == 0 || sdl == 0) {
    DPRINTF(1)("%s%d%: not enough Request Sense data\n", DNAME, DUNIT);
    return(COMMAND_DONE_ERROR);
  }

  /*
   * If we got here, things are going OK. Look at the
   * request sense data.
   *
   * If the sense data length is less than the length of the
   * extended sense struct, check that we got enough data for
   * non-extended sense; if not give up.
   */

  if (sdl < sizeof (struct scsi_extended_sense)) {
    if (sdl < sizeof (struct scsi_sense)) {
      DPRINTF(1)("%s%d%: not enough sense data\n", DNAME, DUNIT);
      return(COMMAND_DONE_ERROR);
    }
    if (devp->sd_sense->es_class != CLASS_EXTENDED_SENSE) {
      /*
       * XXX Non-extended sense classes (0-6) are
       * vendor-unique. Write a routine to decode
       * the error code and call it here. For now,
       * print out the undecoded information.
       */
      printf("%s%d: undecoded sense info:", DNAME, DUNIT);
      clean_print((caddr_t)devp->sd_sense, sdl);
      printf("; assuming a fatal error\n");
      return (COMMAND_DONE_ERROR);
    }
  }

  /*
   * We got enough data for extended sense
   * XXX	es_code isn't very useful - it's usually 0.
   *	Most devices use the additional sense info
   *	bytes, it's more useful to put the vendor
   *	unique error code into un_err_code.
   */

  SCANPDS->un_status = devp->sd_sense->es_key;
  SCANPDS->un_err_code = devp->sd_sense->es_code;
  DPRINTF(2)("Extended sense, status=%x, code=%x\n",
	     SCANPDS->un_status, SCANPDS->un_err_code);

#if 0
  /*
   * The information bytes' meaning depends on the device.
   * For direct access devices it's the unsigned logical block
   * address, for sequential access, printer and processor
   * devices it's the residue (requested - actual lengths).
   */
  if (devp->sd_sense->es_valid) {
    SCANPDS->un_err_blkno = (devp->sd_sense->es_info_1 << 24) |
      (devp->sd_sense->es_info_2 << 16) |
	(devp->sd_sense->es_info_3 << 8)  |
	  (devp->sd_sense->es_info_4);
  } else {
    SCANPDS->un_err_blkno = bp->b_blkno;
  }
#endif

  /* FUJITSU EOM (EOF) detection */
  if (SCANPDS->sio.scan_scanner_type == FUJITSU)
    if (devp->sd_sense->es_eom) {
      SCANPDS->window_size = 0;
      return (COMMAND_DONE);
    }

  switch(SCANPDS->un_status) {
  case KEY_NO_SENSE:
    severity = SCANERR_RETRYABLE;
    rval = QUE_COMMAND;
    break;
  case KEY_RECOVERABLE_ERROR:
    severity = SCANERR_RECOVERED;
    rval = COMMAND_DONE;
    break;
  case KEY_UNIT_ATTENTION:
    severity = (SCANPDS->un_state <= SCAN_STATE_OPENING) ?
      SCANERR_INFORMATIONAL : SCANERR_FATAL;
    rval = COMMAND_DONE_ERROR;
    break;
  case KEY_ABORTED_COMMAND:
    severity = SCANERR_RETRYABLE;
    rval = QUE_COMMAND;
    break;
  case KEY_NOT_READY:
  case KEY_MEDIUM_ERROR:
  case KEY_HARDWARE_ERROR:
  case KEY_ILLEGAL_REQUEST:
  case KEY_WRITE_PROTECT:
  case KEY_BLANK_CHECK:
  case KEY_VENDOR_UNIQUE:
  case KEY_COPY_ABORTED:
  case KEY_EQUAL:
  case KEY_VOLUME_OVERFLOW:
  case KEY_MISCOMPARE:
  case KEY_RESERVED:
    severity = SCANERR_FATAL;
    rval = COMMAND_DONE_ERROR;
    break;
  default:
    /*
     * Undecoded sense key. Hope that retrying will fix
     * the problem, give up if not.
     */
    printf("%s%d: Sense Key 0x%x\n", DNAME, DUNIT, SCANPDS->un_status);
    if (SCANPDS->un_retries++ < SCAN_RETRY_COUNT) {
      severity = SCANERR_RETRYABLE;
      rval = QUE_COMMAND;
    } else {
      severity = SCANERR_FATAL;
      rval = COMMAND_DONE_ERROR;
    }
    break;
  }

  if ((scan_debug > 2) || (severity >= scan_error_reporting)) {
    scanerrmsg(devp, pkt, severity);
  }
  return(rval);
}

/*
 * Check a completed command for errors
 */
static int scan_check_error(devp)
     register struct scsi_device *devp;
{
  struct buf	*bp;
  struct scsi_pkt	*pkt;
  int		rval;

  bp = SCANPDS->un_active;
  pkt = BP_PKT(bp);

  DPRINTF(2)("scan_check_error:\n");

  if (SCBP(pkt)->sts_busy) {
    printf("%s%d: unit busy\n",DNAME,DUNIT);
    if (SCANPDS->un_retries++ < SCAN_RETRY_COUNT) {
      DPRINTF(1)("%s%d: retrying (%d)\n",
		 DNAME, DUNIT, SCANPDS->un_retries);
      timeout(scanrestart, (caddr_t)devp, SCANTIMEOUT);
      rval = JUST_RETURN;
    } else {
      DPRINTF(1)("%s%d: target busy too long\n", DNAME,DUNIT);
      rval = COMMAND_DONE_ERROR;
    }
  } else if (SCBP(pkt)->sts_chk) {
    DPRINTF(1)("%s%d: check condition\n", DNAME, DUNIT);
    rval = QUE_SENSE;
  } else {
    DPRINTF(2)("%s%d: Command done\n", DNAME, DUNIT);
    rval = COMMAND_DONE;
  }
  return(rval);
}

/*
 * Restart a command. Used to retry a command or request sense after
 * a BUSY status. Note the difference between this and scan_retry.
 */
static void scanrestart(arg)
     caddr_t arg;
{
  struct scsi_device *devp = (struct scsi_device *) arg;
  struct buf	*bp;
  struct scsi_pkt	*pkt;
  register int	s;

  s = splr(scanpri);

  if ((bp = SCANPDS->un_active) == NULL) {
    printf("%s%d: restart failed - no active command\n",
	   DNAME, DUNIT);
  }
  else {
    if (SCANPDS->un_state == SCAN_STATE_SENSING) {
      pkt = SCANPDS->un_rsp;
    } else {
      pkt = BP_PKT(bp);
    }
    if (pkt_transport(pkt) == 0) {
      printf("%s%d: restart transport failed\n", DNAME,DUNIT);
      bp->b_resid = bp->b_bcount;
      bp->b_flags |= B_ERROR;
      scandone(devp);
    }
  }
  (void)splx(s);
  return;
}

/**************************************************
 *
 * Error message handling routines
 *
 **************************************************/


static char *scan_cmds[] = {
  "\000test unit ready",	/* 0x00 */
  "\003request sense",		/* 0x03 */
  "\007reassign",		/* 0x07 */
  "\022inquiry",		/* 0x12 */
  "\025mode select",		/* 0x15 */
  "\026reserve",		/* 0x16 */
  "\027release",		/* 0x17 */
  "\032mode sense",		/* 0x1a */
  "\033trigger scan",		/* 0x1b */
  "\044set window",		/* 0x24 */
  "\050read data",		/* 0x28 */
  "\061medium position",	/* 0x31 */
  "\064get buffer status",      /* 0x34 */
  NULL,
};


/*
 * Print a readable error message
 */
static void scanerrmsg(devp, pkt, level)
     register struct scsi_device *devp;
     struct scsi_pkt	*pkt;
     int level;
{
  static char *error_classes[] = {
    "All          ", "Unknown      ", "Informational",
    "Recovered    ", "Retryable    ", "Fatal        ",
  };

  printf("%s%d: Error on command '%s'\n", DNAME, DUNIT,
	 scsi_cmd_decode(CDBP(pkt)->scc_cmd, scan_cmds));
  printf("    \tError Level: %s",
	 error_classes[level]);
  printf("    \tSense Key: %s\n", sense_keys[devp->sd_sense->es_key]);

  if (devp->sd_sense->es_code) {
   printf("\tExtended error code 0x%x, 0x%x\n",
	  devp->sd_sense->es_code, devp->sd_sense->es_qual_code);
  }
  return;
}

/*
 * Print a buffer readably
 */
static void  clean_print(p, l)
     char *p;
     int l;
{
  char a[2];
  register unsigned i = 0, c;
  while (i < l) {
    if((c = *p++) < ' ' || c >= 0177)
      printf("\\%o",c&0xff);
    else {
      a[0] = c; a[1] = 0;
      printf("%s",a);
    }
    i++;
  }
}

/*
 * routine to wakeup find_umax
 */
wake_find_umax(event)
     caddr_t event;
{
  wakeup(event);
}

/*
 * since UMAX scanners return an "operation complete" message even before
 * finishing an unload or trigger operation, poll the scanner with TEST
 * UNIT READY until it's really done and comes back online.
 */
static void find_umax(devp)
     register struct scsi_device *devp;
{
  int tries = 50;

  while (tries--) {
    if (!scansimple(devp, SCMD_TEST_UNIT_READY, 0))
      break;
    timeout(wake_find_umax, devp, 200);
    sleep(devp, scanpri);
  }
}


static int calc_umax_row_len(dpi, ww)
     int dpi;
     int ww;
{
  int st[301];
  int i;
  int rowB = 0;

  for (i = 1; i <= 300; i++)
    st[i] = 1;

  for (i = 1; i <= 300 - dpi; i++)
    st[i * 300 / (300 - dpi)] = 0;

  for (i = 1; i <= (ww % 1200) / 4; i++) {
    if (st[i])
      rowB++;
  }

  return ((ww / 1200) * dpi + rowB);
}

static int hp_has_paper_in_adf(devp)
     register struct scsi_device *devp;
{
  char escape_codes[20];

  sprintf(escape_codes, "%c*s1027E", '\033');

  if (hp_send_cmd(devp, escape_codes))
    return (0);

  if (hp_recv_response(devp, escape_codes, 19))
    return (0);

  if (escape_codes[8] == '1')
    return (1);			/* paper present & ready */
  else
    return (0);			/* no paper, or ADF not ready */
}

static int umax_has_paper_in_adf(devp)
     register struct scsi_device *devp;
{
  struct scsi_pkt *inqp, *rqpkt = SCANPDS->un_rsp;
  caddr_t response;
  int rc;

  if (!(inqp = get_pktiopb(devp->sd_address, &response,
			 CDB_GROUP0, 1, 0x24, B_READ, SLEEP_FUNC))) {
    return (0);
  }

  /*
   * Make the CDB, set NOINTR because this is polled, we'll
   * wait for the command to complete
   */
  makecom_g0_s(inqp, devp, FLAG_NOINTR, SCMD_INQUIRY, 0x24, 0);

  if (scsi_poll(inqp)) {
    DPRINTF(1)("%s%d: Target not responding to SCMD_INQUIRY\n", DNAME, DUNIT);
    free_pktiopb(inqp, response, 0x24);
    return (0);
  }

#if 0
  /*
   * check the Status Completion Block
   */
  switch(SCBP_C(inqp)) {
  case STATUS_GOOD:	/* command succeeded */
    rc = 0;
    break;
  case STATUS_CHECK:	/* got check condition status */
    if (scsi_poll(rqpkt) || SCBP_C(rqpkt) != STATUS_GOOD) {
      DPRINTF(1)("%s%d: (cmd %s) Request Sense failed\n",
		 DNAME, DUNIT, scsi_cmd_decode(CDBP(inqp)->scc_cmd, scan_cmds));
    } else {
      scanerrmsg(devp, inqp, SCANERR_RETRYABLE);
    }
    rc = EIO;
    break;
  default:		/* command failed */
    DPRINTF(1)("%s%d: (cmd %s) command failed, status 0x%x\n",
	       DNAME, DUNIT, scsi_cmd_decode(CDBP(inqp)->scc_cmd, scan_cmds),
	       SCBP_C(inqp));
    rc = EIO;
    break;
  }
#endif

  rc = (response[1] == 1);

  free_pktiopb(inqp, response, 0x24);
  return (rc);
}

static u_char model_to_type_code(model)
     char *model;
{
  if (strncmp("UC630", model, 5) == 0)
    return (UMAX_UC630);

  if (strncmp("UG630", model, 5) == 0)
    return (UMAX_UG630);

  if (strncmp("M3096Gm", model, 7) == 0)
    return (FUJITSU);

  if (strncmp("IS410", model, 5) == 0)
    return (RICOH_IS410);

  if (strncmp("IS50", model, 4) == 0)
    return (RICOH_IS50);

  if (strncmp("2456", model, 4) == 0)
    return (IBM_2456);

  if (strncmp("C1750A", model, 6) == 0)
    return (HP_SCANJET_IIC);

  if (strncmp("C2500A", model, 6) == 0)
    return (HP_SCANJET_IIC);

  return (0);
}

static int calc_window_size(devp)
     register struct scsi_device *devp;
{
  /* Fujitsu M3096G uses EOM bit, doesn't use window_size */

  if (SCANPDS->sio.scan_scanner_type == FUJITSU)
    return (1);

  if (SCANPDS->sio.scan_image_mode < GRAYSCALE) { /* if monochrome */

    if (SCANPDS->sio.scan_scanner_type == HP_SCANJET_IIC) { /* if pad to byte boundary */

      /* this line fails if (wid * xres < 1200) */

      return ((((SCANPDS->sio.scan_window_width * SCANPDS->sio.scan_x_resolution / 1200)+7)/8)
	      * (SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution / 1200));

    } else {			/* if truncate to byte boundary */

      if (SCANPDS->sio.scan_window_width * SCANPDS->sio.scan_x_resolution < 9600 ||
	  SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution < 1200) {

	/* divide late because (wid:95 * res:100 / 9600 == 0) */

	return ((SCANPDS->sio.scan_window_width * SCANPDS->sio.scan_x_resolution *
		 SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution)
		/ 9600 / 1200);
      } else {

	/* divide early because (wid:13200 * res:400 * wid:20400 * res:400) overflows 2^32 */

	return ((SCANPDS->sio.scan_window_width * SCANPDS->sio.scan_x_resolution / 9600)
		* (SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution / 1200));
      }
    }
  } else {					  /* if grayscale or color */

    /* UMAX scaling algorithm makes it a special case */

    if (SCANPDS->sio.scan_scanner_type == UMAX_UC630 ||
	SCANPDS->sio.scan_scanner_type == UMAX_UG630) {
      return (calc_umax_row_len(SCANPDS->sio.scan_x_resolution, SCANPDS->sio.scan_window_width)
	      * ((SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution) / 1200));
    } else {
      return ((SCANPDS->sio.scan_window_width    * SCANPDS->sio.scan_x_resolution / 1200)
	      * (SCANPDS->sio.scan_window_length * SCANPDS->sio.scan_y_resolution / 1200)
	      * (SCANPDS->sio.scan_image_mode == COLOR ? 3 : 1));
    }
  }
}
#endif NSCAN > 0
